/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <sstream>
#include <cmath>
#include <vector>

#define DEF_WIDTH   22

using namespace std;

/*** definitions ***********************************************************/

static   string                           input_file_path;
static   bool                             input_file_path_spec     = false;
static   string                           output_file_path;
static   bool                             output_file_path_spec    = false;
static   vector< pair<string, string> >   mod_key_vals;
static   vector< pair<string, string> >   new_key_vals;
static   int                              key_width                = DEF_WIDTH;

void update_key_value(string & line);

/*** implementations *******************************************************/

int main(int argc, char ** argv)
{

  cout << "Simple EMD configuration file manipulator" << endl
       << "Kopyleft (k) 2007 Mehrtash Babadi (mehrtash@physics.sharif.edu)" << endl << endl;

  /* phase 1. mandatory switches */

  bool stat = true;  /* sufficient arguments */

  if (argc<2)
    stat = false;

  int ctr = 1;
  while (stat && (ctr < argc))
  {

    if ( (string(argv[ctr])=="-i") && (argc>ctr) )
    {
      ctr++;
      input_file_path = argv[ctr];
      input_file_path_spec = true;
    }
    else if ( (string(argv[ctr])=="-o") && (argc>ctr) )
    {
      ctr++;
      output_file_path = argv[ctr];
      output_file_path_spec = true;
    }
    else if ( (string(argv[ctr])=="-c") && (argc>(ctr+1)) )
    {
      pair<string, string> mod_key_val;
      ctr++;
      mod_key_val.first = argv[ctr];
      ctr++;
      mod_key_val.second = argv[ctr];
      mod_key_vals.push_back(mod_key_val);
    }
    else if ( (string(argv[ctr])=="-a") && (argc>(ctr+1)) )
    {
      pair<string, string> new_key_val;
      ctr++;
      new_key_val.first = argv[ctr];
      ctr++;
      new_key_val.second = argv[ctr];
      new_key_vals.push_back(new_key_val);
      if (new_key_val.first.length() > (key_width-2))
        key_width += 5;
    }
    else
      stat = false;

    ctr++;

  }

  if (!input_file_path_spec || !output_file_path_spec)
    stat = false;

  if (!stat)
  {
    cout << "usage: config-manip [OPTIONS]" << endl << endl
         << "       -i <input_file_path>   the path of the input file" << endl
         << "       -o <output_file_path>  the path of the output file" << endl
         << "       -a <key> <value>       add a new entry (@key, @value)" << endl
         << "       -c <key> <value>       change the @value of @key to @mod_value" << endl << endl;
    return EXIT_FAILURE;
  }

  /* phase 2. apply changes */

  ifstream inf(input_file_path.c_str());

  if (!inf)
  {
    cout << "config-manip: error opening the input file!" << endl;
    return EXIT_FAILURE;
  }

  ofstream outf(output_file_path.c_str());

  if (!outf)
  {
    cout << "config-manip: error creating the output file!" << endl;
    return EXIT_FAILURE;
  }

  string curr_line;
  while (!inf.eof())
  {

    getline(inf, curr_line);

    /* apply key-value modifications */
    update_key_value(curr_line);

    outf << curr_line << endl;

  }

  /* new entries */
  for (int i=0; i<new_key_vals.size(); i++)
    outf << left << setw(key_width) << new_key_vals[i].first << new_key_vals[i].second << endl;

  inf.close();
  outf.close();

  cout << "config-manip: modifications applied successfully!" << endl << endl;
  return EXIT_SUCCESS;

}

/***************************************************************************/

void update_key_value(string & line)
{

  const char delim = ' ';

  for (int i=0; i<mod_key_vals.size(); i++)
  {

    int pos = line.find(mod_key_vals[i].first);

    if (pos!=string::npos)
    {
      if (line.length()==(pos+mod_key_vals[i].first.length())) /* the line ends */
      {
        if (pos>0) if (line[pos-1]!=delim) continue;
        line += "    " + mod_key_vals[i].second;
      }
      else
      {
        int cpos = pos + mod_key_vals[i].first.length();
        if (line[cpos]!=delim) continue;
        if (pos>0) if (line[pos-1]!=delim) continue;
        while ((cpos<line.length()) && line[cpos]==delim)
          cpos++;
        line.replace(cpos, line.length(), mod_key_vals[i].second);
      }
    }

  }

}

