/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * todo:                                                                   *
 * - initial velocity setup                                                *
 * - adaptive neighbor list update procedure                               *
 * - CM angular momentum = 0                                               *
 ***************************************************************************/

#include "dynamics.h"

using namespace std;

/***************************************************************************/

emd::uni::dyn::integrator_type   emd::uni::dyn::integrator            =  NVT_MILLER_SYMPLECTIC_NHC;
double                           emd::uni::dyn::dt                    =  1.0; /* fs */
//{{{restart : declaration and initiallization of t
double                           emd::uni::dyn::t                     =  0;
//}}}
long int                         emd::uni::dyn::N_ignore_steps        =  0;
long int                         emd::uni::dyn::N_steps               =  1000;

int                              emd::uni::dyn::nl_update_interval    =  10;
int                              emd::uni::dyn::stats_interval        =  100;

bool                             emd::uni::dyn::energy_conservator    =  false;

double                           emd::uni::dyn::goal_energy;          /* NOTE auto */
double                           emd::uni::dyn::goal_pressure         =  1e-3; /* kcal.mol^-1.ang^-1. */
double                           emd::uni::dyn::goal_temperature      =  300;  /* kelvin */

bool                             emd::uni::dyn::fix_cm                =  true;
bool                             emd::uni::dyn::fix_T;                /* NOTE auto */
bool                             emd::uni::dyn::fix_P;                /* NOTE auto */
bool                             emd::uni::dyn::reset_thermostat      =  true;
bool                             emd::uni::dyn::reset_barostat        =  true;
bool                             emd::uni::dyn::log_to_file           =  false;
bool                             emd::uni::dyn::log_to_stdout         =  true;
bool                             emd::uni::dyn::log_to_stats_file     =  false;
bool                             emd::uni::dyn::log_force_on_surfs    =  true;

/***************************************************************************/

double                           emd::uni::dyn::init_temperature;     /* NOTE must be supplied by user */
double                           emd::uni::dyn::final_temperature;    /* NOTE must be supplied by user */
double                           emd::uni::dyn::final_concentration;  /* NOTE must be supplied by user */

emd::uni::dyn::temp_drift_type   emd::uni::dyn::temp_drift;           /* NOTE must be supplied by user */
bool                             emd::uni::dyn::temperature_drift     =  false;
bool                             emd::uni::dyn::concentration_drift   =  false;

vector<emd::uni::dyn::sph_radius_drift_spec>
                                 emd::uni::dyn::sph_radius_drift_vec;

/***************************************************************************/

std::string                      emd::uni::dyn::log_file_name;        /* NOTE must be supplied by user */
std::string                      emd::uni::dyn::log_file_name_stats;  /* NOTE must be supplied by user */
std::string                      emd::uni::dyn::log_file_name_force_on_surfs; /* NOTE must be supplied by user */

std::string                      emd::uni::dyn::integrator_desc;

/***************************************************************************/
/* thermo/baro -stat */

int                              emd::uni::dyn::LIOUVILLE_PART_ORDER  =  4;

int                              emd::uni::dyn::NHC_T_length          =  5;
double                           emd::uni::dyn::NHC_T_freq            =  0.01;

int                              emd::uni::dyn::NHC_P_length          =  5;
double                           emd::uni::dyn::NHC_P_freq            =  0.0001;

double                           emd::uni::dyn::NHC_g_t;           /* translational degree of freedom */
double                           emd::uni::dyn::NHC_g_r;           /* rotational degrees of freedom */
double                           emd::uni::dyn::NHC_g;             /* degrees of freedom (trans + rot) */

double                           emd::uni::dyn::brendsen_kappa        =  0.5;

double *                         emd::uni::dyn::NHC_T_ksi_t;       /* NHC velocity (translational) */
double *                         emd::uni::dyn::NHC_T_ksi_r;       /* NHC velocity (rotational) */
double *                         emd::uni::dyn::NHC_T_G_t;         /* NHC force (translational) */
double *                         emd::uni::dyn::NHC_T_G_r;         /* NHC force (rotational) */
double *                         emd::uni::dyn::NHC_T_s_t;         /* NHC state variable (translational) */
double *                         emd::uni::dyn::NHC_T_s_r;         /* NHC state variable (rotational) */
double *                         emd::uni::dyn::NHC_T_Q_t;         /* NHC masses (translational) */
double *                         emd::uni::dyn::NHC_T_Q_r;         /* NHC masses (rotational) */

double                           emd::uni::dyn::NHC_P_eps;         /* NHB state (volume) */
double                           emd::uni::dyn::NHC_P_ksi_eps;     /* NHB momentum (volume) */
double                           emd::uni::dyn::NHC_P_G_eps;       /* NHB force (volume) */
double                           emd::uni::dyn::NHC_P_Q_eps;       /* NHB mass (volume) */

double *                         emd::uni::dyn::NHC_P_eta;         /* NHB chain state (chain) */
double *                         emd::uni::dyn::NHC_P_ksi_eta;     /* NHB chain momenta (chain) */
double *                         emd::uni::dyn::NHC_P_G_eta;       /* NHB chain forces (chain) */
double *                         emd::uni::dyn::NHC_P_Q_eta;       /* NHB masses (chain) */

/***************************************************************************/

void emd::uni::dyn::init()
{

  dt_2                        =   dt*dt;

  /***************************************************/

  cout << "* initializing the MD simulation..." << endl;
  cout.flush();

  /***************************************************/

  /* initializing extended memories */
  for (int m=0; m<rigid_bodies_count; m++)
    if ( !rigid_bodies[m].is_ext_alloc() )
      rigid_bodies[m].malloc_ext();

  /***************************************************/

  int stat = emd::uni::eval_volume();

  if (stat==EXIT_FAILURE)
  {

    cout << "* emd::uni::init_universe:post: cannot evaluate volume, NPT simulation"
         << " is not possible." << endl;

    emd::uni::box_closed_flag = false;

  }
  else
  {

    cout << "* emd::uni::init_universe:post: the simulation box is closed, NPT simulation"
         << " is possible." << endl;
    emd::uni::box_closed_flag = true;

    concentration = rigid_bodies_count / volume;

  }

  /* preliminary energy/force/torque evaluation */
  emd::uni::fold_to_box();

  /* init cell grid */
  emd::uni::init_cell_grid();

  /* if cell hashing is applicable, decompose the rigid bodies into cells */
  if (use_cell_info)
  {
    for (unsigned int i=0; i<rigid_bodies_count; i++)
      update_rigid_body_cell_info(i);
  }

  if (use_cell_info)
    cout << "* emd::uni::init_universe_post: cell hashing will be used." << endl;
  else
    cout << "* emd::uni::init_universe_post: cell hashing is not applicable." << endl;

  /* init the neighbor list */
  init_neighbor_list(NL_VERLET);

  /* update the neighbor list */
  emd::uni::update_neighbor_lists();

  emd::uni::clear_evals();
  emd::uni::eval_all_inter(true);
  emd::uni::eval_all_extern(true);
  emd::uni::eval_all_surf(true);
  emd::uni::eval_kinetic();

  goal_energy = kinetic + potential;

  /* CM vector */
  emd::uni::eval_CM_vector();

  if (temperature_drift)
    goal_temperature = init_temperature;

  /***************************************************/

  active_num = 0;
  active_list.reserve(rigid_bodies_count);

  for (int m=0; m<rigid_bodies_count; m++)
    if ( rigid_bodies[m].status == rigid_body::EMD_RIGIDBODY_STATUS_ACTIVE )
    {
      active_num++;
      active_list.push_back(m);
    }

  /* statistics */
  stat_temp_vec.reserve(stats_interval);
  stat_press_vec.reserve(stats_interval);
  stat_conc_vec.reserve(stats_interval);
  stat_en_vec.reserve(stats_interval);
  stat_p2_vec.reserve(stats_interval);

  /***************************************************/

  switch (integrator)
  {

    case NVE_MILLER_SYMPLECTIC:
    {
      init_NVE_miller_symplectic();
      fix_T = false;
      fix_P = false;
      integrator_desc = "NO-SQUISH, NVE";
      break;
    }

    case NVT_MILLER_SYMPLECTIC_NHC:
    {
      init_NVT_miller_symplectic_NHC(reset_thermostat);
      fix_T = true;
      fix_P = false;
      integrator_desc = "NO-SQUISH, NVT, NHC Thermostat";
      break;
    }

    case NPT_MILLER_SYMPLECTIC_NHC:
    {
      init_NPT_miller_symplectic_NHC(reset_thermostat, reset_barostat);
      fix_T = true;
      fix_P = true;
      integrator_desc = "NO-SQUISH, NPT, NHC Thermo/Baro-stat";
      break;
    }

  }

  if (fix_P)
  {
    eval_volume();
    init_volume = volume;
  }

  /***************************************************/

  if (log_to_file)
  {

    log_file() << "#################" << endl
               << "## MD log file ##" << endl
               << "#################" << endl << endl;

    if (fix_T)
      log_file() << "legend   time potential energy temperature pressure order_param";

    if (fix_P)
      log_file() << " concentration hamiltonian" << endl << endl;
    else
      log_file() << " hamiltonian" << endl << endl;

  }

  if (log_to_stats_file)
  {

    stats_log_file() << "    t    "
                     << "      U      "
                     << "     ~U       "
                     << "      T      "
                     << "     ~T       "
                     << "      P      "
                     << "     ~P       "
                     << "      C      "
                     << "     ~C       "
                     << "      O      "
                     << "     ~O       " << endl
                     << "----------------------------------------------------------------------------------------------------------------------------------------------" << endl;

  }

}

/***************************************************************************/

void emd::uni::dyn::step()
{

  /* NOTE phase 0:  fixing CM linear momentum */

  if (fix_cm)
    zero_CM_momentum();

  /* NOTE phase 1:  integration + evaluation of (kinetic energy, virial) */

  switch (integrator)
  {

    case NVE_MILLER_SYMPLECTIC:
    {
      step_NVE_miller_symplectic();
//      step_brendsen_thermostat(dt);
      break;
    }

    case NVT_MILLER_SYMPLECTIC_NHC:
    {
      step_NVT_miller_symplectic_NHC();
      break;
    }

    case NPT_MILLER_SYMPLECTIC_NHC:
    {
      step_NPT_miller_symplectic_NHC();
      break;
    }

  }

  /* NOTE phase 2:  evaluation of temperature and pressure */

  eval_temperature();
  eval_pressure();

  /* NOTE phase 2.1:  evaluation of order parameter */

  eval_Q_tensor(active_list, 2, Q_tensor, Q_eigen_val, Q_eigen_vec);
  order_param = gsl_vector_max( Q_eigen_val );

  /* NOTE phase 3:  stats, log */

  /* update stats lists */
  if (iter>=N_ignore_steps)
  {

    stat_temp_vec.push_back(temperature);
    stat_conc_vec.push_back(concentration);
    stat_press_vec.push_back(pressure);
    stat_en_vec.push_back(potential + kinetic);
    stat_p2_vec.push_back(order_param);
    stats_counter++;

    if (stats_counter == stats_interval)
    {

      stats_counter = 0;

      double en_mean, en_std;
      double temp_mean, temp_std;
      double conc_mean, conc_std;
      double press_mean, press_std;
      double p2_mean, p2_std;

      eval_stats_vec_double(stat_en_vec, en_mean, en_std);
      eval_stats_vec_double(stat_temp_vec, temp_mean, temp_std);
      eval_stats_vec_double(stat_press_vec, press_mean, press_std);
      eval_stats_vec_double(stat_conc_vec, conc_mean, conc_std);
      eval_stats_vec_double(stat_p2_vec, p2_mean, p2_std);

      if (log_to_stats_file)
      {
        stats_log_file() << noshowpos << setprecision(5) << fixed
                         << t << setprecision(5) << scientific << "  "
                         << showpos << en_mean << "  " << noshowpos << en_std << "  "
                         << showpos << temp_mean << "  " << noshowpos << temp_std << "  "
                         << showpos << press_mean << "  " << noshowpos << press_std << "  "
                         << showpos << conc_mean << "  " << noshowpos << conc_std << "  "
                         << showpos << p2_mean << "  " << noshowpos << p2_std << endl;
      }

      if (log_to_stdout)
      {
        cout << showpos << setprecision(3) << scientific << endl
             << " * STATS:" << endl
             << "   - energy:         " << showpos << en_mean << " (~ " << noshowpos << en_std << ")" << endl
             << "   - temperature:    " << showpos << temp_mean << " (~ " << noshowpos << temp_std << ")" << endl
             << "   - pressure:       " << showpos << press_mean << " (~ " << noshowpos << press_std << ")" << endl
             << "   - concentration:  " << showpos << conc_mean << " (~ " << noshowpos << conc_std << ")" << endl
             << "   - order param:    " << showpos << p2_mean << " (~ " << noshowpos << p2_std << ")" << endl << endl;
      }

      stat_en_vec.clear();
      stat_temp_vec.clear();
      stat_press_vec.clear();
      stat_conc_vec.clear();
      stat_p2_vec.clear();

    }

  }

  if (log_to_file && (iter>N_ignore_steps))
  {

    log_file() << "snap   " << noshowpos
               << setprecision(7) << fixed << setw(9) << left << t << showpos << scientific << setprecision(4)
               << "  " << setw(8) << potential
               << "  " << setw(8) << (kinetic + potential)
               << "  " << setw(8) << temperature
               << "  " << setw(8) << pressure
               << "  " << setw(8) << order_param;

    if (fix_P)
      log_file() << "  " << setw(8) << concentration
                 << "  " << setw(8) << NPT_hamiltonian << endl;
    else if (fix_T)
      log_file() << "  " << setw(8) << NVT_hamiltonian << endl;

    else
      log_file() << endl;

  }

  if (log_to_stdout)
  {

    cout << noshowpos
         << " t: " << setprecision(3) << fixed << setw(fix_P?7:9) << t << showpos << scientific << setprecision(2)
         << "  U " << setw(8) << potential
         << "  E " << setw(8) << (kinetic + potential)
         << "  T " << setw(8) << temperature
         << "  P " << setw(8) << pressure
      //   << "  O " << setw(8) << order_param
         << "  C " << setw(8) << concentration;

    if (fix_P)
      cout << "  C " << setw(8) << concentration
           << "  H " << setw(8) << NPT_hamiltonian << endl;
    else if (fix_T)
      cout << "  H " << setw(8) << NVT_hamiltonian << endl;

    else
      cout << endl;

  }

  if (log_force_on_surfs && (iter>N_ignore_steps))
  {

     force_on_surfs_log_file() << noshowpos
                               << setprecision(7) << fixed << setw(9) << left << t;

    for (unsigned int s=0; s<flat_int_surfs_count; s++)
    {

      force_on_surfs_log_file() << "   flat_" << noshowpos << setw(3) << left << s;
      force_on_surfs_log_file() << showpos << scientific << setprecision(4)
                                << "  " << setw(8) << gsl_vector_get(flat_int_surfs[s].force, 0)
                                << "  " << setw(8) << gsl_vector_get(flat_int_surfs[s].force, 1)
                                << "  " << setw(8) << gsl_vector_get(flat_int_surfs[s].force, 2);
    }

    for (unsigned int s=0; s<sph_int_surfs_count; s++)
    {

      force_on_surfs_log_file() << "   sph_" << noshowpos << setw(3) << left << s;
      force_on_surfs_log_file() << showpos << scientific << setprecision(4)
                                << "  " << setw(8) << gsl_vector_get(sph_int_surfs[s].force, 0)
                                << "  " << setw(8) << gsl_vector_get(sph_int_surfs[s].force, 1)
                                << "  " << setw(8) << gsl_vector_get(sph_int_surfs[s].force, 2);
    }

    force_on_surfs_log_file() << endl;

  }

  /* NOTE phase 4  update neighbor list */

  if ( (iter-prev_iter) == nl_update_interval )
  {

    /* fold back within the periodic boundary */
    emd::uni::fold_to_box();

    /* if NPT, resetup the cell decomposition */
    if (fix_P) init_cell_grid();

    /* update cell info */
    if (use_cell_info) update_rigid_bodies_cell_info();

    /* update the neighbor lists */
    emd::uni::update_neighbor_lists();

    prev_iter = iter;

  }

  iter++;

  /* NOTE phase 5  temperature drift */

  if (temperature_drift)

    switch (temp_drift)
    {

      case TEMP_DRIFT_LOG:
      {
        adjust_log_temperature_drift();
        break;
      }

      case TEMP_DRIFT_LIN:
      {
        adjust_lin_temperature_drift();
        break;
      }

    }

  /* NOTE phase 6  concentration drift */

  if (concentration_drift)
    adjust_lin_concentration_drift();

  /* NOTE phase 7  sphere radius drift */

  if (sph_radius_drift_vec.size()>0)
    adjust_lin_sph_radius_drift();

}

/***************************************************************************/

void emd::uni::dyn::conservate_energy(const double& energy0)
{

  double kinetic_req = energy0 - potential;
  double lambda = sqrt(kinetic_req/kinetic);

  int m;
  for (int_it it = active_list.begin(); it!=active_list.end(); it++)
  {
    m = *it;

    rigid_bodies[m].P      = gsl_quaternion_mul_real(rigid_bodies[m].P, lambda);      /* conj momentum of Q */
    rigid_bodies[m].l      = gsl_quaternion_mul_real(rigid_bodies[m].l, lambda);      /* angular momentum */
    rigid_bodies[m].omega  = gsl_quaternion_mul_real(rigid_bodies[m].omega, lambda);  /* angular velocity */
    gsl_vector_scale(rigid_bodies[m].v, lambda);                                      /* CM velocity */

  }

  kinetic *= gsl_pow_2(lambda);

}

/***************************************************************************/

void emd::uni::dyn::zero_CM_momentum()
{

  gsl_vector *   v_cm  =  gsl_vector_calloc(3);
  gsl_vector *   p     =  gsl_vector_calloc(3);
  double         mass  =  0.0;

  int m;
  for (int_it it = active_list.begin(); it!=active_list.end(); it++)
  {
    m = *it;

    gsl_vector_memcpy(p, rigid_bodies[m].v);
    gsl_vector_scale(p, rigid_bodies[m].param->mass);

    gsl_vector_add(v_cm, p);
    mass += rigid_bodies[m].param->mass;
  }

  /* evaluate CM velocity */
  gsl_vector_scale(v_cm, 1.0/mass);

  for (int_it it = active_list.begin(); it!=active_list.end(); it++)
  {
    m = *it;
    gsl_vector_sub(rigid_bodies[m].v, v_cm);
  }

  gsl_vector_free(v_cm);
  gsl_vector_free(p);

}

/***************************************************************************/

void
emd::uni::dyn::adjust_volume (const double& new_vol)
{

  /* phase 1. find the bounding box of the rigid bodies */
  double min_x, min_y, min_z;
  double max_x, max_y, max_z;
  double c_x,   c_y,   c_z;

  min_x = gsl_vector_get(rigid_bodies[0].x, 0);
  min_y = gsl_vector_get(rigid_bodies[0].x, 1);
  min_z = gsl_vector_get(rigid_bodies[0].x, 2);
  max_x = min_x;
  max_y = min_y;
  max_z = min_z;

  for (unsigned int i=0; i<rigid_bodies_count; i++)
  {
    c_x = gsl_vector_get( rigid_bodies[i].x, 0 );
    c_y = gsl_vector_get( rigid_bodies[i].x, 1 );
    c_z = gsl_vector_get( rigid_bodies[i].x, 2 );

    if (c_x>max_x) max_x = c_x;
    if (c_y>max_y) max_y = c_y;
    if (c_z>max_z) max_z = c_z;
    if (c_x<min_x) min_x = c_x;
    if (c_y<min_y) min_y = c_y;
    if (c_z<min_z) min_z = c_z;
  }

  /* phase 2. find the bouding box of the universe */

  /* planes */
  vector<emd::geom::plane_3d>  planes;
  emd::geom::plane_3d          new_plane;
  int                          stat  =  EXIT_SUCCESS;
  vector<geom::point_3d>       points;
  int                          n_faces = 0;
  vector<int>                  tastes;

  if (x_periodic)
  {
    new_plane.nx = 1;  new_plane.ny = 0;  new_plane.nz = 0;
    new_plane.d  = x_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = x_upper_bound;
    planes.push_back(new_plane);
  }

  if (y_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 1;  new_plane.nz = 0;
    new_plane.d  = y_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = y_upper_bound;
    planes.push_back(new_plane);
  }

  if (z_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 0;  new_plane.nz = 1;
    new_plane.d  = z_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = z_upper_bound;
    planes.push_back(new_plane);
  }

  /* flat int surfaces */
  for (int i=0; i<flat_int_surfs_count; i++)
  {
    new_plane.nx = -gsl_vector_get( flat_int_surfs[i].n, 0 );
    new_plane.ny = -gsl_vector_get( flat_int_surfs[i].n, 1 );
    new_plane.nz = -gsl_vector_get( flat_int_surfs[i].n, 2 );
    gsl_blas_ddot( flat_int_surfs[i].n, flat_int_surfs[i].r0, &(new_plane.d) );
    new_plane.d *= -1;
    planes.push_back(new_plane);
  }

  /* interior point */
  geom::point_3d interior_point_3d = { gsl_vector_get(interior_point, 0),
                                       gsl_vector_get(interior_point, 1),
                                       gsl_vector_get(interior_point, 2) };

  stat = emd::geom::find_vertex_points( planes, interior_point_3d, points );

  if (stat==EXIT_FAILURE) return;

  double box_min_x, box_min_y, box_min_z;
  double box_max_x, box_max_y, box_max_z;

  box_min_x = points[0].x;
  box_min_y = points[0].y;
  box_min_z = points[0].z;
  box_max_x = points[0].x;
  box_max_y = points[0].y;
  box_max_z = points[0].z;

  for (int j=0; j<points.size(); j++)
  {
    c_x = points[j].x;
    c_y = points[j].y;
    c_z = points[j].z;

    if (c_x > box_max_x) box_max_x = c_x;
    if (c_y > box_max_y) box_max_y = c_y;
    if (c_z > box_max_z) box_max_z = c_z;
    if (c_x < box_min_x) box_min_x = c_x;
    if (c_y < box_min_y) box_min_y = c_y;
    if (c_z < box_min_z) box_min_z = c_z;
  }

  double dx     =  box_max_x - box_min_x - max_x + min_x;
  double dy     =  box_max_y - box_min_y - max_y + min_y;
  double dz     =  box_max_z - box_min_z - max_z + min_z;
  double d_sum  =  dx + dy + dz;

  /* eval contraction factors */
  double const_vol = 0;
  double lambda, lambda_x, lambda_y, lambda_z;

  for (int i=0; i<sph_int_surfs_count; i++)
    const_vol += ( 4 * M_PI * gsl_pow_3 (gsl_blas_dnrm2( sph_int_surfs[i].r0 ) ) / 3 );

  lambda   = pow( (new_vol + const_vol) / (volume + const_vol), 0.333333333333 );
//  lambda_x = lambda * exp ( 0.5 * (lambda-1) * (2*dx - dy - dz) / (3*d_sum) );
//  lambda_y = lambda * exp ( 0.5 * (lambda-1) * (2*dy - dz - dx) / (3*d_sum) );
//  lambda_z = lambda * exp ( 0.5 * (lambda-1) * (2*dz - dx - dy) / (3*d_sum) );

  lambda_x = lambda;
  lambda_y = lambda;
  lambda_z = lambda;

  /* contract the boundaries and flat surfaces */
  if (x_periodic)
  {
    x_lower_bound = lambda_x*x_lower_bound + (1-lambda_x) * gsl_vector_get(r_cm, 0);
    x_upper_bound = lambda_x*x_upper_bound + (1-lambda_x) * gsl_vector_get(r_cm, 0);
    x_width       = x_upper_bound - x_lower_bound;
    x_center      = 0.5 * (x_upper_bound + x_lower_bound);
  }

  if (y_periodic)
  {
    y_lower_bound = lambda_y*y_lower_bound + (1-lambda_y) * gsl_vector_get(r_cm, 1);
    y_upper_bound = lambda_y*y_upper_bound + (1-lambda_y) * gsl_vector_get(r_cm, 1);
    y_width       = y_upper_bound - y_lower_bound;
    y_center      = 0.5 * (y_upper_bound + y_lower_bound);
  }

  if (z_periodic)
  {
    z_lower_bound = lambda_z*z_lower_bound + (1-lambda_z) * gsl_vector_get(r_cm, 2);
    z_upper_bound = lambda_z*z_upper_bound + (1-lambda_z) * gsl_vector_get(r_cm, 2);
    z_width       = z_upper_bound - z_lower_bound;
    z_center      = 0.5 * (z_upper_bound + z_lower_bound);
  }

  double dist, dot;
  for (long int s=0; s<flat_int_surfs_count; s++)
  {
    gsl_vector_memcpy( tmp_v, flat_int_surfs[s].r0 );
    gsl_vector_sub( tmp_v, r_cm );
    gsl_vector_set( tmp_v, 0, lambda_x * gsl_vector_get(tmp_v, 0) );
    gsl_vector_set( tmp_v, 1, lambda_y * gsl_vector_get(tmp_v, 1) );
    gsl_vector_set( tmp_v, 2, lambda_z * gsl_vector_get(tmp_v, 2) );
    gsl_vector_add( tmp_v, r_cm );

    gsl_vector_memcpy( flat_int_surfs[s].r0, tmp_v );

  }

  /* additive constant */
  gsl_vector_set ( tmp_v, 0, (1 - lambda_x) * gsl_vector_get(r_cm, 0) );
  gsl_vector_set ( tmp_v, 1, (1 - lambda_y) * gsl_vector_get(r_cm, 1) );
  gsl_vector_set ( tmp_v, 2, (1 - lambda_z) * gsl_vector_get(r_cm, 2) );

  /* interior point */
  gsl_vector_set( interior_point, 0, lambda_x * gsl_vector_get(interior_point, 0) );
  gsl_vector_set( interior_point, 1, lambda_y * gsl_vector_get(interior_point, 1) );
  gsl_vector_set( interior_point, 2, lambda_z * gsl_vector_get(interior_point, 2) );
  gsl_vector_add( interior_point, tmp_v );

  /* spherical surfaces */
  for (long int s=0; s<sph_int_surfs_count; s++)
  {

    gsl_vector_set( sph_int_surfs[s].r0, 0, lambda_x * gsl_vector_get(sph_int_surfs[s].r0, 0) );
    gsl_vector_set( sph_int_surfs[s].r0, 1, lambda_y * gsl_vector_get(sph_int_surfs[s].r0, 1) );
    gsl_vector_set( sph_int_surfs[s].r0, 2, lambda_z * gsl_vector_get(sph_int_surfs[s].r0, 2) );

    gsl_vector_add( sph_int_surfs[s].r0, tmp_v );
  }

  /* update the volume variable */
  volume = new_vol;

}

/***************************************************************************/

void
emd::uni::dyn::adjust_log_temperature_drift()
{

  goal_temperature = init_temperature +
    (final_temperature - init_temperature) * log((double)iter) / log((double)N_steps);

}

/***************************************************************************/

void
emd::uni::dyn::adjust_lin_temperature_drift()
{

  goal_temperature = init_temperature +
    (final_temperature - init_temperature) * ((double)iter) / ((double)N_steps);

}

/***************************************************************************/

void
emd::uni::dyn::adjust_lin_concentration_drift()
{

  static double init_concentration = concentration;

  double kappa = ( pow(init_concentration / final_concentration, 1.0 / 3) - 1.0 ) / (double)N_steps;
  concentration  =  init_concentration / gsl_pow_3( 1.0 + iter * kappa );

  double new_vol = rigid_bodies_count / concentration;

  double const_vol = 0;
  for (int i=0; i<sph_int_surfs_count; i++)
    const_vol += ( 4 * M_PI * gsl_pow_3 (sph_int_surfs[i].rad) / 3 );

  double scale_factor = pow( (new_vol + const_vol) / (volume + const_vol), 1.0 / 3 );

  volume = new_vol;

  for (int m=0; m<rigid_bodies_count; m++)
    gsl_vector_scale(rigid_bodies[m].x, scale_factor);

/*  for (int m=0; m<flat_int_surfs_count; m++)
    gsl_vector_scale(flat_int_surfs[m].r0, scale_factor);

  for (int m=0; m<sph_int_surfs_count; m++)
    gsl_vector_scale(sph_int_surfs[m].r0, scale_factor);*/

  gsl_vector_scale(interior_point, scale_factor);

  x_lower_bound *= scale_factor;
  x_upper_bound *= scale_factor;
  x_width       *= scale_factor;
  x_center      *= scale_factor;

  y_lower_bound *= scale_factor;
  y_upper_bound *= scale_factor;
  y_width       *= scale_factor;
  y_center      *= scale_factor;

  z_lower_bound *= scale_factor;
  z_upper_bound *= scale_factor;
  z_width       *= scale_factor;
  z_center      *= scale_factor;

  init_cell_grid();

}

/***************************************************************************/

void
emd::uni::dyn::adjust_lin_sph_radius_drift()
{

  for (int i=0; i<sph_radius_drift_vec.size(); i++)
  {
    int idx          = sph_radius_drift_vec[i].idx;
    double init_rad  = sph_radius_drift_vec[i].init_rad;
    double final_rad = sph_radius_drift_vec[i].final_rad;

    sph_int_surfs[idx].rad = iter*(final_rad-init_rad)/(double)N_steps + init_rad;

  }

  eval_volume();

}

/***************************************************************************/

void
emd::uni::dyn::step_brendsen_thermostat(const double & dt)
{

  /* degress of freedom */
  int N_freedom_t = 3*active_num;
  int N_freedom_r = 3*active_num;
  if (fix_cm) N_freedom_t -= 3;

  double temp_t = 2.0 * kinetic_t / (BOLZ_CONST * N_freedom_t);
  double temp_r = 2.0 * kinetic_r / (BOLZ_CONST * N_freedom_r);

  double lambda_t = sqrt(1.0 + (goal_temperature/temp_t - 1)*dt/brendsen_kappa);
  double lambda_r = sqrt(1.0 + (goal_temperature/temp_r - 1)*dt/brendsen_kappa);

  rigid_body * rb;

  /* loop through the active rigid bodies */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* velocity */
    gsl_vector_scale(rb->v, lambda_t);

    /* conjugate momenum of Q */
    rb->P.dat[0] *= lambda_r;
    rb->P.dat[1] *= lambda_r;
    rb->P.dat[2] *= lambda_r;
    rb->P.dat[3] *= lambda_r;

    /* angular momentum */
    rb->l.dat[0] *= lambda_r;
    rb->l.dat[1] *= lambda_r;
    rb->l.dat[2] *= lambda_r;
    rb->l.dat[3] *= lambda_r;

    /* angular velocity */
    rb->omega.dat[0] *= lambda_r;
    rb->omega.dat[1] *= lambda_r;
    rb->omega.dat[2] *= lambda_r;
    rb->omega.dat[3] *= lambda_r;

  }

  kinetic_t *= gsl_pow_2(lambda_t);
  kinetic_r *= gsl_pow_2(lambda_r);

  kinetic = kinetic_t + kinetic_r;

}

/***************************************************************************/

void
emd::uni::dyn::eval_temperature()
{

  /* degress of freedom */
  int N_freedom = 6*active_num;
  if (fix_cm) N_freedom -= 3;

  temperature = 2.0 * kinetic / (BOLZ_CONST * N_freedom);

}

/***************************************************************************/

void
emd::uni::dyn::eval_pressure()
{
  pressure = (2 * kinetic_t + vir_x + vir_y + vir_z) / (3 * volume);

}

/***************************************************************************/

void
emd::uni::dyn::eval_stats_vec_double(std::vector<double> & vec,
  double & mean, double & std)
{

  mean = 0;
  std  = 0;

  /* mean */
  for (vector<double>::iterator it = vec.begin(); it != vec.end(); it++)
    mean += *it;

  mean /= vec.size();

  /* std */
  for (vector<double>::iterator it = vec.begin(); it != vec.end(); it++)
    std += gsl_pow_2( *it - mean );

  std /= vec.size();

  std = sqrt(std);

}

/***************************************************************************/

ofstream&
emd::uni::dyn::log_file()
{
  static ofstream log( log_file_name.c_str() );
  return log;
}

/***************************************************************************/

ofstream&
emd::uni::dyn::stats_log_file()
{
  static ofstream log( log_file_name_stats.c_str() );
  return log;
}

/***************************************************************************/

ofstream&
emd::uni::dyn::force_on_surfs_log_file()
{
  static ofstream log( log_file_name_force_on_surfs.c_str() );
  return log;
}

/***************************************************************************/

/**** NOTE **** NOTE **** NOTE **** NOTE **** NOTE **** NOTE **** NOTE *****
 *                                                                         *
 * time-reversible and symplectic integrator, using Miller's rigid-body    *
 * Hamiltonian for NVE, NVT and NPT ensembles [1].                         *
 * The thermostats and barostats are both based on Nose-Hoover Chain       *
 * method.                                                                 *
 *                                                                         *
 ***************************************************************************/

/* helper functions */

double
emd::uni::dyn::sinch_10 (const double & x)
{

  double x_2 = x*x;
  double x_4 = x_2*x_2;
  double x_8 = x_4*x_4;

  return ( 1 + x_2/6 + x_4/120 + x_2*x_4/5040 + x_8/362880 + x_8*x_2/39916800 );

}

/***************************************************************************/

/* NOTE: R(t) must be up to date! */
gsl_quaternion
emd::uni::dyn::get_principal_frame_torque (const emd::rigid_body & rb)
{

  gsl_quaternion Q_torque;

  /* N_{principal} = R^T N_{lab} */
  gsl_blas_dgemv ( CblasTrans, 1, rb.R, rb.torque, 0, tmp_v );

  Q_torque.dat[0] = 0;
  Q_torque.dat[1] = gsl_vector_get(tmp_v, 0);
  Q_torque.dat[2] = gsl_vector_get(tmp_v, 1);
  Q_torque.dat[3] = gsl_vector_get(tmp_v, 2);

  return Q_torque;

}

/***************************************************************************/

gsl_quaternion
emd::uni::dyn::apply_liouville_op_step_P ( const gsl_quaternion & Q,
  const gsl_quaternion & P, const gsl_quaternion & N, const double & delta_t )
{

  gsl_quaternion tmp_quat = gsl_quaternion_mul( Q, N );
  gsl_quaternion res;

  res.dat[0] = P.dat[0] + 2 * delta_t * tmp_quat.dat[0];
  res.dat[1] = P.dat[1] + 2 * delta_t * tmp_quat.dat[1];
  res.dat[2] = P.dat[2] + 2 * delta_t * tmp_quat.dat[2];
  res.dat[3] = P.dat[3] + 2 * delta_t * tmp_quat.dat[3];

  return res;

}

/***************************************************************************/

void
emd::uni::dyn::apply_no_squish_rot_P_Q_1 (emd::rigid_body & rb, const double & delta_t)
{

  double ksi_1 = gsl_quaternion_dot_prod( rb.P, gsl_quaternion_d1(rb.Q) ) / (4 * rb.param->I1);

  double cos_term = cos (ksi_1 * delta_t);
  double sin_term = sin (ksi_1 * delta_t);

  gsl_quaternion Q_tmp = gsl_quaternion_d1(rb.Q);
  gsl_quaternion P_tmp = gsl_quaternion_d1(rb.P);

  rb.Q.dat[0] = cos_term * rb.Q.dat[0] + sin_term * Q_tmp.dat[0];
  rb.Q.dat[1] = cos_term * rb.Q.dat[1] + sin_term * Q_tmp.dat[1];
  rb.Q.dat[2] = cos_term * rb.Q.dat[2] + sin_term * Q_tmp.dat[2];
  rb.Q.dat[3] = cos_term * rb.Q.dat[3] + sin_term * Q_tmp.dat[3];

  rb.P.dat[0] = cos_term * rb.P.dat[0] + sin_term * P_tmp.dat[0];
  rb.P.dat[1] = cos_term * rb.P.dat[1] + sin_term * P_tmp.dat[1];
  rb.P.dat[2] = cos_term * rb.P.dat[2] + sin_term * P_tmp.dat[2];
  rb.P.dat[3] = cos_term * rb.P.dat[3] + sin_term * P_tmp.dat[3];

}

/***************************************************************************/

void
emd::uni::dyn::apply_no_squish_rot_P_Q_2 (emd::rigid_body & rb, const double & delta_t)
{

  double ksi_2 = gsl_quaternion_dot_prod( rb.P, gsl_quaternion_d2(rb.Q) ) / (4 * rb.param->I2);

  double cos_term = cos (ksi_2 * delta_t);
  double sin_term = sin (ksi_2 * delta_t);

  gsl_quaternion Q_tmp = gsl_quaternion_d2(rb.Q);
  gsl_quaternion P_tmp = gsl_quaternion_d2(rb.P);

  rb.Q.dat[0] = cos_term * rb.Q.dat[0] + sin_term * Q_tmp.dat[0];
  rb.Q.dat[1] = cos_term * rb.Q.dat[1] + sin_term * Q_tmp.dat[1];
  rb.Q.dat[2] = cos_term * rb.Q.dat[2] + sin_term * Q_tmp.dat[2];
  rb.Q.dat[3] = cos_term * rb.Q.dat[3] + sin_term * Q_tmp.dat[3];

  rb.P.dat[0] = cos_term * rb.P.dat[0] + sin_term * P_tmp.dat[0];
  rb.P.dat[1] = cos_term * rb.P.dat[1] + sin_term * P_tmp.dat[1];
  rb.P.dat[2] = cos_term * rb.P.dat[2] + sin_term * P_tmp.dat[2];
  rb.P.dat[3] = cos_term * rb.P.dat[3] + sin_term * P_tmp.dat[3];

}

/***************************************************************************/

void
emd::uni::dyn::apply_no_squish_rot_P_Q_3 (emd::rigid_body & rb, const double & delta_t)
{

  double ksi_3 = gsl_quaternion_dot_prod( rb.P, gsl_quaternion_d3(rb.Q) ) / (4 * rb.param->I3);

  double cos_term = cos (ksi_3 * delta_t);
  double sin_term = sin (ksi_3 * delta_t);

  gsl_quaternion Q_tmp = gsl_quaternion_d3(rb.Q);
  gsl_quaternion P_tmp = gsl_quaternion_d3(rb.P);

  rb.Q.dat[0] = cos_term * rb.Q.dat[0] + sin_term * Q_tmp.dat[0];
  rb.Q.dat[1] = cos_term * rb.Q.dat[1] + sin_term * Q_tmp.dat[1];
  rb.Q.dat[2] = cos_term * rb.Q.dat[2] + sin_term * Q_tmp.dat[2];
  rb.Q.dat[3] = cos_term * rb.Q.dat[3] + sin_term * Q_tmp.dat[3];

  rb.P.dat[0] = cos_term * rb.P.dat[0] + sin_term * P_tmp.dat[0];
  rb.P.dat[1] = cos_term * rb.P.dat[1] + sin_term * P_tmp.dat[1];
  rb.P.dat[2] = cos_term * rb.P.dat[2] + sin_term * P_tmp.dat[2];
  rb.P.dat[3] = cos_term * rb.P.dat[3] + sin_term * P_tmp.dat[3];

}

/***************************************************************************/

/* stepper */
void
emd::uni::dyn::init_NVE_miller_symplectic()
{
  /* NOTE nothing here */
}

/***************************************************************************/

/* stepper */
void
emd::uni::dyn::step_NVE_miller_symplectic()
{

  /* temporary workspace */

  rigid_body *                    rb;

  /* loop through the active rigid bodies */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* v(t + dt/2) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* x(t + dt) */
    gsl_vector_memcpy( tmp_v, rb->v );
    gsl_vector_scale( tmp_v, dt );
    gsl_vector_add( rb->x, tmp_v );

    /* external torque P(t + dt/2) [1] */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* torque free rotation Q(t + dt), P(t + dt/2) [2] */
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_1 ( *rb, dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );

    /* normalize Q(t + dt) */
    gsl_quaternion_normalize( &(rb->Q) );

    /* R(t + dt) */
    gsl_quaternion_unit_quat_to_mat( &(rb->Q), rb->R );

  }

  /* F(t + dt), N(t + dt) */
  emd::uni::clear_evals();
  emd::uni::eval_all_inter(true);
  emd::uni::eval_all_extern(true);
  emd::uni::eval_all_surf(true);

  /* loop through the active rigid bodies */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* v(t + dt) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* P(t + dt) */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* l(t + dt) = P(t + dt) / ( 2 * Q(t + dt) ) */
    rb->l = gsl_quaternion_mul( gsl_quaternion_inverse(rb->Q), rb->P );
    rb->l.dat[0] *= 0.5;
    rb->l.dat[1] *= 0.5;
    rb->l.dat[2] *= 0.5;
    rb->l.dat[3] *= 0.5;

    /* omega(t + dt) = I_body^{-1} l(t + dt) */
    rb->omega.dat[0] = 0;
    rb->omega.dat[1] = rb->l.dat[1] / rb->param->I1;
    rb->omega.dat[2] = rb->l.dat[2] / rb->param->I2;
    rb->omega.dat[3] = rb->l.dat[3] / rb->param->I3;

  }

  /* eval. the kinetic energy */
  eval_kinetic();

  t = t + dt;

}

/***************************************************************************/

/* init */
void
emd::uni::dyn::init_NVT_miller_symplectic_NHC(const bool & reset_thermostat)
{

  double        kT = BOLZ_CONST * goal_temperature;

  /* NHC workspace init */
  NHC_T_ksi_t   = new double[NHC_T_length];
  NHC_T_ksi_r   = new double[NHC_T_length];
  NHC_T_G_t     = new double[NHC_T_length];
  NHC_T_G_r     = new double[NHC_T_length];
  NHC_T_s_t     = new double[NHC_T_length];
  NHC_T_s_r     = new double[NHC_T_length];
  NHC_T_Q_t     = new double[NHC_T_length];
  NHC_T_Q_r     = new double[NHC_T_length];

  NHC_g_t       = 3*active_num;
  NHC_g_r       = 3*active_num;

  if (fix_cm)
    NHC_g_t -= 3;

  if (reset_thermostat)
  {

    for (int i=0; i<NHC_T_length; i++)
    {
      NHC_T_ksi_t[i] = 0;
      NHC_T_ksi_r[i] = 0;
      NHC_T_G_t[i]   = 0;
      NHC_T_G_r[i]   = 0;
      NHC_T_s_t[i]   = 0;
      NHC_T_s_r[i]   = 0;
    }

    NHC_T_Q_t[0] = NHC_g_t * BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
    NHC_T_Q_r[0] = NHC_g_r * BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );

    for (int i=1; i<NHC_T_length; i++)
    {
      NHC_T_Q_t[i] = BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
      NHC_T_Q_r[i] = BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
    }

    for (int i=1; i<NHC_T_length; i++)
    {
      NHC_T_ksi_t[i] = 0;
      NHC_T_ksi_r[i] = 0;
    }

  }

  /* eval thermostat forces */
  eval_NHC_thermostat_forces();

}

/***************************************************************************/

/* evaluate NHC thermostat forces
 * NOTE: it is assumed that the angular momenta and velocities have been updated */
void
emd::uni::dyn::eval_NHC_thermostat_forces()
{

  rigid_body *  rb;
  double        kT = BOLZ_CONST * goal_temperature;

  /* first element of the chain */

  NHC_T_G_t[0] = 2 * kinetic_t - NHC_g_t * kT;
  NHC_T_G_r[0] = 2 * kinetic_r - NHC_g_r * kT;

  /* the rest of the chain */
  for (int k=1; k<NHC_T_length; k++)
  {
    NHC_T_G_t[k] = NHC_T_Q_t[k-1] * gsl_pow_2(NHC_T_ksi_t[k-1]) - kT;
    NHC_T_G_r[k] = NHC_T_Q_r[k-1] * gsl_pow_2(NHC_T_ksi_r[k-1]) - kT;
  }

}

/***************************************************************************/

/* step thermostat variables */
void
emd::uni::dyn::step_NHC_thermostat_new (const double & delta_t)
{

  double       dt_s, exp_term, sinch_term;
  double       kT = BOLZ_CONST * goal_temperature;
  rigid_body * rb;

  /* 3rd-order Yoshida-Suzuki */ 
  for (int i=0; i<3; i++)
  {

    /* set the stepping time */
    dt_s = YS_3[i] * delta_t;

    /* eval thermostat forces */
    eval_NHC_thermostat_forces();

    /* update the thermostat velocities */

    NHC_T_ksi_t[NHC_T_length - 1] += 0.5 * dt_s * NHC_T_G_t[NHC_T_length - 1] / NHC_T_Q_t[NHC_T_length - 1];
    NHC_T_ksi_r[NHC_T_length - 1] += 0.5 * dt_s * NHC_T_G_r[NHC_T_length - 1] / NHC_T_Q_r[NHC_T_length - 1];

    for (int k=1; k<NHC_T_length; k++)
    {
      exp_term   = exp( - 0.25 * dt_s * NHC_T_ksi_t[NHC_T_length - k] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_T_ksi_t[NHC_T_length - k] );

      NHC_T_ksi_t[NHC_T_length - k - 1] = exp_term * exp_term * NHC_T_ksi_t[NHC_T_length - k - 1]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_T_G_t[NHC_T_length - k - 1] / NHC_T_Q_t[NHC_T_length - k - 1];

      exp_term = exp( - 0.25 * dt_s * NHC_T_ksi_r[NHC_T_length - k] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_T_ksi_r[NHC_T_length - k] );

      NHC_T_ksi_r[NHC_T_length - k - 1] = exp_term * exp_term * NHC_T_ksi_r[NHC_T_length - k - 1]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_T_G_r[NHC_T_length - k - 1] / NHC_T_Q_r[NHC_T_length - k - 1];
    }

    double exp_term_t = exp( - dt_s * NHC_T_ksi_t[0] );
    double exp_term_r = exp( - dt_s * NHC_T_ksi_r[0] );

    /* scale the velocities */

    /* loop through the active rigid bodies */
    for (int_it it = active_list.begin(); it != active_list.end(); it++)
    {

      rb = &(rigid_bodies[*it]);

      /* velocity */
      gsl_vector_scale(rb->v, exp_term_t);

      /* conjugate momenum of Q */
      rb->P.dat[0] *= exp_term_r;
      rb->P.dat[1] *= exp_term_r;
      rb->P.dat[2] *= exp_term_r;
      rb->P.dat[3] *= exp_term_r;

      /* angular momentum */
      rb->l.dat[0] *= exp_term_r;
      rb->l.dat[1] *= exp_term_r;
      rb->l.dat[2] *= exp_term_r;
      rb->l.dat[3] *= exp_term_r;

      /* angular velocity */
      rb->omega.dat[0] *= exp_term_r;
      rb->omega.dat[1] *= exp_term_r;
      rb->omega.dat[2] *= exp_term_r;
      rb->omega.dat[3] *= exp_term_r;

    }

    /* update the thermostat positions */
    for (int k=0; k<NHC_T_length; k++)
    {
      NHC_T_s_t[k] += dt_s * NHC_T_ksi_t[k];
      NHC_T_s_r[k] += dt_s * NHC_T_ksi_r[k];
    }

    /* update thermostat forces */
    kinetic_t *= gsl_pow_2(exp_term_t);
    kinetic_r *= gsl_pow_2(exp_term_r);
    kinetic = kinetic_t + kinetic_r;

    NHC_T_G_t[0] = 2 * kinetic_t - NHC_g_t * kT;
    NHC_T_G_r[0] = 2 * kinetic_r - NHC_g_r * kT;

    /* update thermostat velocities */
    for (int k=0; k<NHC_T_length-1; k++)
    {

      /* translational */
      exp_term   = exp( - 0.25 * dt_s * NHC_T_ksi_t[k+1] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_T_ksi_t[k+1] );

      NHC_T_ksi_t[k] = exp_term * exp_term * NHC_T_ksi_t[k]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_T_G_t[k] / NHC_T_Q_t[k];

      NHC_T_G_t[k+1] = (NHC_T_Q_t[k] * gsl_pow_2(NHC_T_ksi_t[k]) - kT);

      /* rotational */
      exp_term = exp( - 0.25 * dt_s * NHC_T_ksi_r[k+1] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_T_ksi_r[k+1] );

      NHC_T_ksi_r[k] = exp_term * exp_term * NHC_T_ksi_r[k]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_T_G_r[k] / NHC_T_Q_r[k];

      NHC_T_G_r[k+1] = (NHC_T_Q_r[k] * gsl_pow_2(NHC_T_ksi_r[k]) - kT);

    }

    /* translational */
    NHC_T_ksi_t[NHC_T_length - 1] +=
      0.5 * dt_s * NHC_T_G_t[NHC_T_length - 1] / NHC_T_Q_t[NHC_T_length - 1];

    /* rotational */
    NHC_T_ksi_r[NHC_T_length - 1] +=
      0.5 * dt_s * NHC_T_G_r[NHC_T_length - 1] / NHC_T_Q_r[NHC_T_length - 1];

  }

}

/***************************************************************************/

/* stepper */
void
emd::uni::dyn::step_NVT_miller_symplectic_NHC()
{

  rigid_body *                    rb;

  /* phase 1. (NHC) */
  for (int i=0; i<THERMO_PART_ORDER; i++)
    step_NHC_thermostat_new ( 0.5 * dt / THERMO_PART_ORDER );

  /* phase 2. (particles) */

  /* loop through the active rigid bodies and do a half step */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* v(t + dt/2) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* x(t + dt) */
    gsl_vector_memcpy( tmp_v, rb->v );
    gsl_vector_scale( tmp_v, dt );
    gsl_vector_add( rb->x, tmp_v );

    /* external torque P(t + dt/2) [1] */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* torque free rotation Q(t + dt), P(t + dt/2) [2] */
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_1 ( *rb, dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );

    /* normalize Q(t + dt) */
    gsl_quaternion_normalize( &(rb->Q) );

    /* R(t + dt) */
    gsl_quaternion_unit_quat_to_mat( &(rb->Q), rb->R );

  }

  /* F(t + dt), N(t + dt) */
  emd::uni::clear_evals();
  emd::uni::eval_all_inter(true);
  emd::uni::eval_all_extern(true);
  emd::uni::eval_all_surf(true);

  /* loop through the active rigid bodies */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* v(t + dt) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* P(t + dt) */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* l(t + dt) = P(t + dt) / ( 2 * Q(t + dt) ) */
    rb->l = gsl_quaternion_mul( gsl_quaternion_inverse(rb->Q), rb->P );
    rb->l.dat[0] *= 0.5;
    rb->l.dat[1] *= 0.5;
    rb->l.dat[2] *= 0.5;
    rb->l.dat[3] *= 0.5;

    /* omega(t + dt) = I_body^{-1} l(t + dt) */
    rb->omega.dat[0] = 0;
    rb->omega.dat[1] = rb->l.dat[1] / rb->param->I1;
    rb->omega.dat[2] = rb->l.dat[2] / rb->param->I2;
    rb->omega.dat[3] = rb->l.dat[3] / rb->param->I3;

  }

  /* evaluate the kinetic energy */
  eval_kinetic();

  /* phase 3. (NHC) */
  for (int i=0; i<THERMO_PART_ORDER; i++)
    step_NHC_thermostat_new ( 0.5 * dt / THERMO_PART_ORDER );

  /* eval NVT hamiltonian */
  eval_NVT_hamiltonian();

  t = t + dt;

}

/***************************************************************************/

void
emd::uni::dyn::eval_NVT_hamiltonian()
{

  double kT = BOLZ_CONST * goal_temperature;

  /* NVE part */
  NVT_hamiltonian = potential + kinetic;

  /* T part */
  NVT_hamiltonian += (NHC_g_t * NHC_T_s_t[0] + NHC_g_r * NHC_T_s_r[0]) * kT;

  for (int i=1; i<NHC_T_length; i++)
    NVT_hamiltonian += (NHC_T_s_t[i] + NHC_T_s_r[i]) * kT;

  for (int i=0; i<NHC_T_length; i++)
  {
    NVT_hamiltonian += 0.5 * NHC_T_Q_t[i] * gsl_pow_2( NHC_T_ksi_t[i] );
    NVT_hamiltonian += 0.5 * NHC_T_Q_r[i] * gsl_pow_2( NHC_T_ksi_r[i] );
  }

}

/***************************************************************************/

/* init */
void
emd::uni::dyn::init_NPT_miller_symplectic_NHC(const bool & reset_thermostat,
  const bool & reset_barostat)
{

  double        kT = BOLZ_CONST * goal_temperature;

  /* NHC workspace init (T) */
  NHC_T_ksi_t    = new double[NHC_T_length];
  NHC_T_ksi_r    = new double[NHC_T_length];
  NHC_T_G_t      = new double[NHC_T_length];
  NHC_T_G_r      = new double[NHC_T_length];
  NHC_T_s_t      = new double[NHC_T_length];
  NHC_T_s_r      = new double[NHC_T_length];
  NHC_T_Q_t      = new double[NHC_T_length];
  NHC_T_Q_r      = new double[NHC_T_length];

  /* NHC workspace init (P) */
  NHC_P_eta      = new double[NHC_P_length];
  NHC_P_ksi_eta  = new double[NHC_P_length];
  NHC_P_G_eta    = new double[NHC_P_length];
  NHC_P_Q_eta    = new double[NHC_P_length];

  NHC_g_t       = 3*active_num;
  NHC_g_r       = 3*active_num;

  if (fix_cm)
    NHC_g_t -= 3;

  NHC_g         = NHC_g_t + NHC_g_r;

  if (reset_thermostat)
  {

    for (int i=0; i<NHC_T_length; i++)
    {
      NHC_T_ksi_t[i] = 0;
      NHC_T_ksi_r[i] = 0;
      NHC_T_G_t[i]   = 0;
      NHC_T_G_r[i]   = 0;
      NHC_T_s_t[i]   = 0;
      NHC_T_s_r[i]   = 0;
    }

    NHC_T_Q_t[0] = NHC_g_t * BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
    NHC_T_Q_r[0] = NHC_g_r * BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );

    for (int i=1; i<NHC_T_length; i++)
    {
      NHC_T_Q_t[i] = BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
      NHC_T_Q_r[i] = BOLZ_CONST * goal_temperature / ( NHC_T_freq * NHC_T_freq );
    }

    for (int i=1; i<NHC_T_length; i++)
    {
      NHC_T_ksi_t[i] = sqrt( kT / NHC_T_Q_t[i-1] );
      NHC_T_ksi_r[i] = sqrt( kT / NHC_T_Q_r[i-1] );
    }

  }

  if (reset_barostat)
  {

    /* volume */
    NHC_P_eps     = 0;
    NHC_P_Q_eps   = ( NHC_g + NHC_d ) * BOLZ_CONST * goal_temperature / ( NHC_P_freq * NHC_P_freq );
    NHC_P_ksi_eps = 0;

    /* chain */
    for (int i=0; i<NHC_T_length; i++)
    {
      NHC_P_eta[i]     = 0;
      NHC_P_ksi_eta[i] = 0;
      NHC_P_G_eta[i]   = 0;
    }

    /* chain masses */
    NHC_P_Q_eta[0] = gsl_pow_2(NHC_d) * BOLZ_CONST * goal_temperature / ( NHC_P_freq * NHC_P_freq );

    for (int i=1; i<NHC_P_length; i++)
      NHC_P_Q_eta[i] = BOLZ_CONST * goal_temperature / ( NHC_P_freq * NHC_P_freq );

    for (int i=0; i<NHC_P_length; i++)
      NHC_P_ksi_eta[i] = 0;

  }

  /* eval thermostat forces */
  eval_NHC_thermostat_forces();

  /* eval thermostat forces */
  eval_NHC_barostat_forces();

}

/***************************************************************************/

void
emd::uni::dyn::eval_NHC_barostat_forces()
{

  rigid_body *  rb;
  double        kT = BOLZ_CONST * goal_temperature;

  /* G_eps */
  NHC_P_G_eps = (2.0 * kinetic_t) + (vir_x + vir_y + vir_z)
                - (NHC_d * init_volume * exp( NHC_d * NHC_P_eps ) * goal_pressure)
                + NHC_d * (2.0 * kinetic) / NHC_g;

  NHC_P_G_eta[0] = NHC_P_Q_eps * gsl_pow_2( NHC_P_ksi_eps ) - kT;

  for (int k=1; k<NHC_P_length; k++)
    NHC_P_G_eta[k] = NHC_P_Q_eta[k-1] * gsl_pow_2( NHC_P_ksi_eta[k-1] ) - kT;

}

/***************************************************************************/
void
emd::uni::dyn::step_NHC_barostat_new (const double & delta_t)
{

  double       dt_s, exp_term, sinch_term;
  double       kT = BOLZ_CONST * goal_temperature;
  rigid_body * rb;

  /* 3rd-order Yoshida-Suzuki */ 
  for (int i=0; i<3; i++)
  {

    /* set the stepping time */
    dt_s = YS_3[i] * delta_t;

    /* eval barostat forces */
    eval_NHC_barostat_forces();

    /* update barostat velocities */
    exp_term   = exp( -0.25 * dt_s * NHC_P_ksi_eta[0] );
    sinch_term = sinch_10( 0.25 * dt_s * NHC_P_ksi_eta[0] );

    NHC_P_ksi_eps = exp_term * exp_term * NHC_P_ksi_eps +
      0.5 * dt_s * exp_term * sinch_term * NHC_P_G_eps / NHC_P_Q_eps;

    NHC_P_ksi_eta[NHC_P_length - 1] += 0.5 * dt_s * NHC_P_G_eta[NHC_P_length - 1] / NHC_P_Q_eta[NHC_P_length - 1];

    for (int k=1; k<NHC_P_length; k++)
    {
      exp_term   = exp( - 0.25 * dt_s * NHC_P_ksi_eta[NHC_P_length - k] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_P_ksi_eta[NHC_P_length - k] );

      NHC_P_ksi_eta[NHC_P_length - k - 1] = exp_term * exp_term * NHC_P_ksi_eta[NHC_P_length - k - 1]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_P_G_eta[NHC_P_length - k - 1] / NHC_P_Q_eta[NHC_P_length - k - 1];
    }

    double exp_term_t = exp( - dt_s * (1 + NHC_d/NHC_g) * NHC_P_ksi_eps );
    double exp_term_r = exp( - dt_s * (NHC_d/NHC_g) * NHC_P_ksi_eps );

    /* scale the velocities */

    /* loop through the active rigid bodies */
    for (int_it it = active_list.begin(); it != active_list.end(); it++)
    {

      rb = &(rigid_bodies[*it]);

      /* velocity */
      gsl_vector_scale(rb->v, exp_term_t);

      /* conjugate momenum of Q */
      rb->P.dat[0] *= exp_term_r;
      rb->P.dat[1] *= exp_term_r;
      rb->P.dat[2] *= exp_term_r;
      rb->P.dat[3] *= exp_term_r;

      /* angular momentum */
      rb->l.dat[0] *= exp_term_r;
      rb->l.dat[1] *= exp_term_r;
      rb->l.dat[2] *= exp_term_r;
      rb->l.dat[3] *= exp_term_r;

      /* angular velocity */
      rb->omega.dat[0] *= exp_term_r;
      rb->omega.dat[1] *= exp_term_r;
      rb->omega.dat[2] *= exp_term_r;
      rb->omega.dat[3] *= exp_term_r;

    }

    /* update the barostat positions */
    NHC_P_eps += dt_s * NHC_P_ksi_eps;
    for (int k=0; k<NHC_P_length; k++)
      NHC_P_eta[k] += dt_s * NHC_P_ksi_eta[k];

    /* update barostat forces */
    kinetic_t *= gsl_pow_2(exp_term_t);
    kinetic_r *= gsl_pow_2(exp_term_r);
    kinetic = kinetic_t + kinetic_r;

    NHC_P_G_eps = (2.0 * kinetic_t) + (vir_x + vir_y + vir_z)
                  - (NHC_d * init_volume * exp( NHC_d * NHC_P_eps ) * goal_pressure)
                  + NHC_d * (2.0 * kinetic) / NHC_g;

    exp_term   = exp( -0.25 * dt_s * NHC_P_ksi_eta[0] );
    sinch_term = sinch_10( 0.25 * dt_s * NHC_P_ksi_eta[0] );

    NHC_P_ksi_eps = exp_term * exp_term * NHC_P_ksi_eps +
      0.5 * dt_s * exp_term * sinch_term * NHC_P_G_eps / NHC_P_Q_eps;

    NHC_P_G_eta[0] = NHC_P_Q_eps * gsl_pow_2( NHC_P_ksi_eps ) - kT;

    /* update barostat velocities */
    for (int k=0; k<NHC_P_length-1; k++)
    {

      exp_term   = exp( - 0.25 * dt_s * NHC_P_ksi_eta[k+1] );
      sinch_term = sinch_10( 0.25 * dt_s * NHC_P_ksi_eta[k+1] );

      NHC_P_ksi_eta[k] = exp_term * exp_term * NHC_P_ksi_eta[k]
        + 0.5 * dt_s * exp_term * sinch_term * NHC_P_G_eta[k] / NHC_P_Q_eta[k];

      NHC_P_G_eta[k+1] = (NHC_P_Q_eta[k] * gsl_pow_2(NHC_P_ksi_eta[k]) - kT);

    }

    NHC_P_ksi_eta[NHC_P_length - 1] +=
      0.5 * dt_s * NHC_P_G_eta[NHC_P_length - 1] / NHC_P_Q_eta[NHC_P_length - 1];

  }

}

/***************************************************************************/

/* stepper */
void
emd::uni::dyn::step_NPT_miller_symplectic_NHC()
{

  rigid_body *                    rb;
  double                          exp_term, exp_term_2, sinch_term;

  /* phase 1. (NHC-T) */
  for (int i=0; i<THERMO_PART_ORDER; i++)
    step_NHC_thermostat_new ( 0.5 * dt / THERMO_PART_ORDER );

  /* phase 2. (NHC-P) */
  for (int i=0; i<BARO_PART_ORDER; i++)
    step_NHC_barostat_new ( 0.5 * dt / BARO_PART_ORDER );

  exp_term   = exp( 0.5 * dt * NHC_P_ksi_eps );
  exp_term_2 = exp_term * exp_term;
  sinch_term = sinch_10( 0.5 * dt * NHC_P_ksi_eps );

  /* phase 3. (particles) */

  /* loop through the active rigid bodies and do a half step */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* v(t + dt/2) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* external torque P(t + dt/2) */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* x(t + dt) */
    gsl_vector_scale( rb->x, exp_term_2 );
    gsl_vector_memcpy( tmp_v, r_cm );
    gsl_vector_scale( tmp_v, 1.0 - exp_term_2 );
    gsl_vector_add( rb->x, tmp_v );

    gsl_vector_memcpy( tmp_v, rb->v );
    gsl_vector_scale( tmp_v, dt * exp_term * sinch_term );
    gsl_vector_add( rb->x, tmp_v );

    /* torque free rotation Q(t + dt), P(t + dt/2) [2] */
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_1 ( *rb, dt );
    apply_no_squish_rot_P_Q_2 ( *rb, 0.5 * dt );
    apply_no_squish_rot_P_Q_3 ( *rb, 0.5 * dt );

    /* normalize Q(t + dt) */
    gsl_quaternion_normalize( &(rb->Q) );

    /* R(t + dt) */
    gsl_quaternion_unit_quat_to_mat( &(rb->Q), rb->R );

  }

  /* F(t + dt), N(t + dt) */
  emd::uni::clear_evals();
  emd::uni::eval_all_inter(true);
  emd::uni::eval_all_extern(true);
  emd::uni::eval_all_surf(true);

  /* loop through the active rigid bodies */
  for (int_it it = active_list.begin(); it != active_list.end(); it++)
  {

    rb = &(rigid_bodies[*it]);

    /* P(t + dt) */
    rb->P = apply_liouville_op_step_P ( rb->Q, rb->P, get_principal_frame_torque(*rb), 0.5 * dt );

    /* v(t + dt) */
    gsl_vector_memcpy( tmp_v, rb->force );
    gsl_vector_scale( tmp_v, 0.5 * dt / rb->param->mass );
    gsl_vector_add( rb->v, tmp_v );

    /* l(t + dt) = P(t + dt) / ( 2 * Q(t + dt) ) */
    rb->l = gsl_quaternion_mul( gsl_quaternion_inverse(rb->Q), rb->P );
    rb->l.dat[0] *= 0.5;
    rb->l.dat[1] *= 0.5;
    rb->l.dat[2] *= 0.5;
    rb->l.dat[3] *= 0.5;

    /* omega(t + dt) = I_body^{-1} l(t + dt) */
    rb->omega.dat[0] = 0;
    rb->omega.dat[1] = rb->l.dat[1] / rb->param->I1;
    rb->omega.dat[2] = rb->l.dat[2] / rb->param->I2;
    rb->omega.dat[3] = rb->l.dat[3] / rb->param->I3;

  }

  /* evaluate the kinetic energy */
  eval_kinetic();

  /* phase 4. (NHC-P) */
  for (int i=0; i<BARO_PART_ORDER; i++)
    step_NHC_barostat_new ( 0.5 * dt / BARO_PART_ORDER );

  /* phase 5. (NHC-T) */
  for (int i=0; i<THERMO_PART_ORDER; i++)
    step_NHC_thermostat_new ( 0.5 * dt / THERMO_PART_ORDER );

  /* adjust the volume */
  double new_vol = init_volume * exp( NHC_d * NHC_P_eps );

  adjust_volume ( new_vol );
  volume        = new_vol;
  concentration = rigid_bodies_count / volume;

  eval_NPT_hamiltonian();

  t = t + dt;

}

/***************************************************************************/

void
emd::uni::dyn::eval_NPT_hamiltonian()
{

  double kT = BOLZ_CONST * goal_temperature;

  /* NVE part */
  NPT_hamiltonian = potential + kinetic;

  /* T part */
  NPT_hamiltonian += (NHC_g_t * NHC_T_s_t[0] + NHC_g_r * NHC_T_s_r[0]) * kT;

  for (int i=1; i<NHC_T_length; i++)
    NPT_hamiltonian += (NHC_T_s_t[i] + NHC_T_s_r[i]) * kT;

  for (int i=0; i<NHC_T_length; i++)
  {
    NPT_hamiltonian += 0.5 * NHC_T_Q_t[i] * gsl_pow_2( NHC_T_ksi_t[i] );
    NPT_hamiltonian += 0.5 * NHC_T_Q_r[i] * gsl_pow_2( NHC_T_ksi_r[i] );
  }

  /* P part */
  NPT_hamiltonian += ( goal_pressure*volume + 0.5 * NHC_P_Q_eps * gsl_pow_2(NHC_P_ksi_eps) );

  for (int i=0; i<NHC_P_length; i++)
    NPT_hamiltonian += ( kT * NHC_P_eta[i] + 0.5 * NHC_P_Q_eta[i] * gsl_pow_2(NHC_P_ksi_eta[i]) );

}

/***************************************************************************/

void emd::uni::dyn::free_memory()
{

  /* TODO */

#ifdef DYNAMICS_DEBUG
  cout << "* emd::uni::dyn: destructed." << endl;
#endif

}
