/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "dynamics_job_manager.h"

using namespace std;

/***************************************************************************/

int
emd::dynman::init()
  {

  }

/***************************************************************************/

/* parses the input file */
int
emd::dynman::load_file(const std::string & input_file_name)
  {

    /* init parser, ... */
    init();

    /* opening the specified file as ro */
    ifstream inf(input_file_name.c_str(), ios::in);

    /* parsing */
    int stat =  emd::file_parser(inf, sizeof(KEYWORDS)/sizeof(string*), KEYWORDS, ACTIONS, '#', true);

    /* init dynamics */
    emd::uni::dyn::init();

    /* load nhc variables */
    if (nhc_file_supplied) load_NHC_file(nhc_file);

    cout << "* MD configuration: " << endl << endl
    << "  - integrator type:     " << emd::uni::dyn::integrator_desc << endl
    << "  - time step:           " << emd::uni::dyn::dt << " (arb)" << endl
    << "  - total steps:         " << steps << endl;
    if (emd::uni::dyn::fix_T)
      cout << "  - goal temperature:    " << emd::uni::dyn::goal_temperature << " (arb)" << endl;
    if (emd::uni::dyn::fix_P)
      cout << "  - goal pressure:       " << emd::uni::dyn::goal_pressure << " (arb)" << endl;
    cout << "  - NL update interval:  " << emd::uni::dyn::nl_update_interval << endl
    << "  - snapshot frequency:  " << snapshot_freq << endl
    << "  - stats frequency:     " << emd::uni::dyn::stats_interval << endl
    << "  - RE-squared cutoff:   " << emd::uni::re2::r_c << " (arb)" << endl
    << "  - RE-squared skin:     " << emd::uni::re2::r_s << " (arb)" << endl
    << "  - EES cutoff:          " << emd::uni::ees::r_c << " (arb)" << endl
    << "  - EES skin:            " << emd::uni::ees::r_s << " (arb)" << endl
    << "  - fixed CM:            " << (emd::uni::dyn::fix_cm ? "yes" : "no") << endl
    << "  - controlled temp.:    " << (emd::uni::dyn::fix_T ? "yes" : "no") << endl
    << "  - controlled pressure: " << (emd::uni::dyn::fix_P ? "yes" : "no") << endl
    << "  - log to file:         " << (emd::uni::dyn::log_to_file ? "yes" : "no") << endl
    << "  - log to stdout:       " << (emd::uni::dyn::log_to_stdout ? "yes" : "no") << endl
    << "  - log stats to file:   " << (emd::uni::dyn::log_to_stats_file ? "yes" : "no") << endl
    << "  - snapshot universe:   " << (save_universe_snapshot ? "yes" : "no") << endl
    << "  - snapshot povray:     " << (save_povray_snapshot ? "yes" : "no") << endl
    << "  - save final universe: " << (save_output ? "yes" : "no") << endl;
    if (emd::uni::dyn::log_to_file)
      cout << "  - log file name:       " << emd::uni::dyn::log_file_name << endl;
    if (emd::uni::dyn::log_to_stats_file)
      cout << "  - stats log file name: " << emd::uni::dyn::log_file_name_stats << endl;
    if (save_output)
      cout << "  - output file name:    " << universe_output_file << endl << endl;

    ready_flag = true;

    if (emd::uni::dyn::fix_P && !emd::uni::box_closed_flag)
      ready_flag = false;

    return stat;

  }

/***************************************************************************/

/* parses the input file */
int
emd::dynman::load_NHC_file(const std::string & input_file_name)
  {

    /* opening the specified file as ro */
    ifstream inf(input_file_name.c_str(), ios::in);

    /* parsing */
    int stat =  emd::file_parser(inf, sizeof(KEYWORDS)/sizeof(string*), KEYWORDS, ACTIONS, '#', true);

    return stat;

  }

/***************************************************************************/

void
emd::dynman::start()
  {
    //{{{restart : changes t to t_start
    if(save_restart) 
      emd::uni::dyn::t = emd::uni::loader::t_start;
    //}}}
    int snapshot_ctr = 0;
    //{{{restart :change where the output labeling begins
    int restart_ctr = 0;
    int cur_step = int(emd::uni::dyn::t/emd::uni::dyn::dt + 0.5);
    int snapshot_num = (cur_step - ignore_steps)/snapshot_freq;
    if (snapshot_num < 0) snapshot_num = 0;
    //}}}

    if (ready_flag)
      {

        cout << "* starting MD simulation..." << endl << endl;
/* iterations */
        //{{{restart :changed the starting point 
        for (long int iter = cur_step ; iter<steps; iter++)
        //}}}
          {

            emd::uni::dyn::step();

            //{{{restart: restart_ctr++
            restart_ctr++;
            //}}}

            //{{{restart: save restart file
            if (restart_ctr == restart_save_freq){
              if(save_restart)
                emd::uni::saver::save_universe_to_file(restart_file);
              restart_ctr = 0;
            }
            //}}}
            /* cease loging on demand */        
    if (iter < ignore_steps) continue;
            snapshot_ctr++;

            if (snapshot_ctr == snapshot_freq)
              {
        
                ostringstream os;
                os << setw(4) << setfill('0') << snapshot_num;


                if (save_universe_snapshot)

                  emd::uni::saver::save_universe_to_file( uni_snapshot_seq_file + os.str() + ".uni" );

                if (save_povray_snapshot)
                  emd::uni::saver::save_universe_to_povray_file( pov_snapshot_seq_file + os.str() + ".pov" );

                if (save_nhc_snapshot)
                  emd::uni::saver::save_nhc_vars_to_file( uni_snapshot_seq_file + os.str() + ".nhc" );

                snapshot_ctr = 0;
                snapshot_num++;
              }
          }

      }

    else

      {

        cout << "* exiting..." << endl;

      }


  }

/***************************************************************************/

void
emd::dynman::stop()
  {
    emd::uni::dyn::t=0;
    if (save_output)
      emd::uni::saver::save_universe_to_file( universe_output_file );

    cout << "\n* simulation finished successfully." << endl << endl;

  }

/***************************************************************************/

/* NOTE parsers */

/***************************************************************************/

int
emd::dynman::parse_universe_input_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    int stat = emd::uni::load_universe(tokenized_buffer[0]);

    if (stat==EXIT_FAILURE)
      return EXIT_FAILURE;

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_universe_output_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    universe_output_file = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_file_name = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_pov_snapshot_seq_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    pov_snapshot_seq_file = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_uni_snapshot_seq_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    uni_snapshot_seq_file = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_fix_cm(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::fix_cm = (bool)( atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_save_universe_snapshot(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    save_universe_snapshot = (bool)( atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_save_povray_snapshot(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    save_povray_snapshot = (bool)( atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_integrator_type(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    int integ_type = atoi( tokenized_buffer[0].c_str() );

    switch (integ_type)
    {

    case 0:
        {
          emd::uni::dyn::integrator = emd::uni::dyn::NVE_MILLER_SYMPLECTIC;
          break;
        }

    case 1:
        {
          emd::uni::dyn::integrator = emd::uni::dyn::NVT_MILLER_SYMPLECTIC_NHC;
          break;
        }

    case 2:
        {
          emd::uni::dyn::integrator = emd::uni::dyn::NPT_MILLER_SYMPLECTIC_NHC;
          break;
        }

    }

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_time_step(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::dt = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_ignore_steps(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    ignore_steps = atoi( tokenized_buffer[0].c_str() );
    emd::uni::dyn::N_ignore_steps = ignore_steps;

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_steps(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    steps = atoi( tokenized_buffer[0].c_str() );
    emd::uni::dyn::N_steps = steps;

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_goal_temperature(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::goal_temperature = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_goal_pressure(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::goal_pressure = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nl_update_interval(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::nl_update_interval = atoi( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_snapshot_freq(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    snapshot_freq = atoi( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_to_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_to_file = (bool)(atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_to_stdout(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_to_stdout = (bool)(atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_force_on_surfs(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_force_on_surfs = (bool)(atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_file_force_on_surfs(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_file_name_force_on_surfs = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_re2_cutoff(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::re2::r_c = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_gb_cutoff(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::gb::r_c = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_ees_cutoff(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::ees::r_c = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_re2_skin(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::re2::r_s = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_gb_skin(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::gb::r_s = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_ees_skin(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::ees::r_s = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_save_output(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    save_output = (bool)(atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_to_stats_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_to_stats_file = (bool)(atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_stats_log_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::log_file_name_stats = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_stats_interval(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::stats_interval = atoi(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_length(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_T_length = atoi(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_length(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_P_length = atoi(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_freq(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_T_freq = atof(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_freq(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_P_freq = atof(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    nhc_file_supplied = true;
    nhc_file = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_save_nhc_snapshot(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    save_nhc_snapshot = (bool)(atoi(tokenized_buffer[0].c_str()));

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_ksi_t(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_T_ksi_t[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_ksi_r(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_T_ksi_r[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_s_t(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_T_s_t[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_t_s_r(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_T_s_r[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_eta(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_P_eta[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_ksi_eta(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    for (int i=0; i<tokenized_buffer.size(); i++)
      emd::uni::dyn::NHC_P_ksi_eta[i] = atof(tokenized_buffer[i].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_eps(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_P_eps = atof(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_nhc_p_ksi_eps(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::NHC_P_ksi_eps = atof(tokenized_buffer[0].c_str());

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_log_temperature_drift(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<2)
      return EXIT_FAILURE;

    emd::uni::dyn::temperature_drift = true;
    emd::uni::dyn::temp_drift        = emd::uni::dyn::TEMP_DRIFT_LOG;
    emd::uni::dyn::init_temperature  = atof( tokenized_buffer[0].c_str() );
    emd::uni::dyn::final_temperature = atof( tokenized_buffer[1].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_lin_temperature_drift(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<2)
      return EXIT_FAILURE;

    emd::uni::dyn::temperature_drift = true;
    emd::uni::dyn::temp_drift        = emd::uni::dyn::TEMP_DRIFT_LIN;
    emd::uni::dyn::init_temperature  = atof( tokenized_buffer[0].c_str() );
    emd::uni::dyn::final_temperature = atof( tokenized_buffer[1].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_lin_concentration_drift(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    emd::uni::dyn::concentration_drift = true;
    emd::uni::dyn::final_concentration = atof( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_lin_sph_radius_drift(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<3)
      return EXIT_FAILURE;

    emd::uni::dyn::sph_radius_drift_spec spec;

    spec.idx       = atoi( tokenized_buffer[0].c_str() );
    spec.init_rad  = atof( tokenized_buffer[1].c_str() );
    spec.final_rad = atof( tokenized_buffer[2].c_str() );

    emd::uni::dyn::sph_radius_drift_vec.push_back(spec);

    return EXIT_SUCCESS;
  }

/***************************************************************************/
//{{{restart: definition of parse_restart_save_freq and parse_restart_save_freq and parse_restart_file
int
emd::dynman::parse_save_restart(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    save_restart = (bool)( atoi( tokenized_buffer[0].c_str() ) );

    return EXIT_SUCCESS;
  }
/***************************************************************************/
int
emd::dynman::parse_restart_save_freq(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    restart_save_freq = atoi( tokenized_buffer[0].c_str() );

    return EXIT_SUCCESS;
  }

/***************************************************************************/

int
emd::dynman::parse_restart_file(const std::string & param_string)
  {

    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    restart_file = tokenized_buffer[0];

    return EXIT_SUCCESS;
  }

/***************************************************************************/

//}}}:
