/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "emd-analyzer.h"
//test
using namespace std;
using namespace boost::filesystem;


/*** implementations *******************************************************/
int main(int argc, char *argv[])
  {
    /* splash */
    cout << "Ellipsoids Molecular Dynamics [Analyzer Module] ("
    << VERSION_STRING << ")" << endl;
    cout << "Kopyleft (k) 2007 Mehrtash Babadi (mehrtash@physics.sharif.edu)"
    << endl << endl;

    /* argument parsing and pre-processing */
    int stat = true;


    int ctr = 1;
    while (stat && (ctr < argc))
      {

        if (string(argv[ctr])=="-l")
          if (argc>ctr)
            {
              ctr++;
              log_file_path = argv[ctr];
              log_file_path_spec = true;
            }
          else
            stat = false;

        if (string(argv[ctr])=="-uni")
          if (argc>ctr)
            {
              ctr++;
              uni_file_path_spec = true;
              get_path_name(argv,argc,ctr,uni_file_path);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-s")
          if (argc>ctr)
            {
              ctr++;
              get_path_name(argv,argc,ctr,uni_snapshot_dir_path);
              uni_snapshot_dir_path_spec = true;
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-o")
          if (argc>ctr)
            {
              ctr++;
              output_dir_path = argv[ctr];
              output_dir_path_spec = true;
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-param")
          if (argc>ctr)
            {
              ctr++;
              emd::uni::loader::set_ext_param_file(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-eval_3d_rdf")
          {
            eval_3d_rdf = true;
            if (argc>ctr)
              {
                if (string(argv[ctr+1])=="-DFT")
                  {
                    eval_DFT_3d_rdf = true;
                    ctr++;
                  }
              }
            else
              stat = false;
          }
        else if (string(argv[ctr])=="-eval_2d_rdf")
          {
            eval_2d_rdf = true;
            if (argc>ctr)
              {
                if (string(argv[ctr+1])=="-DFT")
                  {
                    eval_DFT_2d_rdf = true;
                    ctr++;
                  }
              }
            else
              stat = false;
          }
        else if (string(argv[ctr])=="-sf")
          eval_sf = true;

        else if (string(argv[ctr])=="-order_param")
          eval_order_param = true;

        else if (string(argv[ctr])=="-count_regions")
          count_regions = true;

        else if (string(argv[ctr])=="-eval_tilt")
          eval_tilt = true;

        else if (string(argv[ctr])=="-density_fourier")
          eval_density_fourier = true;

        else if (string(argv[ctr])=="-structure_factor")
          eval_structure_factor = true;

        else if (string(argv[ctr])=="-neighbor_dist")
          if (argc>ctr)
            {
              ctr++;
              neighbor_dist = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-axis_idx")
          if (argc>ctr)
            {
              ctr++;
              axis_idx = atoi(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-max_rad")
          if (argc>ctr)
            {
              ctr++;
              rdf_max_rad = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-divs")
          if (argc>ctr)
            {
              ctr++;
              rdf_divs = atoi(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-freq_divs")
          if (argc>ctr)
            {
              ctr++;
              freq_divs = atoi(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-min_wl")
          if (argc>ctr)
            {
              ctr++;
              min_wl = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-iporder")
          if (argc>ctr)
            {
              ctr++;
              in_plane_order = atoi(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-snap_skip_rate")
          if (argc>ctr)
            {
              ctr++;
              snap_skip_rate = atoi(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-dd_penalty_coeff")
          if (argc>ctr)
            {
              ctr++;
              penalty_coeff = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-eval_smectic_config")
          eval_smectic_config = true;

        else if (string(argv[ctr])=="-eval_bond_order")
          eval_bond_order = true;

        else if (string(argv[ctr])=="-bond_order_r_min")
          if (argc>ctr)
            {
              ctr++;
              bond_order_r_min = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-bond_order_r_max")
          if (argc>ctr)
            {
              ctr++;
              bond_order_r_max = atof(argv[ctr]);
            }
          else
            stat = false;

        else if (string(argv[ctr])=="-eval_plane_ave")
          if (argc>(ctr+4))
            {

              eval_plane_ave = true;

              plane_ave_normal = atoi(argv[++ctr]);
              plane_ave_center = atof(argv[++ctr]);
              plane_ave_width = atof(argv[++ctr]);
              plane_ave_divs_1 = atoi(argv[++ctr]);
              plane_ave_divs_2 = atoi(argv[++ctr]);

              /* allocate memory */
              bin_data_vec = new bin_data[plane_ave_divs_1*plane_ave_divs_2];

            }
          else
            stat = false;

        else if (string(argv[ctr])=="-rel_dir")
          if (argc>(ctr+1))
            {
              eval_rel_dir = true;
              ctr++;
              class_id1 = atoi(argv[ctr]);
              ctr++;
              class_id2 = atoi(argv[ctr]);
            }
          else
            stat = false;

        //changed:

        else if (string(argv[ctr])=="-class_id")
          if (argc>ctr)
            {
              ctr++;
              class_id = atoi(argv[ctr]);
            }
          else
            stat = false;
        ctr++;

      }

    if (!output_dir_path_spec)
      stat = false;

    if (!stat)
      {
        cout << "usage: emd-analyzer [OPTIONS]" << endl << endl
        << "     -l <log_file_path>         The path of the log file"
        << endl
        << "     -s <snapshot_dir_path>     The path of the directory of universe snapshots"
        << endl
        << "     -o <output_dir_path>       The path of the output directory"
        << endl
        << "     -eval_3d_rdf               Evaluate scalar, para. and perp. RDFs"
        << endl
        << "     -eval_2d_rdf               Evaluate in-plane 2D RDF"
        << endl
        << "     -eval_tilt                 Evaluate the director-plane tilt angle"
        << endl
        << "     -axis_idx <val>            Axis index for P_2 (0, 1, 2) [default: 2]"
        << endl
        << "     -max_rad <val>             Maximum range of RDFs [default: 8.0]"
        << endl
        << "     -divs <val>                Number of division points in the range of RDFs [default: 200]"
        << endl
        << "     -min_wl <val>              Minumum possible wavelength [default: 1.0]"
        << endl
        << "     -iporder <val>             In-plane RDF order [default: 0]"
        << endl
        << "     -snap_skip_rate <val>      Snapshot skipping rates [default: 0]"
        << endl
        << "     -dd_penalty_coeff <val>    Director deviation penalty coefficient [default: 1.0]"
        << endl
        << "     -eval_smectic_config       Configuration of a smectic plane"
        << endl
        << "     -eval_bond_order           Bond orientational order"
        << endl
        << "     -bond_order_r_min          Bond order min cutoff [default: 2.0]"
        << endl
        << "     -bond_order_r_max          Bond order max cutoff [default: 3.0]"
        << endl
        << "     -eval_plane_ave <normal_idx> <center> <width> <divs_1> <divs_2>"
        << endl
        << "                                Evaluate averages (density, OP) in a plane"
        << endl
        //changed:
        << "     -DFT                       Evaluate the DFT of time averaged RDFs "
        << endl
        << "     -uni <uni_file_path>       The path of the uni file "
        <<endl
        << "     -sf                        Evaluate structure factor and rdf for a single snapshot "
        <<endl
        << "     -class_id <val>            Class(label) of the particle to be analyzed"
        <<endl
        << "     -count_regions             Counts the number of regions created by particle <class_id> after spinodal decomposition  "  
        <<endl
        << "     -neighbor_dist <val>       The size of the cell for determination of regions"
        <<endl
        << "     -param <param_file_name>   If this field is supplied the param file will be read from param_file_name rather than the location universe files point to "
        <<endl 
        << "     -density_fourier           Calculates the fourier transformation of the density "
        << endl
        << "     -freq_divs <val>           Number of division points in the range of frequencies [default: 20]"
        <<endl 
        << "     -structure_factor          Calculates the structure factors for number density and concentration "
        <<endl 
        << "     -order_param               Calculates the order parameter for particle <class_id> "
        <<endl
        << "     -rel_dir <class_id1> <class_id2> "
        <<endl
        <<"                                 Calculate the angle between director of particle <class_id1> and <class_id2> "
        << endl<<endl;

        return EXIT_FAILURE;
      }

    /* phase 1. parse the log file */

    if (log_file_path_spec)
      {

        ifstream log_file(log_file_path.c_str());

        if (!log_file_path_spec || !log_file)
          {
            cout << "emd-analyzer: error opening the log file!" << endl;
            return EXIT_FAILURE;
          }

        stat = emd::file_parser(log_file, sizeof(LOG_KEYWORDS)/sizeof(string*),
            LOG_KEYWORDS, LOG_ACTIONS, '#', true);

        /* phase 1.1  stats of the log-file */

        cout << "emd-analyzer: analyzing the log file... " << endl;
        path log_stat_out_path = path(output_dir_path)/path("stats.log");
        log_stat_of.open(log_stat_out_path.native_file_string().c_str());

        if (!log_stat_of)
          {
            cout << "emd-analyzer: error creating output files (stats.log)!"
            << endl;
            return EXIT_FAILURE;
          }

        log_stat_of << "#########################" << endl
        << "## log file statistics ##" << endl
        << "#########################" << endl << endl;

        if (time_vec.size()>0)
          log_stat_of << left << setw(DEF_WIDTH) << "total analyzed steps: "
          << time_vec.size() << endl << left << setw(DEF_WIDTH)
          << "total time: " << (time_vec[time_vec.size()-1]-time_vec[0]) << endl;

        if (potential_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(potential_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "potential mean: "
            << mean << endl << left << setw(DEF_WIDTH) << "potential std: "
            << std << endl;
          }

        if (energy_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(energy_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "energy mean: " << mean
            << endl << left << setw(DEF_WIDTH) << "energy std: " << std
            << endl;
          }

        if (temperature_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(temperature_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "temperature mean: "
            << mean << endl << left << setw(DEF_WIDTH)
            << "temperature std: " << std << endl;
          }

        if (temperature_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(pressure_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "pressure mean: " << mean
            << endl << left << setw(DEF_WIDTH) << "pressure std: " << std
            << endl;
          }

        if (order_param_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(order_param_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "order parameter mean: "
            << mean << endl << left << setw(DEF_WIDTH)
            << "order parameter std: " << std << endl;
          }

        if (hamiltonian_vec.size()>0)
          {
            double mean, std;
            eval_stats_vec_double(hamiltonian_vec, mean, std);
            log_stat_of << left << setw(DEF_WIDTH) << "hamiltonian mean: "
            << mean << endl << left << setw(DEF_WIDTH)
            << "hamiltonian std: " << std << endl;
          }
        log_stat_of.close();
      }
    //Single-input analyzing

    if (uni_file_path_spec)
      {
        emd::uni::load_universe(uni_file_path, false);
        cout<<" Input universe : "<<uni_file_path<<endl<<endl;
        if(eval_rel_dir)
          {
            gsl_vector * temp_director = gsl_vector_alloc(3);
            eval_director(class_id1);
            cout<<endl;
            gsl_vector_memcpy(temp_director,director);
            eval_director(class_id2);
            double dot_product;
            gsl_blas_ddot(temp_director,director,&dot_product);
            double cos_theta = 
              fabs(dot_product)/gsl_blas_dnrm2(temp_director)/gsl_blas_dnrm2(director);
            cout<<endl<<"      Cos(theta) = "<<cos_theta<<endl;
            cout<<      "      theta      = "<<acos(cos_theta)*180/M_PI<<"  degrees"<<endl;
          }

        if(eval_structure_factor)
          {
            eval_director(class_id);

            gsl_vector * para_Snn = gsl_vector_alloc(freq_divs);
            gsl_vector * scalar_Snn = gsl_vector_alloc(freq_divs);
            gsl_vector * perp_Snn = gsl_vector_alloc(freq_divs);     
            gsl_vector * para_Scc = gsl_vector_alloc(freq_divs);
            gsl_vector * scalar_Scc = gsl_vector_alloc(freq_divs);
            gsl_vector * perp_Scc = gsl_vector_alloc(freq_divs);     

            double dk;
            emd::uni::total_structure_factor(director,freq_divs,min_wl,dk,
                para_Snn , perp_Snn , scalar_Snn,
                para_Scc , perp_Scc , scalar_Scc);

            emd::uni::free_memory();

            path Snn_path = path(output_dir_path)/path("Snn.log");
            path Scc_path = path(output_dir_path)/path("Scc.log");

            ofstream Snn_ouf(Snn_path.native_file_string().c_str());
            if(!Snn_ouf){
              cerr<<"Error creating "<<Snn_path.native_file_string()<<endl;
              exit(0);
            }

            ofstream Scc_ouf(Scc_path.native_file_string().c_str());
            if(!Scc_ouf){
              cerr<<"Error creating "<<Scc_path.native_file_string()<<endl;
              exit(0);
            }

            Scc_ouf<<setw(15)<<" # freq "
            <<setw(15)<<"scalar"<<setw(15)<<"para"<<setw(15)<<"perp"
            <<endl<<endl;

            Snn_ouf<<setw(15)<<" # freq "
            <<setw(15)<<"scalar"<<setw(15)<<"para"<<setw(15)<<"perp"
            <<endl<<endl;

            for(unsigned int i=1;i<freq_divs;i++){

              Snn_ouf<<setw(15)<<i*dk
              <<setw(15)<<gsl_vector_get(scalar_Snn,i)
              <<setw(15)<<gsl_vector_get(para_Snn,i)
              <<setw(15)<<gsl_vector_get(perp_Snn,i)
              <<endl;

              Scc_ouf<<setw(15)<<i*dk
              <<setw(15)<<gsl_vector_get(scalar_Scc,i)
              <<setw(15)<<gsl_vector_get(para_Scc,i)
              <<setw(15)<<gsl_vector_get(perp_Scc,i)
              <<endl;
            }

            gsl_vector_free(scalar_Snn);
            gsl_vector_free(para_Snn);
            gsl_vector_free(perp_Snn);
            gsl_vector_free(scalar_Scc);
            gsl_vector_free(para_Scc);
            gsl_vector_free(perp_Scc);

            Scc_ouf.close();
            Snn_ouf.close();

          }
        if(eval_density_fourier){

          eval_director(class_id);


          gsl_vector * para_freq = gsl_vector_alloc(freq_divs);
          gsl_vector * scalar_freq = gsl_vector_alloc(freq_divs);
          gsl_vector * perp_freq = gsl_vector_alloc(freq_divs);
          double dk;
          emd::uni::total_density_fourier(class_id,director,freq_divs
              ,min_wl,dk,para_freq,perp_freq,scalar_freq);
          emd::uni::free_memory();
          path dens_four_out_path = path(output_dir_path)/path("dens_four.log");

          dens_four_ouf.open(dens_four_out_path.native_file_string().c_str());
          if(!dens_four_ouf){
            cerr<<"Error creating "<<dens_four_out_path.native_file_string()<<endl;
            exit(0);
          }
          dens_four_ouf<<setw(15)<<" # freq "
          <<setw(15)<<"scalar"<<setw(15)<<"para"<<setw(15)<<"perp"
          <<endl<<endl;
          for(unsigned int i=1;i<freq_divs;i++){
            dens_four_ouf<<setw(15)<<i*dk
            <<setw(15)<<gsl_vector_get(scalar_freq,i)
            <<setw(15)<<gsl_vector_get(para_freq,i)
            <<setw(15)<<gsl_vector_get(perp_freq,i)
            <<endl;
          }
          gsl_vector_free(scalar_freq);
          gsl_vector_free(para_freq);
          gsl_vector_free(perp_freq);
          dens_four_ouf.close();
        }

        if(count_regions){

          cout<<"Number of regions as a result of spinodal decomposition for particle "
          <<class_id<<" is : "
          <<emd::uni::eval_number_of_regions(class_id,neighbor_dist,neighbor_dist,neighbor_dist)
          <<endl;

        }
        if (eval_sf)
          {
            eval_director(class_id);

            gsl_vector * curr_para_dist = gsl_vector_calloc(rdf_divs);
            gsl_vector * curr_perp_dist = gsl_vector_calloc(rdf_divs);
            gsl_vector * curr_scalar_dist = gsl_vector_calloc(rdf_divs);

            emd::uni::eval_radial_dist(class_id, rdf_max_rad, rdf_divs,
                director, curr_para_dist, curr_perp_dist, curr_scalar_dist);

            gsl_vector * curr_in_plane_dist = gsl_vector_calloc(rdf_divs);

            eval_plane();

            emd::uni::eval_in_plane_radial_dist(class_id, in_plane_order,
                rdf_max_rad, rdf_divs, k_x, k_y, k_z, phi, curr_in_plane_dist);

            emd::uni::free_memory();

            path single_rdf_out_path = path(output_dir_path)
            /path("single_rdf.log");
            ofstream single_rdf_of(single_rdf_out_path.native_file_string().c_str());

            single_rdf_of<<setw(10)<<"#r" <<setw(20)<<"scalar" <<setw(20)
            <<"para" <<setw(20)<<"perp" <<setw(20)<<"in_plane" <<endl;

            for (int i=0; i<rdf_divs; i++)
              {
                single_rdf_of<<setw(10)<< ((i+0.5)*rdf_max_rad/rdf_divs)
                <<setw(20)<<gsl_vector_get(curr_scalar_dist, i) <<setw(20)
                <<gsl_vector_get(curr_para_dist, i) <<setw(20)
                <<gsl_vector_get(curr_perp_dist, i) <<setw(20)
                <<gsl_vector_get(curr_in_plane_dist, i) <<endl;
              }
            scalar_rdf_of.close();

            // structure factor
            path single_sf_out_path = path(output_dir_path)
            /path("single_sf.log");
            ofstream single_sf_of(single_sf_out_path.native_file_string().c_str());

            single_sf_of<<setw(10)<<"#freq" <<setw(20)<<"scalar abs" <<setw(20)
            <<"scalar arg" <<setw(20)<<"para abs" <<setw(20)<<"para arg"
            <<setw(20)<<"perp abs " <<setw(20)<<"perp arg " <<setw(20)
            <<"in_plane abs" <<setw(20)<<"in_plane arg" <<endl;

            gsl_complex scalar_dft;
            gsl_complex para_dft;
            gsl_complex perp_dft;
            gsl_complex in_plane_dft;

            for (int k=1; k<rdf_divs/2; k++)
              {

                emd::uni::discrete_fourier_transform(curr_scalar_dist, k,
                    scalar_dft);
                emd::uni::discrete_fourier_transform(curr_para_dist, k,
                    para_dft);
                emd::uni::discrete_fourier_transform(curr_perp_dist, k,
                    perp_dft);
                emd::uni::discrete_fourier_transform(curr_in_plane_dist, k,
                    in_plane_dft);

                single_sf_of <<setw(10)<< 2*M_PI*k/rdf_divs <<setw(20)
                <<gsl_complex_abs(scalar_dft) <<setw(20)
                <<gsl_complex_arg(scalar_dft) <<setw(20)
                <<gsl_complex_abs(para_dft) <<setw(20)
                <<gsl_complex_arg(para_dft) <<setw(20)
                <<gsl_complex_abs(perp_dft) <<setw(20)
                <<gsl_complex_arg(perp_dft) <<setw(20)
                <<gsl_complex_abs(in_plane_dft) <<setw(20)
                <<gsl_complex_arg(in_plane_dft) <<endl;
              }
            single_sf_of.close();
          }
      }

    //End of Single-input analyzing
    if (uni_snapshot_dir_path_spec && (eval_2d_rdf || eval_3d_rdf || eval_tilt
        || eval_smectic_config || eval_bond_order || eval_plane_ave || eval_order_param))
      {
        /* phase 1.2  analyzing the universe snapshots */

        cout << "emd-analyzer: analyzing the universe snapshots... " << endl;
        path path_uni_snapshot_dir_path(uni_snapshot_dir_path);

        if (!is_directory(path_uni_snapshot_dir_path))
          {
            cout << "emd-analyzer: error opening the snapshots directory!"
            << endl;
            return EXIT_FAILURE;
          }

        /* look into the snapshots directory */
        directory_iterator end_iter;
        string curr_uni_file;

        int skip_counter = 0;
        for (directory_iterator dir_iter(path_uni_snapshot_dir_path); dir_iter
        != end_iter; dir_iter++)
          {

            try
            {
              if (!is_directory(*dir_iter))
                {

                  /* skipping */
                  if (skip_counter<snap_skip_rate)
                    {
                      skip_counter++;
                      continue;
                    }
                  else skip_counter = 0;

                  curr_uni_file = dir_iter -> path().native_file_string();//path().native_file_string();
                  cout << "emd-analyzer: analyzing " << curr_uni_file << "..." << endl;

                  if (emd::uni::load_universe(curr_uni_file, false)!=EXIT_SUCCESS)
                    {
                      emd::uni::free_memory();
                      continue;
                    }

                  N_snapshots++;

                  /* flags to keep track of director/plane calculations */
                  director_is_eval = false;
                  plane_is_eval = false;
                  if(concentration==0)
                    {
                      if (emd::uni::eval_volume()==EXIT_FAILURE)
                        exit(0);
                      concentration = emd::uni::rigid_bodies_count / emd::uni::volume;
                    }

                  if (eval_bond_order)
                    {

                      if (!director_is_eval)
                        {
                          eval_director(class_id);
                          director_is_eval = true;
                        }

                      if (!plane_is_eval)
                        {
                          eval_plane();
                          plane_is_eval = true;
                        }

                      double bond_order = emd::uni::eval_bond_order (k_x, k_y, k_z, bond_order_r_min,
                          bond_order_r_max);
                      bond_orders.push_back(bond_order);

                      cout << "  * bond order: " << bond_order << endl;

                    }

                  if (eval_smectic_config)
                    {

                      vector<emd::rigid_body_ref> indices;
                      emd::rigid_body_ref ref;

                      if (!director_is_eval)
                        {
                          eval_director(class_id);
                          director_is_eval = true;
                        }

                      if (!plane_is_eval)
                        {
                          eval_plane();
                          plane_is_eval = true;
                        }

                      for ( unsigned int m=0; m < emd::uni::rigid_bodies_count; m++ )

                        {
                          //changed:
                          if(emd::uni::rigid_bodies[m].class_id != class_id) continue;

                          phi = k_x*gsl_vector_get(emd::uni::rigid_bodies[m].x, 0) +
                          k_y*gsl_vector_get(emd::uni::rigid_bodies[m].x, 1) +
                          k_z*gsl_vector_get(emd::uni::rigid_bodies[m].x, 2);
                          int pl_idx = (int)floor(0.5 + phi/(2*M_PI));

                          if ( pl_idx == 0 )
                            {
                              ref.idx = m;

                              ref.n_x = 0;
                              ref.n_y = 0;
                              ref.n_z = 0;

                              indices.push_back(ref);
                            }

                        }

                      path pov_plane_out_path = path(output_dir_path)/path(dir_iter->leaf()+"-plane.pov");
                      emd::uni::saver::save_universe_to_povray_file(indices, pov_plane_out_path.native_file_string().c_str() );

                    }

                  if (eval_tilt)
                    {

                      if (!director_is_eval)
                        {
                          eval_director(class_id);
                          director_is_eval = true;
                        }

                      if (!plane_is_eval)
                        {
                          eval_plane();
                          plane_is_eval = true;
                        }

                      double c_tilt_angle =
                        (k_x*gsl_vector_get(director, 0) +
                            k_y*gsl_vector_get(director, 1) +
                            k_z*gsl_vector_get(director, 2)) / (sqrt(k_x*k_x + k_y*k_y + k_z*k_z) * gsl_blas_dnrm2(director));

                      c_tilt_angle = 180*acos(c_tilt_angle)/M_PI;

                      cout << "  * tilt angle: " << c_tilt_angle << endl;

                      tilt_angles.push_back(c_tilt_angle);

                    }
                  if(eval_order_param)
                    {
                      cout<<"   * evaluating the order parameter..."<<endl;
                      vector<unsigned int> particle_indices;
                      for ( unsigned int m = 0; m < emd::uni::rigid_bodies_count; m++ )
                        {
                          if(emd::uni::rigid_bodies[m].class_id==class_id)
                            {
                              particle_indices.push_back(m);
                            }
                        }
                      emd::uni::eval_Q_tensor(particle_indices, axis_idx, c_Q_tensor, Q_eigen_val, Q_eigen_vec);
                      double order_param = gsl_vector_max( Q_eigen_val );
                      order_params.push_back(order_param);
                    }

                  if (eval_3d_rdf)
                    {

                      if (!director_is_eval)
                        {
                          eval_director(class_id);
                          director_is_eval = true;
                        }

                      /* evaluate the RDFs */
                      gsl_vector * curr_para_dist = gsl_vector_calloc(rdf_divs);
                      gsl_vector * curr_perp_dist = gsl_vector_calloc(rdf_divs);
                      gsl_vector * curr_scalar_dist = gsl_vector_calloc(rdf_divs);

                      cout << "  * evaluating the 3D distribution functions..." << endl;

                      emd::uni::eval_radial_dist(class_id,rdf_max_rad, rdf_divs, director, curr_para_dist,
                          curr_perp_dist, curr_scalar_dist);

                      scalar_rdfs.push_back(curr_scalar_dist);
                      para_rdfs.push_back(curr_para_dist);
                      perp_rdfs.push_back(curr_perp_dist);

                    }

                  if (eval_2d_rdf)
                    {

                      gsl_vector * curr_in_plane_dist = gsl_vector_calloc(rdf_divs);

                      if (!director_is_eval)
                        {
                          eval_director(class_id);
                          director_is_eval = true;
                        }

                      if (!plane_is_eval)
                        {
                          eval_plane();
                          plane_is_eval = true;
                        }

                      cout << "  * evaluating in-plane 2D distribution function..." << endl;

                      emd::uni::eval_in_plane_radial_dist(class_id,in_plane_order, rdf_max_rad, rdf_divs, k_x, k_y, k_z,
                          phi, curr_in_plane_dist);

                      /* add to the RDF vectors */
                      in_plane_rdfs.push_back(curr_in_plane_dist);

                    }

                  if (eval_plane_ave)
                    {

                      cout << "  * evaluating in-plain averages..." << endl;

                      /* find bounds in the plane */
                      switch (plane_ave_normal)
                        {
                      case AXIS_X:
                        {
                          plane_ave_y_lower = emd::uni::y_lower_bound;
                          plane_ave_y_upper = emd::uni::y_upper_bound;
                          plane_ave_z_lower = emd::uni::z_lower_bound;
                          plane_ave_z_upper = emd::uni::z_upper_bound;
                          break;
                        }
                      case AXIS_Y:
                        {
                          plane_ave_y_lower = emd::uni::z_lower_bound;
                          plane_ave_y_upper = emd::uni::z_upper_bound;
                          plane_ave_z_lower = emd::uni::x_lower_bound;
                          plane_ave_z_upper = emd::uni::x_upper_bound;
                          break;
                        }
                      case AXIS_Z:
                        {
                          plane_ave_y_lower = emd::uni::x_lower_bound;
                          plane_ave_y_upper = emd::uni::x_upper_bound;
                          plane_ave_z_lower = emd::uni::y_lower_bound;
                          plane_ave_z_upper = emd::uni::y_upper_bound;
                          break;
                        }
                        }

                      for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
                        {
                          //changed:

                          if(emd::uni::rigid_bodies[m].class_id != class_id) continue;

                          double x, y, z;
                          int idx_y, idx_z, idx_bin;

                          /* permute the coordinates with respect to the normal to the plane */
                          switch (plane_ave_normal)
                            {
                          case AXIS_X:
                            {
                              x = gsl_vector_get(emd::uni::rigid_bodies[m].x, 0);
                              y = gsl_vector_get(emd::uni::rigid_bodies[m].x, 1);
                              z = gsl_vector_get(emd::uni::rigid_bodies[m].x, 2);
                              break;
                            }
                          case AXIS_Y:
                            {
                              x = gsl_vector_get(emd::uni::rigid_bodies[m].x, 1);
                              y = gsl_vector_get(emd::uni::rigid_bodies[m].x, 2);
                              z = gsl_vector_get(emd::uni::rigid_bodies[m].x, 0);
                              break;
                            }
                          case AXIS_Z:
                            {
                              x = gsl_vector_get(emd::uni::rigid_bodies[m].x, 2);
                              y = gsl_vector_get(emd::uni::rigid_bodies[m].x, 0);
                              z = gsl_vector_get(emd::uni::rigid_bodies[m].x, 1);
                              break;
                            }
                            }

                          /* find the idx */
                          if (abs(x - plane_ave_center) > 0.5*plane_ave_width) continue;

                          idx_y = (int)floor(plane_ave_divs_1*(y - plane_ave_y_lower)/(plane_ave_y_upper - plane_ave_y_lower));
                          idx_z = (int)floor(plane_ave_divs_2*(z - plane_ave_z_lower)/(plane_ave_z_upper - plane_ave_z_lower));
                          idx_bin = idx_z*plane_ave_divs_1 + idx_y;

                          /* find the Q_tensor */
                          vector<unsigned int> c_ref;
                          c_ref.push_back(m);
                          emd::uni::eval_Q_tensor(c_ref, axis_idx, c_Q_tensor);

                          /* update the bin */
                          gsl_matrix_add(bin_data_vec[idx_bin].Q_tensor, c_Q_tensor); /* Q_tensor */
                          bin_data_vec[idx_bin].N_samples++; /* number of samples */

                        }

                    }

                  /* free the memory */
                  emd::uni::free_memory();

                }

            }
            catch (const exception & ex)
              {
                cout << "emd-analyzer: " << ex.what() << endl;
              }

          }

        /* NOTE post-processing and output */

        /***********************************************************************/
        if (eval_bond_order)
          {

            double mean, std;
            eval_stats_vec_double(bond_orders, mean, std);

            path bond_out_path = path(output_dir_path)/path("bond.log");
            ofstream bond_of(bond_out_path.native_file_string().c_str());

            bond_of << left << setw(DEF_WIDTH) << "bond order mean: " << mean
            << endl << left << setw(DEF_WIDTH) << "bond order std: " << std
            << endl;

            bond_of.close();

          }

        /***********************************************************************/
        if (eval_tilt)
          {

            double mean, std;
            eval_stats_vec_double(tilt_angles, mean, std);

            path tilt_out_path = path(output_dir_path)/path("tilt.log");
            ofstream tilt_of(tilt_out_path.native_file_string().c_str());

            tilt_of << left << setw(DEF_WIDTH) << "tilt angle mean: " << mean
            << endl << left << setw(DEF_WIDTH) << "tilt angle std: " << std
            << endl;

            tilt_of.close();

          }

        /***********************************************************************/
        if(eval_order_param)
          {
            path param_path = path(output_dir_path)/path("param.log");
            param_ouf.open(param_path.native_file_string().c_str(),ios::app);

            double mean, std;
            eval_stats_vec_double(order_params, mean, std);
            param_ouf<<setw(20)<<concentration<<setw(20)<<mean<<setw(20)<<std<<endl;

          }
        if (eval_3d_rdf)
          {

            scalar_rdf_mean = gsl_vector_calloc(rdf_divs);
            scalar_rdf_std = gsl_vector_calloc(rdf_divs);
            para_rdf_mean = gsl_vector_calloc(rdf_divs);
            para_rdf_std = gsl_vector_calloc(rdf_divs);
            perp_rdf_mean = gsl_vector_calloc(rdf_divs);
            perp_rdf_std = gsl_vector_calloc(rdf_divs);

            for (int i=0; i<rdf_divs; i++)
              {

                vector<double> curr_row;
                double mean, std;

                /* scalar */
                for (int j=0; j<scalar_rdfs.size(); j++)
                  curr_row.push_back(gsl_vector_get(scalar_rdfs[j], i));

                eval_stats_vec_double(curr_row, mean, std);
                gsl_vector_set(scalar_rdf_mean, i, mean);
                gsl_vector_set(scalar_rdf_std, i, std);

                curr_row.clear();

                /* para */
                for (int j=0; j<para_rdfs.size(); j++)
                  curr_row.push_back(gsl_vector_get(para_rdfs[j], i));

                eval_stats_vec_double(curr_row, mean, std);
                gsl_vector_set(para_rdf_mean, i, mean);
                gsl_vector_set(para_rdf_std, i, std);

                curr_row.clear();

                /* perp */
                for (int j=0; j<perp_rdfs.size(); j++)
                  curr_row.push_back(gsl_vector_get(perp_rdfs[j], i));

                eval_stats_vec_double(curr_row, mean, std);
                gsl_vector_set(perp_rdf_mean, i, mean);
                gsl_vector_set(perp_rdf_std, i, std);

                curr_row.clear();

              }

            /* output to file (scalar_rdf.log) */
            path scalar_rdf_out_path = path(output_dir_path)
            /path("scalar_rdf.log");
            scalar_rdf_of.open(scalar_rdf_out_path.native_file_string().c_str());

            if (!scalar_rdf_of)
              {
                cout
                << "emd-analyzer: error creating output files (scalar_rdf.log)!"
                << endl;
                return EXIT_FAILURE;
              }

            for (int i=0; i<rdf_divs; i++)
              {
                scalar_rdf_of << ((i+0.5)*rdf_max_rad/rdf_divs) << "  "
                << gsl_vector_get(scalar_rdf_mean, i) << "  "
                << gsl_vector_get(scalar_rdf_std, i) << endl;
              }
            scalar_rdf_of.close();
            //scalar structure factor
            if (eval_DFT_3d_rdf)
              {
                path scalar_SF_out_path = path(output_dir_path)
                /path("scalar_SF.log");
                scalar_SF_of.open(scalar_SF_out_path.native_file_string().c_str());

                if (!scalar_SF_of)
                  {
                    cout
                    << "emd-analyzer: error creating output files (scalar_SF.log)!"
                    << endl;
                    return EXIT_FAILURE;
                  }
                gsl_complex comp;
                for (int k=1; k<rdf_divs/2; k++)
                  {
                    emd::uni::discrete_fourier_transform(scalar_rdf_mean, k,
                        comp);
                    scalar_SF_of << 2*M_PI*k/rdf_divs << "        "
                    << gsl_complex_abs(comp)<<"        "
                    << gsl_complex_arg(comp)<< endl;
                  }
                scalar_SF_of.close();
              }

            /* output to file (para_rdf.log) */
            path para_rdf_out_path = path(output_dir_path)/path("para_rdf.log");
            para_rdf_of.open(para_rdf_out_path.native_file_string().c_str());

            if (!para_rdf_of)
              {
                cout
                << "emd-analyzer: error creating output files (para_rdf.log)!"
                << endl;
                return EXIT_FAILURE;
              }

            for (int i=0; i<rdf_divs; i++)
              {
                para_rdf_of << ((i+0.5)*rdf_max_rad/rdf_divs) << "  "
                << gsl_vector_get(para_rdf_mean, i) << "  "
                << gsl_vector_get(para_rdf_std, i) << endl;
              }
            para_rdf_of.close();

            //para structure factor
            if (eval_DFT_3d_rdf)
              {
                path para_SF_out_path = path(output_dir_path)
                /path("para_SF.log");
                para_SF_of.open(para_SF_out_path.native_file_string().c_str());

                if (!para_SF_of)
                  {
                    cout
                    << "emd-analyzer: error creating output files (para_SF.log)!"
                    << endl;
                    return EXIT_FAILURE;
                  }
                gsl_complex comp;
                for (int k=1; k<rdf_divs/2; k++)
                  {
                    emd::uni::discrete_fourier_transform(para_rdf_mean, k, comp);
                    para_SF_of << 2*M_PI*k/rdf_divs << "        "
                    << gsl_complex_abs(comp)<<"        "
                    << gsl_complex_arg(comp)<< endl;
                  }
                para_SF_of.close();
              }

            /* output to file (perp_rdf.log) */
            path perp_rdf_out_path = path(output_dir_path)/path("perp_rdf.log");
            perp_rdf_of.open(perp_rdf_out_path.native_file_string().c_str());

            if (!perp_rdf_of)
              {
                cout
                << "emd-analyzer: error creating output files (perp_rdf.log)!"
                << endl;
                return EXIT_FAILURE;
              }

            for (int i=0; i<rdf_divs; i++)
              {
                perp_rdf_of << ((i+0.5)*rdf_max_rad/rdf_divs) << "  "
                << gsl_vector_get(perp_rdf_mean, i) << "  "
                << gsl_vector_get(perp_rdf_std, i) << endl;
              }
            perp_rdf_of.close();

            //perp structure factor
            if (eval_DFT_3d_rdf)
              {
                path perp_SF_out_path = path(output_dir_path)
                /path("perp_SF.log");
                perp_SF_of.open(perp_SF_out_path.native_file_string().c_str());

                if (!perp_SF_of)
                  {
                    cout
                    << "emd-analyzer: error creating output files (perp_SF.log)!"
                    << endl;
                    return EXIT_FAILURE;
                  }
                gsl_complex comp;
                for (int k=1; k<rdf_divs/2; k++)
                  {
                    emd::uni::discrete_fourier_transform(perp_rdf_mean, k, comp);
                    perp_SF_of << 2*M_PI*k/rdf_divs << "        "
                    << gsl_complex_abs(comp)<<"        "
                    << gsl_complex_arg(comp)<< endl;
                  }
                perp_SF_of.close();
              }

          }

        /***********************************************************************/
        if (eval_2d_rdf)
          {

            in_plane_rdf_mean = gsl_vector_calloc(rdf_divs);
            in_plane_rdf_std = gsl_vector_calloc(rdf_divs);

            for (int i=0; i<rdf_divs; i++)
              {

                vector<double> curr_row;
                double mean, std;

                /* in-plane */
                for (int j=0; j<in_plane_rdfs.size(); j++)
                  curr_row.push_back(gsl_vector_get(in_plane_rdfs[j], i));

                eval_stats_vec_double(curr_row, mean, std);
                gsl_vector_set(in_plane_rdf_mean, i, mean);
                gsl_vector_set(in_plane_rdf_std, i, std);

                curr_row.clear();

              }

            /* output to file (in_plane_rdf.log) */
            path in_plane_rdf_out_path = path(output_dir_path)
            /path("in_plane_rdf.log");
            in_plane_rdf_of.open(in_plane_rdf_out_path.native_file_string().c_str());

            if (!in_plane_rdf_of)
              {
                cout
                << "emd-analyzer: error creating output files (in_plane_rdf.log)!"
                << endl;
                return EXIT_FAILURE;
              }

            for (int i=0; i<rdf_divs; i++)
              {
                in_plane_rdf_of << ((i+0.5)*rdf_max_rad/rdf_divs) << "  "
                << gsl_vector_get(in_plane_rdf_mean, i) << "  "
                << gsl_vector_get(in_plane_rdf_std, i) << endl;
              }
            in_plane_rdf_of.close();

            //in_plane structure factor
            if (eval_DFT_3d_rdf)
              {
                path in_plane_SF_out_path = path(output_dir_path)
                /path("in_plane_SF.log");
                in_plane_SF_of.open(in_plane_SF_out_path.native_file_string().c_str());

                if (!in_plane_SF_of)
                  {
                    cout
                    << "emd-analyzer: error creating output files (in_plane_SF.log)!"
                    << endl;
                    return EXIT_FAILURE;
                  }
                gsl_complex comp;
                for (int k=1; k<rdf_divs/2; k++)
                  {
                    emd::uni::discrete_fourier_transform(in_plane_rdf_mean, k,
                        comp);
                    in_plane_SF_of << 2*M_PI*k/rdf_divs << "        "
                    << gsl_complex_abs(comp)<<"        "
                    << gsl_complex_arg(comp)<< endl;
                  }
                in_plane_SF_of.close();
              }

          }

        /*************************************************************************/
        if (eval_plane_ave)
          {

            /* number of bins */
            int N_bins = plane_ave_divs_1*plane_ave_divs_2;

            /* bin volume */
            double bin_vol = plane_ave_width* (plane_ave_y_upper
                - plane_ave_y_lower)* (plane_ave_z_upper - plane_ave_z_lower)
                /N_bins;

            for (int i=0; i<N_bins; i++)
              {

                if (bin_data_vec[i].N_samples==0)
                  continue;

                /* Q_tensor */
                gsl_matrix_scale(bin_data_vec[i].Q_tensor, 1.0
                    /(double)bin_data_vec[i].N_samples);

                /* density */
                bin_data_vec[i].density = ((double)bin_data_vec[i].N_samples)
                /(bin_vol*N_snapshots);

                /* the eigensystem of Q_tensor */
                gsl_matrix_memcpy(c_Q_tensor, bin_data_vec[i].Q_tensor);
                gsl_eigen_symmv(c_Q_tensor, Q_eigen_val, Q_eigen_vec, eigen_WS);
                int max_eig_idx = gsl_vector_max_index(Q_eigen_val);

                /* P_2 */
                bin_data_vec[i].P_2 = gsl_vector_get(Q_eigen_val, max_eig_idx);

                /* director */
                gsl_vector_set(bin_data_vec[i].director, 0, gsl_matrix_get(
                    Q_eigen_vec, 0, max_eig_idx));
                gsl_vector_set(bin_data_vec[i].director, 1, gsl_matrix_get(
                    Q_eigen_vec, 1, max_eig_idx));
                gsl_vector_set(bin_data_vec[i].director, 2, gsl_matrix_get(
                    Q_eigen_vec, 2, max_eig_idx));

              }

            /* output to povray file (plane_ave.pov) */
            path plane_ave_pov_path = path(output_dir_path)
            /path("plane_ave.pov");
            plane_ave_pov_of.open(plane_ave_pov_path.native_file_string().c_str());

            /* header */
            plane_ave_pov_of << endl << "#include \"colors.inc\"" << endl
            << "#include \"default_view.inc\"" << endl << endl;

            for (int i=0; i<N_bins; i++)
              {

                if (bin_data_vec[i].N_samples==0)
                  continue;

                double bin_x, bin_y, bin_z;
                double bin_y_width = (plane_ave_y_upper - plane_ave_y_lower)
                /((double)plane_ave_divs_1);
                double bin_z_width = (plane_ave_z_upper - plane_ave_z_lower)
                /((double)plane_ave_divs_2);
                int idx_z = i/plane_ave_divs_1;
                int idx_y = i - idx_z*plane_ave_divs_1;

                double p_1_x, p_1_y, p_1_z;
                double p_2_x, p_2_y, p_2_z;
                double p_3_x, p_3_y, p_3_z;
                double p_4_x, p_4_y, p_4_z;

                /* find the bin center */
                switch (plane_ave_normal)
                  {
                case AXIS_X:
                  {
                    bin_x = plane_ave_center;
                    bin_y = bin_y_width*(idx_y + 0.5) + plane_ave_y_lower;
                    bin_z = bin_z_width*(idx_z + 0.5) + plane_ave_z_lower;

                    p_1_x = bin_x;
                    p_1_y = bin_y - 0.5*bin_y_width;
                    p_1_z = bin_z - 0.5*bin_z_width;

                    p_2_x = bin_x;
                    p_2_y = bin_y - 0.5*bin_y_width;
                    p_2_z = bin_z + 0.5*bin_z_width;

                    p_3_x = bin_x;
                    p_3_y = bin_y + 0.5*bin_y_width;
                    p_3_z = bin_z + 0.5*bin_z_width;

                    p_4_x = bin_x;
                    p_4_y = bin_y + 0.5*bin_y_width;
                    p_4_z = bin_z - 0.5*bin_z_width;

                    break;
                  }
                case AXIS_Y:
                  {
                    bin_y = plane_ave_center;
                    bin_z = bin_y_width*(idx_y + 0.5) + plane_ave_y_lower;
                    bin_x = bin_z_width*(idx_z + 0.5) + plane_ave_z_lower;

                    p_1_y = bin_y;
                    p_1_z = bin_z - 0.5*bin_y_width;
                    p_1_x = bin_x - 0.5*bin_z_width;

                    p_2_y = bin_y;
                    p_2_z = bin_z - 0.5*bin_y_width;
                    p_2_x = bin_x + 0.5*bin_z_width;

                    p_3_y = bin_y;
                    p_3_z = bin_z + 0.5*bin_y_width;
                    p_3_x = bin_x + 0.5*bin_z_width;

                    p_4_y = bin_y;
                    p_4_z = bin_z + 0.5*bin_y_width;
                    p_4_x = bin_x - 0.5*bin_z_width;

                    break;
                  }
                case AXIS_Z:
                  {
                    bin_z = plane_ave_center;
                    bin_x = bin_y_width*(idx_y + 0.5) + plane_ave_y_lower;
                    bin_y = bin_z_width*(idx_z + 0.5) + plane_ave_z_lower;

                    p_1_z = bin_z;
                    p_1_x = bin_x - 0.5*bin_y_width;
                    p_1_y = bin_y - 0.5*bin_z_width;

                    p_2_z = bin_z;
                    p_2_x = bin_x - 0.5*bin_y_width;
                    p_2_y = bin_y + 0.5*bin_z_width;

                    p_3_z = bin_z;
                    p_3_x = bin_x + 0.5*bin_y_width;
                    p_3_y = bin_y + 0.5*bin_z_width;

                    p_4_z = bin_z;
                    p_4_x = bin_x + 0.5*bin_y_width;
                    p_4_y = bin_y - 0.5*bin_z_width;

                    break;
                  }
                  }

                double dx = plane_ave_dir_max_len* gsl_vector_get(
                    bin_data_vec[i].director, 0);
                double dy = plane_ave_dir_max_len* gsl_vector_get(
                    bin_data_vec[i].director, 1);
                double dz = plane_ave_dir_max_len* gsl_vector_get(
                    bin_data_vec[i].director, 2);

                double x_0 = bin_x - 0.5*dx;
                double y_0 = bin_y - 0.5*dy;
                double z_0 = bin_z - 0.5*dz;
                double x_1 = bin_x + 0.5*dx;
                double y_1 = bin_y + 0.5*dy;
                double z_1 = bin_z + 0.5*dz;

                double trans = bin_data_vec[i].P_2;

                plane_ave_pov_of << "union {\n" << "  cylinder {\n" << "    <"
                << x_0 << ", " << y_0 << ", " << z_0 << ">,\n" << "    <"
                << x_1 << ", " << y_1 << ", " << z_1 << ">,\n" << "    "
                << plane_ave_dir_rad << endl << "  }\n"
                << "  pigment { color rgb <" << 0 << ", " << 0 << ", " << 0
                << "> }\n" << "}\n\n";

                plane_ave_pov_of << "union {\n" << "  polygon {\n"
                << "    5,\n" << "    <" << p_1_x << ", " << p_1_y << ", "
                << p_1_z <<">,\n" << "    <" << p_2_x << ", " << p_2_y
                << ", " << p_2_z <<">,\n" << "    <" << p_3_x << ", "
                << p_3_y << ", " << p_3_z <<">,\n" << "    <" << p_4_x
                << ", " << p_4_y << ", " << p_4_z <<">,\n" << "    <"
                << p_1_x << ", " << p_1_y << ", " << p_1_z <<">\n"
                << "  }\n" << "  pigment { color rgb <" << trans << ", "
                << trans << ", " << trans << "> transmit 0.1}\n" << "}\n\n";

              }

            plane_ave_pov_of.close();

          }

        cout << "done!" << endl;

        cout << endl;

        /* free the heap allocations */

      }

    return EXIT_SUCCESS;

  }



/***************************************************************************/
void eval_director(const int Class_id)
  {
    cout << "  Particle Number : "<<Class_id<<endl;
    cout << "  * evaluating the director..." << endl;

    /* find a list of active bodies */
    vector<unsigned int> active_list;
    active_list.reserve(emd::uni::rigid_bodies_count);
    for (int m=0; m<emd::uni::rigid_bodies_count; m++)
      if (emd::uni::rigid_bodies[m].status
          == emd::rigid_body::EMD_RIGIDBODY_STATUS_ACTIVE)
        {
          //changed:

          if (emd::uni::rigid_bodies[m].class_id != Class_id)
            continue;

          active_list.push_back(m);
        }

    /* locate the director */
    emd::uni::eval_Q_tensor(active_list, axis_idx, Q_tensor, Q_eigen_val,
        Q_eigen_vec);
    int max_eig_idx = gsl_vector_max_index(Q_eigen_val);
    gsl_vector_set(director, 0, gsl_matrix_get(Q_eigen_vec, 0, max_eig_idx));
    gsl_vector_set(director, 1, gsl_matrix_get(Q_eigen_vec, 1, max_eig_idx));
    gsl_vector_set(director, 2, gsl_matrix_get(Q_eigen_vec, 2, max_eig_idx));

    cout << "  * order param: " << gsl_vector_get(Q_eigen_val, max_eig_idx)
    << endl;
    cout << "  * director: (" << gsl_vector_get(director, 0) << ", "
    << gsl_vector_get(director, 1) << ", " << gsl_vector_get(director, 2)
    << ")" << endl;

  }

/***************************************************************************/

void eval_plane()
  {

    cout << "  * detecting the planes... " << endl;
    emd::uni::detect_planes(class_id, min_wl, k_x, k_y, k_z, phi, director,
        penalty_coeff);
    cout << "  * (k, phi) = (" << k_x << ", " << k_y << ", " << k_z << ", "
    << phi << ")" << endl;

  }

/***************************************************************************/

void eval_stats_vec_double(std::vector<double> & vec, double & mean,
    double & std)
  {

    mean = 0;
    std = 0;

    /* mean */
    for (vector<double>::iterator it = vec.begin(); it != vec.end(); it++)
      mean += *it;

    mean /= vec.size();

    /* std */
    for (vector<double>::iterator it = vec.begin(); it != vec.end(); it++)
      std += (*it - mean)*(*it - mean);

    std /= vec.size();

    std = sqrt(std);

  }

/*** log parser actions ************************************************/

int parse_log_legend(const string & param_string)
  {
    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size()<1)
      return EXIT_FAILURE;

    log_fields.clear();

    for (int i=0; i<tokenized_buffer.size(); i++)
      if (tokenized_buffer[i]=="time")
        log_fields.push_back(FIELD_TIME);
      else if (tokenized_buffer[i]=="potential")
        log_fields.push_back(FIELD_POTENTIAL);
      else if (tokenized_buffer[i]=="energy")
        log_fields.push_back(FIELD_ENERGY);
      else if (tokenized_buffer[i]=="temperature")
        log_fields.push_back(FIELD_TEMPERATURE);
      else if (tokenized_buffer[i]=="pressure")
        log_fields.push_back(FIELD_PRESSURE);
      else if (tokenized_buffer[i]=="order_param")
        log_fields.push_back(FIELD_ORDER_PARAM);
      else if (tokenized_buffer[i]=="hamiltonian")
        log_fields.push_back(FIELD_HAMILTONIAN);
      else
        return EXIT_FAILURE;

    cout << "emd-analyzer: log file fields: " << endl << "  ";

    for (int i=0; i<log_fields.size(); i++)
      switch (log_fields[i])
        {
      case FIELD_TIME:
        cout << "TIME ";
        break;
      case FIELD_POTENTIAL:
        cout << "POTENTIAL ";
        break;
      case FIELD_ENERGY:
        cout << "ENERGY ";
        break;
      case FIELD_TEMPERATURE:
        cout << "TEMPERATURE ";
        break;
      case FIELD_PRESSURE:
        cout << "PRESSURE ";
        break;
      case FIELD_ORDER_PARAM:
        cout << "ORDER_PARAM ";
        break;
      case FIELD_HAMILTONIAN:
        cout << "HAMILTONIAN ";
        break;
        }

    cout << endl;

    return EXIT_SUCCESS;
  }

int parse_log_snap(const string & param_string)
  {
    vector<string> tokenized_buffer;

    /* tokenize the buffer */
    string_tokenize(param_string, tokenized_buffer);

    /* a certain number of parameters are required! */
    if (tokenized_buffer.size() != log_fields.size())
      return EXIT_FAILURE;

    for (int i=0; i<tokenized_buffer.size(); i++)
      switch (log_fields[i])
        {
      case FIELD_TIME:
        time_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_POTENTIAL:
        potential_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_ENERGY:
        energy_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_TEMPERATURE:
        temperature_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_PRESSURE:
        pressure_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_ORDER_PARAM:
        order_param_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
      case FIELD_HAMILTONIAN:
        hamiltonian_vec.push_back(atof(tokenized_buffer[i].c_str()));
        break;
        }

    return EXIT_SUCCESS;
  }

/***********************************************************************/

bin_data::bin_data()
  {

    N_samples = 0;
    density = 0;
    Q_tensor = gsl_matrix_calloc(3, 3);
    director = gsl_vector_calloc(3);

  }

bin_data::~bin_data()
  {

    gsl_matrix_free(Q_tensor);
    gsl_vector_free(director);

  }

/***************************************************************************/
void get_path_name(char *inp_argv[] ,int inp_argc, int &ctr ,string &path_name )
  {
    path_name = inp_argv[ctr];

    while((ctr+1 < inp_argc) && string(inp_argv[ctr+1]).find("-")!=0 )
      {      
        ctr++;
        path_name = path_name + " " + string(inp_argv[ctr]);  
      }
  }
/***************************************************************************/

