/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef EMD_ANALYZER_H
#define EMD_ANALYZER_H

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <sstream>
#include <cmath>


#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include "consts.h"
#include "stringext.h"
#include "gslext.h"
#include "uniparser.h"
#include "rigid_body.h"
#include "moluniverse.h"
#include "universesaver.h"

#define DEF_WIDTH    30


/***********************************************************************/
/* container class for averages in a bin */
class bin_data
{
public:

  long int       N_samples;
  double         density;
  double         P_2;
  gsl_matrix *   Q_tensor;
  gsl_vector *   director;

  bin_data();
  ~bin_data();

};

/***********************************************************************/

enum log_field_type
{
  FIELD_TIME,
  FIELD_POTENTIAL,
  FIELD_ENERGY,
  FIELD_TEMPERATURE,
  FIELD_PRESSURE,
  FIELD_ORDER_PARAM,
  FIELD_HAMILTONIAN
};

static const int       AXIS_X               = 0;
static const int       AXIS_Y               = 1;
static const int       AXIS_Z               = 2;

static std::string     log_file_path;                         /* must be supplied */
static bool            log_file_path_spec         = false;
static std::string     uni_snapshot_dir_path;                 /* must be supplied */
static bool            uni_snapshot_dir_path_spec = false;
static std::string     output_dir_path;                       /* must be supplied */
static bool            output_dir_path_spec       = false;
static std::string     uni_file_path;                         /* must be supplied */
static bool            uni_file_path_spec         = false;
static bool            eval_density_fourier       = false;
static bool            eval_structure_factor      = false;      
static bool            eval_order_param           = false;
static bool            eval_rel_dir               = false;

static bool            eval_3d_rdf          =  false;
static bool            eval_2d_rdf          =  false;
static bool            eval_smectic_config  =  false;
static bool            eval_bond_order      =  false;
static bool            eval_tilt            =  false;
static bool            eval_plane_ave       =  false;
        
static bool            eval_DFT_3d_rdf      = false;
static bool            eval_DFT_2d_rdf      = false;

static bool            eval_sf              = false;
static bool            director_is_eval;
static bool            plane_is_eval;
static bool            count_regions        = false;
/* configuration */
static double          rdf_max_rad          =  8.0;
static int             rdf_divs             =  200;
static int             freq_divs             =  20;
static int             axis_idx             =  2;
static double          min_wl               =  1.0;
static int             in_plane_order       =  0;
static double          penalty_coeff        =  1.0;
static int             snap_skip_rate       =  0;
static double          bond_order_r_min     =  2.0;
static double          bond_order_r_max     =  3.0;
static int 	       class_id             =  1;
static int             class_id1            = 1;
static int             class_id2            = 2;
     
static double          neighbor_dist         =  3;
static int             plane_ave_normal;             /* NOTE must be supplied by user */
static double          plane_ave_center;             /* NOTE must be supplied by user */
static double          plane_ave_width;              /* NOTE must be supplied by user */
static int             plane_ave_divs_1;             /* NOTE must be supplied by user */
static int             plane_ave_divs_2;             /* NOTE must be supplied by user */

static double          plane_ave_y_lower;
static double          plane_ave_y_upper;
static double          plane_ave_z_lower;
static double          plane_ave_z_upper;

static double          plane_ave_dir_max_len  = 1.0;
static double          plane_ave_dir_rad      = 0.05;
static double           concentration=0;

static unsigned int    N_snapshots          =  0;

static std::vector<log_field_type>
                       log_fields;

static std::vector<double>  time_vec,
                       potential_vec,
                       energy_vec,
                       temperature_vec,
                       pressure_vec,
                       order_param_vec,
                       hamiltonian_vec;

gsl_eigen_symmv_workspace *  eigen_WS       = gsl_eigen_symmv_alloc(3);
static gsl_vector *          Q_eigen_val    = gsl_vector_alloc(3);
static gsl_matrix *          Q_tensor       = gsl_matrix_alloc(3, 3);
static gsl_matrix *          c_Q_tensor     = gsl_matrix_alloc(3, 3);
static gsl_matrix *          Q_eigen_vec    = gsl_matrix_alloc(3, 3);
static gsl_vector *          director       = gsl_vector_alloc(3);
static double                k_x, k_y, k_z, phi;

static std::vector<gsl_vector*>   scalar_rdfs;
static std::vector<gsl_vector*>   para_rdfs;
static std::vector<gsl_vector*>   perp_rdfs;
static std::vector<gsl_vector*>   in_plane_rdfs;
static std::vector<double>        tilt_angles;
static std::vector<double>        bond_orders;
bin_data*                         bin_data_vec;
static std::vector<double>        order_params;

static gsl_vector * scalar_rdf_mean;
static gsl_vector * scalar_rdf_std;
static gsl_vector * para_rdf_mean;
static gsl_vector * para_rdf_std;
static gsl_vector * perp_rdf_mean;
static gsl_vector * perp_rdf_std;
static gsl_vector * in_plane_rdf_mean;
static gsl_vector * in_plane_rdf_std;

std::ofstream log_stat_of;
std::ofstream scalar_rdf_of;
std::ofstream para_rdf_of;
std::ofstream perp_rdf_of;
std::ofstream in_plane_rdf_of;
std::ofstream plane_ave_pov_of;
std::ofstream scalar_SF_of;
std::ofstream para_SF_of;
std::ofstream perp_SF_of;
std::ofstream in_plane_SF_of;
std::ofstream dens_four_ouf;
std::ofstream param_ouf;

static int parse_log_legend(const std::string & input_string);
static int parse_log_snap(const std::string & input_string);

void eval_stats_vec_double(std::vector<double> & vec, double & mean, double & std);

void eval_director(const int Class_id);
void eval_plane();
void get_path_name(char *inp_argv[] ,int inp_argc, int &ctr ,std::string &path_name );

/* actions */
typedef int (*act)(const std::string & input_string);

static std::string LOG_KEYWORDS[] =
{
  "legend",
  "snap"
};

static act LOG_ACTIONS[] =
{
  &parse_log_legend,
  &parse_log_snap
};

/***********************************************************************/

#endif /* EMD_ANALYZER_H */
