/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <iomanip>
#include <cmath>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_math.h>

#include <quaternion/gsl_quaternion.h>

#ifndef GSLEXT_H
#define GSLEXT_H

/*********************************************
 * matrix and vector related extra functions *
 *********************************************/

double gsl_3matrix_det(const gsl_matrix * A);

int gsl_3matrix_invert(const gsl_matrix * A, gsl_matrix * Ainv);

int gsl_matrix_euler_rot(const double & theta, const double & phi,
  const double & psi, gsl_matrix * dest);

/* SO(3) rot. generator with respect to x */
int gsl_matrix_rot_x(gsl_matrix * dest);

/* SO(3) rot. generator with respect to y */
int gsl_matrix_rot_y(gsl_matrix * dest);

/* SO(3) rot. generator with respect to z */
int gsl_matrix_rot_z(gsl_matrix * dest);

/* normalize a vector */
int gsl_vector_normalize(gsl_vector * vec);

/* rotate a vector about x axis */
int gsl_vector_rot_x(const double & theta, const gsl_vector * src, gsl_vector * dest);

/* rotate a vector about x axis */
int gsl_vector_rot_y(const double & theta, const gsl_vector * src, gsl_vector * dest);

/* rotate a vector about x axis */
int gsl_vector_rot_z(const double & theta, const gsl_vector * src, gsl_vector * dest);

/* print a gsl_matrix */
std::ostream & operator<<(std::ostream & os, const gsl_matrix & A);

/* print a gsl_vector */
std::ostream & operator<<(std::ostream & os, const gsl_vector & v);

/***************
 * quaternions *
 ***************/

/* print a gsl_quaternion */
std::ostream & operator<<(std::ostream & os, const gsl_quaternion & q);

#endif /* GSLEXT_H */
