/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <int_dipole_comp.h>

using namespace std;

/***************************************************************************/

double   emd::inter::dipole::U;

gsl_vector *       emd::inter::dipole::e_x       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::dipole::e_y       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::dipole::e_z       = gsl_vector_calloc(3);

gsl_matrix *       emd::inter::dipole::R_x       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::dipole::R_y       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::dipole::R_z       = gsl_matrix_calloc(3, 3);

/***************************************************************************/

void
emd::inter::dipole::init ()
{

  gsl_vector_set_basis(e_x, 0);
  gsl_vector_set_basis(e_y, 1);
  gsl_vector_set_basis(e_z, 2);

  gsl_matrix_rot_x(R_x);
  gsl_matrix_rot_y(R_y);
  gsl_matrix_rot_z(R_z);

  init_flag = true;

}

/***************************************************************************/

bool
emd::inter::dipole::is_init ()
{
  return init_flag;
}

/***************************************************************************/

int
emd::inter::dipole::free_memory ()
{

  gsl_vector_free(e_x);
  gsl_vector_free(e_y);
  gsl_vector_free(e_z);

  gsl_matrix_free(R_x);
  gsl_matrix_free(R_y);
  gsl_matrix_free(R_z);

}

/***************************************************************************/

/* interaction energy */
double
emd::inter::dipole::get_dipole_int_energy (const gsl_vector * p0, const gsl_vector * E0)
{

  gsl_vector_memcpy(p_lab, p0);
  gsl_vector_memcpy(E, E0);

  gsl_blas_ddot(E0, p0, &U);

  U *= -1;

  return U;

}

/***************************************************************************/

/* exerted torque with respect to CM */
double
emd::inter::dipole::get_dipole_int_torque (const gsl_matrix * gen)
{

  double torque;

  gsl_blas_dgemv(CblasNoTrans, 1, gen, p_lab, 0, dvec);
  gsl_blas_ddot(E, dvec, &torque);

  return torque;

}

/***************************************************************************/
