/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef INT_DIPOLE_COMP_H
#define INT_DIPOLE_COMP_H

#include <iostream>
#include <cmath>

#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include "gslext.h"
#include "paramdb.h"
#include "rigid_body.h"

namespace emd
{

namespace inter
{

namespace dipole
{

  static bool               init_flag = false;

  static gsl_vector *       E       = gsl_vector_alloc(3);
  static gsl_vector *       p_lab   = gsl_vector_alloc(3);
  extern double             U;

  static gsl_vector *       dvec    = gsl_vector_alloc(3);


  /* bases and generators: that would be handy to have them around */

  extern gsl_vector *       e_x;
  extern gsl_vector *       e_y;
  extern gsl_vector *       e_z;

  extern gsl_matrix *       R_x;
  extern gsl_matrix *       R_y;
  extern gsl_matrix *       R_z;

  /***************************************************************************/

  void init();

  bool is_init();

  int free_memory();

  /**** NOTE ********************************************************************************
   *                                                                                        *
   * In all cases, the transpose of the provided rotation matrix @A will be used!           *
   * (The orientation operators of the rigid bodies transform -a vector- (in the lab frame) *
   * to the representation of a same vector in the body frame; however, the expressions     *
   * given in the paper transform -the lab frame- to the body frame, hence transforming     *
   * a vector in the body frame to the lab frame)                                           *
   *                                                                                        *
   ******************************************************************************************/

  /*** energy/force/torque related *************/

  /* interaction energy */
  double
  get_dipole_int_energy (const gsl_vector * p, const gsl_vector * E);

  /* exerted torque with respect to CM */
  double
  get_dipole_int_torque (const gsl_matrix * gen);

}; /* dipole */

}; /* inter */

}; /* emd */

/***********************************************/

#endif /* INT_DIPOLE_COMP_H */
