/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "int_dipole_uni.h"

using namespace std;

/***************************************************************************/

double
emd::uni::dipole::eval_dipole_single(const int& m, const bool& dyn)
{

  /* init dipole workspace */
  if ( !emd::inter::dipole::is_init() ) emd::inter::dipole::init();

  gsl_vector * p_lab      = gsl_vector_alloc(3);
  gsl_vector * tmp_torque = gsl_vector_alloc(3);

  /* find dipole at lab */
  gsl_blas_dgemv(CblasNoTrans, 1, rigid_bodies[m].R, rigid_bodies[m].dipole_cm, 0, p_lab);

  /* find the energy */
  double curr_U = emd::inter::dipole::get_dipole_int_energy(p_lab, emd::uni::E_field);
  rigid_bodies[m].potential += curr_U;

  if (dyn)
  {

    /* torque */
    gsl_vector_set(tmp_torque, 0, emd::inter::dipole::get_dipole_int_torque(emd::inter::dipole::R_x));
    gsl_vector_set(tmp_torque, 1, emd::inter::dipole::get_dipole_int_torque(emd::inter::dipole::R_y));
    gsl_vector_set(tmp_torque, 2, emd::inter::dipole::get_dipole_int_torque(emd::inter::dipole::R_z));

    /* adding up the current torque contribution */
    gsl_vector_add(rigid_bodies[m].torque, tmp_torque);

  }

  gsl_vector_free(p_lab);
  gsl_vector_free(tmp_torque);

  return curr_U;

}

/*******************************************************************************************/

double
emd::uni::dipole::eval_dipole(const bool& dyn)
{

  double    old_potential = potential;

  for (int m=0; m<rigid_bodies_count; m++)
    if ( (rigid_bodies[m].int_mask & MASK_DIPOLE_INT) == MASK_DIPOLE_INT )
      potential += eval_dipole_single(m, dyn);


  return (potential - old_potential);

}

/*******************************************************************************************/

int
emd::uni::dipole::free_memory()
{

  return EXIT_SUCCESS;

}
