/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * todo:                                                                   *
 * - max_offset automatic evaluations                                      *
 *                                                                         *
 ***************************************************************************/

#ifndef INT_GB_UNI_H
#define INT_GB_UNI_H

#include <iostream>
#include <iomanip>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "paramdb.h"
#include "neighbor_data.h"
#include "moluniverse.h"

/****************************************/

#include "int_gb_comp.h"

/****************************************/

#define CHECK_OVERFLOW_GB_INT
#define OVERFLOW_ENERGY_GB_INT      1000

/****************************************/

namespace emd
{

namespace uni
{

namespace gb
{

  /************************************************************************************/

  typedef    std::vector<emd::uni::neighbor_data>::iterator   neighbor_it;

  /************************************************************************************/

  extern    std::vector<neighbor_data> *    n_list_gb;       /* neighbor list for gb */

  extern    bool      nl_init_flag;   /* init flag of neighbor list */
  extern    double    r_c;            /* cutoff distance */
  extern    double    r_s;            /* skin distance */
  extern    double    max_offset;     /* "thickness* */

  /************************************************************************************
   * desc: evaluate GB non-bonded interactions with respect to the current neighbor   *
   *        list                                                                      *
   *        - first derivatives will also be evaluated if @dyn is true                *
   *        - the previous interaction evals will be discarded if @clear is true      *
   *                                                                                  *
   *        gb_ws must be initialized (init())                                        *
   ************************************************************************************/
  double eval_GB(const bool& dyn, int& stat);

  double eval_GB_single(const int& m1, const bool& exclusive,
    const bool& dyn, int& stat);

  /******* energy/force/nl related utility functions **********************************/

  /* malloc for the neighbor list */
  int init_neighbor_list();

  /* update the neighbor list (with respect to the specified algorthim in emd::uni) */
  void update_neighbor_list();

  /* prints out the neighbor list */
  void print_neighbor_list();

  /* the number of interactions */
  unsigned long get_int_count();

  /* update "the verlet-type" neighbor list */
  void update_neighbor_list_verlet();
  void update_neighbor_list_verlet_using_cell_info();

  /* DEPRECATED */
  /************************************************************************************
   * this function must be called after a call to eval_GB_single in order to         *
   *  sync the neighbor data enegies.                                                 *
   ************************************************************************************/
  void sync_neighbor_data_energies(const int& idx);

  /******* force/torque workspace *****************************************************/

  static gsl_vector *    tmp_r12     =  gsl_vector_alloc(3);
  static gsl_vector *    tmp_force   =  gsl_vector_alloc(3);
  static gsl_vector *    tmp_torque  =  gsl_vector_alloc(3);
  static gsl_vector *    tmp_v       =  gsl_vector_alloc(3);
  static gsl_vector *    img_vec     =  gsl_vector_alloc(3);

  /************************************************************************************/

  int free_memory();

}; /* gb */

}; /* uni */

}; /* emd */

#endif /* MOLUNIVERSE_H */
