/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <paramdb.h>

using namespace std;

/**** memory allocation for extern vars ************************************/

emd::paramdb::param_cache_rigid_body *        emd::paramdb::rigid_body_param;
emd::paramdb::param_cache_int_surf *          emd::paramdb::int_surf_param;
emd::paramdb::param_cache_RE2_int **          emd::paramdb::re2_int_param;
emd::paramdb::param_cache_GB_int **           emd::paramdb::gb_int_param;
emd::paramdb::param_cache_EES_int **          emd::paramdb::ees_int_param;
std::string                                   emd::paramdb::param_file_name;

/***************************************************************************/

emd::paramdb::param_cache_rigid_body::param_cache_rigid_body ()
{

  I_body     = gsl_matrix_calloc(3, 3);
  I_body_inv = gsl_matrix_calloc(3, 3);

#ifdef PARAMCACHE_RIGIDBODY_DEBUG
  std::cout << "* emd::paramdb::param_cache_rigid_body: constructed (" << this << ")." << std::endl;
#endif

}

emd::paramdb::param_cache_rigid_body::~param_cache_rigid_body ()
{

  gsl_matrix_free(I_body);
  gsl_matrix_free(I_body_inv);

#ifdef PARAMCACHE_RIGIDBODY_DEBUG
  std::cout << "* emd::paramdb::param_cache_rigid_body: destructed (" << this << ")." << std::endl;
#endif

}

/***************************************************************************/

emd::paramdb::param_cache_RE2_int::param_cache_RE2_int ()
{

  S1     = gsl_matrix_calloc(3, 3);
  S1p2   = gsl_matrix_calloc(3, 3);
  S1pm2  = gsl_matrix_calloc(3, 3);
  E1     = gsl_matrix_calloc(3, 3);
  E1p2   = gsl_matrix_calloc(3, 3);

  S2     = gsl_matrix_calloc(3, 3);
  S2p2   = gsl_matrix_calloc(3, 3);
  S2pm2  = gsl_matrix_calloc(3, 3);
  E2     = gsl_matrix_calloc(3, 3);
  E2p2   = gsl_matrix_calloc(3, 3);

#ifdef PARAMCACHE_RE2_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_RE2_int: constructed (" << this << ")." << std::endl;
#endif

}

emd::paramdb::param_cache_RE2_int::~param_cache_RE2_int ()
{

  gsl_matrix_free(S1);
  gsl_matrix_free(S1p2);
  gsl_matrix_free(S1pm2);
  gsl_matrix_free(E1);
  gsl_matrix_free(E1p2);

  gsl_matrix_free(S2);
  gsl_matrix_free(S2p2);
  gsl_matrix_free(S2pm2);
  gsl_matrix_free(E2);
  gsl_matrix_free(E2p2);

#ifdef PARAMCACHE_RE2_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_RE2_int: destructed (" << this << ")." << std::endl;
#endif

}

/***************************************************************************/

emd::paramdb::param_cache_GB_int::param_cache_GB_int ()
{

  S1     = gsl_matrix_calloc(3, 3);
  S1p2   = gsl_matrix_calloc(3, 3);
  E1     = gsl_matrix_calloc(3, 3);
  E1p2   = gsl_matrix_calloc(3, 3);

  S2     = gsl_matrix_calloc(3, 3);
  S2p2   = gsl_matrix_calloc(3, 3);
  E2     = gsl_matrix_calloc(3, 3);
  E2p2   = gsl_matrix_calloc(3, 3);

#ifdef PARAMCACHE_GB_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_GB_int: constructed (" << this << ")." << std::endl;
#endif

}

emd::paramdb::param_cache_GB_int::~param_cache_GB_int ()
{

  gsl_matrix_free(S1);
  gsl_matrix_free(S1p2);
  gsl_matrix_free(E1);
  gsl_matrix_free(E1p2);

  gsl_matrix_free(S2);
  gsl_matrix_free(S2p2);
  gsl_matrix_free(E2);
  gsl_matrix_free(E2p2);

#ifdef PARAMCACHE_GB_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_GB_int: destructed (" << this << ")." << std::endl;
#endif

}

/***************************************************************************/

emd::paramdb::param_cache_EES_int::param_cache_EES_int ()
{

  S  =  gsl_matrix_calloc(3, 3);
  E  =  gsl_matrix_calloc(3, 3);

#ifdef PARAMCACHE_EES_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_EES_int: constructed (" << this << ")." << std::endl;
#endif

}

emd::paramdb::param_cache_EES_int::~param_cache_EES_int ()
{

  gsl_matrix_free(S);
  gsl_matrix_free(E);

#ifdef PARAMCACHE_EES_INT_DEBUG
  std::cout << "* emd::paramdb::param_cache_EES_int: destructed (" << this << ")." << std::endl;
#endif

}

/***************************************************************************/

/*
 * loads a parameter file
 *
 *   @param_file_name            parameter file
 *   @clear                      clear the current vectors
 *
 */
int
emd::paramdb::load_param_file (const string & param_file_name, const bool & clear)
{

  /* keywords */
  KEYWORDS[0] = "rigid_class";
  KEYWORDS[1] = "int_surf_class";
  KEYWORDS[2] = "re2_homo_int";
  KEYWORDS[3] = "re2_hetero_int";
  KEYWORDS[4] = "gb_homo_int";
  KEYWORDS[5] = "gb_hetero_int";
  KEYWORDS[6] = "ees_int";

  /* actions (sub parsers) */
  ACTIONS[0]  = &emd::paramdb::parse_param_rigid_body;
  ACTIONS[1]  = &emd::paramdb::parse_param_int_surf;
  ACTIONS[2]  = &emd::paramdb::parse_param_RE2_homo_int;
  ACTIONS[3]  = &emd::paramdb::parse_param_RE2_hetero_int;
  ACTIONS[4]  = &emd::paramdb::parse_param_GB_homo_int;
  ACTIONS[5]  = &emd::paramdb::parse_param_GB_hetero_int;
  ACTIONS[6]  = &emd::paramdb::parse_param_EES_int;

  int exit_flag;

  if (clear)
  {

    /* objects */
    vec_param_rigid_body.clear();
    vec_param_int_surf.clear();

    /* interaction parameters */
    vec_param_RE2_homo_int.clear();
    vec_param_RE2_hetero_int.clear();
    vec_param_GB_homo_int.clear();
    vec_param_GB_hetero_int.clear();
    vec_param_EES_int.clear();

  }

  /* opening the specified file as ro */
  ifstream inf(param_file_name.c_str(), ios::in);

  exit_flag = file_parser(inf, static_cast<int>(N_KEYWORDS), KEYWORDS, ACTIONS, '#', true);

  if (exit_flag==EXIT_FAILURE)
    return EXIT_FAILURE;

  /* post-processing */
  exit_flag = param_post_proc();

  if (exit_flag==EXIT_FAILURE)
    return EXIT_FAILURE;

  emd::paramdb::param_file_name = param_file_name;
  emd::paramdb::init = true;

  return EXIT_SUCCESS;
}

/* parsers */

/**********************************************************************************/

int
emd::paramdb::parse_param_rigid_body (const string & param_buffer)
{

  vector <string>              tokenized_buffer;
  param_rigid_body         tmp_param_rigid_body;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  if (tokenized_buffer.size()<6)
    return EXIT_FAILURE;

  tmp_param_rigid_body.class_id = atoi(tokenized_buffer[0].c_str());                 /* class_id */
  tmp_param_rigid_body.name     = tokenized_buffer[1];                               /* name */
  tmp_param_rigid_body.mass     = atof(tokenized_buffer[2].c_str())*MASS_FACTOR;     /* mass */
  tmp_param_rigid_body.I_body_x = atof(tokenized_buffer[3].c_str())*MASS_FACTOR;     /* I_body_x */
  tmp_param_rigid_body.I_body_y = atof(tokenized_buffer[4].c_str())*MASS_FACTOR;     /* I_body_y */
  tmp_param_rigid_body.I_body_z = atof(tokenized_buffer[5].c_str())*MASS_FACTOR;     /* I_body_z */

  tokenized_buffer.clear();

  vec_param_rigid_body.push_back(tmp_param_rigid_body);

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_rigid_body: adding a new RE2 body class:" << endl
       << "  class_id:    " << tmp_param_rigid_body.class_id << endl
       << "  name:        " << tmp_param_rigid_body.name << endl;
  cout << setprecision(8) << showpos << showpoint << left
       << "  mass:        " << setw(8) << tmp_param_rigid_body.mass
       << " (kcal.mol^-1.fs^2.ang^-2)" << endl
       << "  I_body_x:    " << setw(8) << tmp_param_rigid_body.I_body_x
       << " (kcal.mol^-1.fs^2)" << endl
       << "  I_body_y:    " << setw(8) << tmp_param_rigid_body.I_body_y
       << " (kcal.mol^-1.fs^2)" << endl
       << "  I_body_z:    " << setw(8) << tmp_param_rigid_body.I_body_z
       << " (kcal.mol^-1.fs^2)" << endl;

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_int_surf (const string & param_buffer)
{

  vector <string>              tokenized_buffer;
  param_int_surf               tmp_param_int_surf;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;

  tmp_param_int_surf.class_id  = atoi(tokenized_buffer[0].c_str());      /* class_id */

  tokenized_buffer.clear();

  vec_param_int_surf.push_back(tmp_param_int_surf);

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_int_surf: adding a new interacting surface class:" << endl
       << "  class_id:    " << tmp_param_int_surf.class_id << endl;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_RE2_homo_int (const string & param_buffer)
{

  vector <string>      tokenized_buffer;
  param_RE2_homo_int   tmp_param_RE2_homo_int;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  if (tokenized_buffer.size()<10)
    return EXIT_FAILURE;

  tmp_param_RE2_homo_int.class_id = atoi(tokenized_buffer[0].c_str());   /* class id */
  tmp_param_RE2_homo_int.h_type   = atoi(tokenized_buffer[1].c_str());   /* h formula */
  tmp_param_RE2_homo_int.hamaker  = atof(tokenized_buffer[2].c_str());   /* hamaker constant */
  tmp_param_RE2_homo_int.sigma_c  = atof(tokenized_buffer[3].c_str());   /* atomic sigma */
  tmp_param_RE2_homo_int.sig1     = atof(tokenized_buffer[4].c_str());   /* sigma_x */
  tmp_param_RE2_homo_int.sig2     = atof(tokenized_buffer[5].c_str());   /* sigma_y */
  tmp_param_RE2_homo_int.sig3     = atof(tokenized_buffer[6].c_str());   /* sigma_z */
  tmp_param_RE2_homo_int.e1       = atof(tokenized_buffer[7].c_str());   /* e_x */
  tmp_param_RE2_homo_int.e2       = atof(tokenized_buffer[8].c_str());   /* e_y */
  tmp_param_RE2_homo_int.e3       = atof(tokenized_buffer[9].c_str());   /* e_z */

  tokenized_buffer.clear();

  vec_param_RE2_homo_int.push_back(tmp_param_RE2_homo_int);

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_RE2_homo_int: adding a new RE2 homogeneous "
       << "interaction parameter set:" << endl
       << "  class_id:    " << tmp_param_RE2_homo_int.class_id << endl
       << "  h_type:      " << tmp_param_RE2_homo_int.h_type << endl;
  cout << setprecision(8) << showpos << showpoint << left
       << setw(8) << "  hamaker:     " << tmp_param_RE2_homo_int.hamaker << " (kcal.mol^-1)" << endl
       << setw(8) << "  sigma_c:     " << tmp_param_RE2_homo_int.sigma_c << " (ang)" << endl
       << setw(8) << "  sig1:        " << tmp_param_RE2_homo_int.sig1 << " (ang)" << endl
       << setw(8) << "  sig2:        " << tmp_param_RE2_homo_int.sig2 << " (ang)" << endl
       << setw(8) << "  sig3:        " << tmp_param_RE2_homo_int.sig3 << " (ang)" << endl
       << setw(8) << "  e1:          " << tmp_param_RE2_homo_int.e1 << endl
       << setw(8) << "  e2:          " << tmp_param_RE2_homo_int.e2 << endl
       << setw(8) << "  e3:          " << tmp_param_RE2_homo_int.e3 << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_RE2_hetero_int (const string & param_buffer)
{

  vector <string>       tokenized_buffer;
  param_RE2_hetero_int  tmp_param_RE2_hetero_int;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* at least, 5 parameters must be provided */
  if (tokenized_buffer.size()<5)
    return EXIT_FAILURE;

  tmp_param_RE2_hetero_int.class_id_1 = atoi(tokenized_buffer[0].c_str());   /* class id 1 */
  tmp_param_RE2_hetero_int.class_id_2 = atoi(tokenized_buffer[1].c_str());   /* class id 2 */
  tmp_param_RE2_hetero_int.h_type     = atoi(tokenized_buffer[2].c_str());   /* h type */
  tmp_param_RE2_hetero_int.hamaker    = atof(tokenized_buffer[3].c_str());   /* heterogeneous hamaker constant */
  tmp_param_RE2_hetero_int.sigma_c    = atof(tokenized_buffer[4].c_str());   /* heterogeneous atomic sigma */

  /********************************** optional parameters ****
   *  NO FAILURE at this point                               *
   *  homogenous parameters will be used if no heterogenous  *
   *  one is found                                           *
   ***********************************************************/

  if (tokenized_buffer.size()>=17)
  {
    tmp_param_RE2_hetero_int.spec = true;
    tmp_param_RE2_hetero_int.sig11 = atof(tokenized_buffer[5].c_str());   /* sigma_x (1) */
    tmp_param_RE2_hetero_int.sig12 = atof(tokenized_buffer[6].c_str());   /* sigma_y (1) */
    tmp_param_RE2_hetero_int.sig13 = atof(tokenized_buffer[7].c_str());   /* sigma_z (1) */
    tmp_param_RE2_hetero_int.e11   = atof(tokenized_buffer[8].c_str());   /* e_x (1) */
    tmp_param_RE2_hetero_int.e12   = atof(tokenized_buffer[9].c_str());   /* e_y (1) */
    tmp_param_RE2_hetero_int.e13   = atof(tokenized_buffer[10].c_str());   /* e_z (1) */
    tmp_param_RE2_hetero_int.sig21 = atof(tokenized_buffer[11].c_str());  /* sigma_x (2) */
    tmp_param_RE2_hetero_int.sig22 = atof(tokenized_buffer[12].c_str());  /* sigma_y (2) */
    tmp_param_RE2_hetero_int.sig23 = atof(tokenized_buffer[13].c_str());  /* sigma_z (2) */
    tmp_param_RE2_hetero_int.e21   = atof(tokenized_buffer[14].c_str());  /* e_x (2) */
    tmp_param_RE2_hetero_int.e22   = atof(tokenized_buffer[15].c_str());  /* e_y (2) */
    tmp_param_RE2_hetero_int.e23   = atof(tokenized_buffer[16].c_str());  /* e_z (2) */

  }
  else
    tmp_param_RE2_hetero_int.spec = false;

  vec_param_RE2_hetero_int.push_back(tmp_param_RE2_hetero_int);

  tokenized_buffer.clear();

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_RE2_hetero_int: adding a new RE2 heterogeneous interaction parameter set:"
       << endl
       << "  class_id_1:  " << tmp_param_RE2_hetero_int.class_id_1 << endl
       << "  class_id_2:  " << tmp_param_RE2_hetero_int.class_id_2 << endl
       << "  h_type:      " << tmp_param_RE2_hetero_int.h_type << endl;

  cout << setprecision(8) << showpos << showpoint << left
       << "  hamaker:     " << tmp_param_RE2_hetero_int.hamaker << " (kcal.mol^-1)" << endl
       << "  sigma_c:     " << tmp_param_RE2_hetero_int.sigma_c << " (ang)" << endl;

  if (tmp_param_RE2_hetero_int.spec)
  {
    cout << setw(8) << "  Specific parameters:" << endl
         << setw(8) << "  - sig11:       " << tmp_param_RE2_hetero_int.sig11 << " (ang)" << endl
         << setw(8) << "  - sig12:       " << tmp_param_RE2_hetero_int.sig12 << " (ang)" << endl
         << setw(8) << "  - sig13:       " << tmp_param_RE2_hetero_int.sig13 << " (ang)" << endl
         << setw(8) << "  - e11:         " << tmp_param_RE2_hetero_int.e11 << endl
         << setw(8) << "  - e12:         " << tmp_param_RE2_hetero_int.e12 << endl
         << setw(8) << "  - e13:         " << tmp_param_RE2_hetero_int.e13 << endl
         << setw(8) << "  - sig21:       " << tmp_param_RE2_hetero_int.sig21 << " (ang)" << endl
         << setw(8) << "  - sig22:       " << tmp_param_RE2_hetero_int.sig22 << " (ang)" << endl
         << setw(8) << "  - sig23:       " << tmp_param_RE2_hetero_int.sig23 << " (ang)" << endl
         << setw(8) << "  - e21:         " << tmp_param_RE2_hetero_int.e21 << endl
         << setw(8) << "  - e22:         " << tmp_param_RE2_hetero_int.e22 << endl
         << setw(8) << "  - e23:         " << tmp_param_RE2_hetero_int.e23 << endl;
  }

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_GB_homo_int (const string & param_buffer)
{

  vector <string>      tokenized_buffer;
  param_GB_homo_int    tmp_param_GB_homo_int;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  if (tokenized_buffer.size()<11)
    return EXIT_FAILURE;

  tmp_param_GB_homo_int.class_id = atoi(tokenized_buffer[0].c_str());   /* class id */
  tmp_param_GB_homo_int.eps      = atof(tokenized_buffer[1].c_str());   /* energy */
  tmp_param_GB_homo_int.sig      = atof(tokenized_buffer[2].c_str());   /* sigma */
  tmp_param_GB_homo_int.mu       = atoi(tokenized_buffer[3].c_str());   /* mu */
  tmp_param_GB_homo_int.nu       = atoi(tokenized_buffer[4].c_str());   /* nu */
  tmp_param_GB_homo_int.sig1     = atof(tokenized_buffer[5].c_str());   /* sigma_x */
  tmp_param_GB_homo_int.sig2     = atof(tokenized_buffer[6].c_str());   /* sigma_y */
  tmp_param_GB_homo_int.sig3     = atof(tokenized_buffer[7].c_str());   /* sigma_z */
  tmp_param_GB_homo_int.e1       = atof(tokenized_buffer[8].c_str());   /* e_x */
  tmp_param_GB_homo_int.e2       = atof(tokenized_buffer[9].c_str());   /* e_y */
  tmp_param_GB_homo_int.e3       = atof(tokenized_buffer[10].c_str());  /* e_z */

  tokenized_buffer.clear();

  vec_param_GB_homo_int.push_back(tmp_param_GB_homo_int);

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_GB_homo_int: adding a new GB homogeneous "
       << "interaction parameter set:" << endl
       << "  class_id:    " << tmp_param_GB_homo_int.class_id << endl;
  cout << setprecision(8) << showpos << showpoint << left
       << setw(8) << "  eps:         " << tmp_param_GB_homo_int.eps << endl
       << setw(8) << "  sig:         " << tmp_param_GB_homo_int.sig << endl
       << setw(8) << "  mu:          " << tmp_param_GB_homo_int.mu << endl
       << setw(8) << "  nu:          " << tmp_param_GB_homo_int.nu << endl
       << setw(8) << "  sig1:        " << tmp_param_GB_homo_int.sig1 << endl
       << setw(8) << "  sig2:        " << tmp_param_GB_homo_int.sig2 << endl
       << setw(8) << "  sig3:        " << tmp_param_GB_homo_int.sig3 << endl
       << setw(8) << "  e1:          " << tmp_param_GB_homo_int.e1 << endl
       << setw(8) << "  e2:          " << tmp_param_GB_homo_int.e2 << endl
       << setw(8) << "  e3:          " << tmp_param_GB_homo_int.e3 << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_GB_hetero_int (const string & param_buffer)
{

  vector <string>       tokenized_buffer;
  param_GB_hetero_int   tmp_param_GB_hetero_int;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* at least, 6 parameters must be provided */
  if (tokenized_buffer.size()<7)
    return EXIT_FAILURE;

  tmp_param_GB_hetero_int.class_id_1 = atoi(tokenized_buffer[0].c_str());   /* class id 1 */
  tmp_param_GB_hetero_int.class_id_2 = atoi(tokenized_buffer[1].c_str());   /* class id 2 */
  tmp_param_GB_hetero_int.eps        = atof(tokenized_buffer[2].c_str());   /* heterogeneous hamaker constant */
  tmp_param_GB_hetero_int.sig        = atof(tokenized_buffer[3].c_str());   /* heterogeneous atomic sigma */
  tmp_param_GB_hetero_int.mu         = atoi(tokenized_buffer[4].c_str());   /* heterogeneous hamaker constant */
  tmp_param_GB_hetero_int.nu         = atoi(tokenized_buffer[5].c_str());   /* heterogeneous atomic sigma */

  /********************************** optional parameters ****
   *  NO FAILURE at this point                               *
   *  homogenous parameters will be used if no heterogenous  *
   *  one is found                                           *
   ***********************************************************/

  /* NOTE */
  if (tokenized_buffer.size()>=18)
  {
    tmp_param_GB_hetero_int.spec = true;
    tmp_param_GB_hetero_int.sig11 = atof(tokenized_buffer[6].c_str());   /* sigma_x (1) */
    tmp_param_GB_hetero_int.sig12 = atof(tokenized_buffer[7].c_str());   /* sigma_y (1) */
    tmp_param_GB_hetero_int.sig13 = atof(tokenized_buffer[8].c_str());   /* sigma_z (1) */
    tmp_param_GB_hetero_int.e11   = atof(tokenized_buffer[9].c_str());   /* e_x (1) */
    tmp_param_GB_hetero_int.e12   = atof(tokenized_buffer[10].c_str());   /* e_y (1) */
    tmp_param_GB_hetero_int.e13   = atof(tokenized_buffer[11].c_str());   /* e_z (1) */
    tmp_param_GB_hetero_int.sig21 = atof(tokenized_buffer[12].c_str());  /* sigma_x (2) */
    tmp_param_GB_hetero_int.sig22 = atof(tokenized_buffer[13].c_str());  /* sigma_y (2) */
    tmp_param_GB_hetero_int.sig23 = atof(tokenized_buffer[14].c_str());  /* sigma_z (2) */
    tmp_param_GB_hetero_int.e21   = atof(tokenized_buffer[15].c_str());  /* e_x (2) */
    tmp_param_GB_hetero_int.e22   = atof(tokenized_buffer[16].c_str());  /* e_y (2) */
    tmp_param_GB_hetero_int.e23   = atof(tokenized_buffer[17].c_str());  /* e_z (2) */

  }
  else
    tmp_param_GB_hetero_int.spec = false;

  vec_param_GB_hetero_int.push_back(tmp_param_GB_hetero_int);

  tokenized_buffer.clear();

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::parse_param_GB_hetero_int: adding a new GB heterogeneous interaction parameter set:"
       << endl
       << "  class_id_1:  " << tmp_param_GB_hetero_int.class_id_1 << endl
       << "  class_id_2:  " << tmp_param_GB_hetero_int.class_id_2 << endl;
  cout << setprecision(8) << showpos << showpoint << left
       << "  eps:         " << tmp_param_GB_hetero_int.eps << endl
       << "  sig:         " << tmp_param_GB_hetero_int.sig << endl
       << "  mu:          " << tmp_param_GB_hetero_int.mu << endl
       << "  nu:          " << tmp_param_GB_hetero_int.nu << endl;

  if (tmp_param_GB_hetero_int.spec)
  {
    cout << setw(8) << "  Specific parameters:" << endl
         << setw(8) << "  - sig11:       " << tmp_param_GB_hetero_int.sig11 << endl
         << setw(8) << "  - sig12:       " << tmp_param_GB_hetero_int.sig12 << endl
         << setw(8) << "  - sig13:       " << tmp_param_GB_hetero_int.sig13 << endl
         << setw(8) << "  - e11:         " << tmp_param_GB_hetero_int.e11 << endl
         << setw(8) << "  - e12:         " << tmp_param_GB_hetero_int.e12 << endl
         << setw(8) << "  - e13:         " << tmp_param_GB_hetero_int.e13 << endl
         << setw(8) << "  - sig21:       " << tmp_param_GB_hetero_int.sig21 << endl
         << setw(8) << "  - sig22:       " << tmp_param_GB_hetero_int.sig22 << endl
         << setw(8) << "  - sig23:       " << tmp_param_GB_hetero_int.sig23 << endl
         << setw(8) << "  - e21:         " << tmp_param_GB_hetero_int.e21 << endl
         << setw(8) << "  - e22:         " << tmp_param_GB_hetero_int.e22 << endl
         << setw(8) << "  - e23:         " << tmp_param_GB_hetero_int.e23 << endl;
  }

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::parse_param_EES_int (const string & param_buffer)
{

  vector <string>      tokenized_buffer;
  param_EES_int        tmp_param_EES_int;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  if (tokenized_buffer.size()<13)
    return EXIT_FAILURE;

  tmp_param_EES_int.class_id_surf        = atoi(tokenized_buffer[0].c_str());   /* class id */
  tmp_param_EES_int.class_id_rigid_body  = atoi(tokenized_buffer[1].c_str());   /* class id */
  tmp_param_EES_int.A                    = atof(tokenized_buffer[2].c_str());   /* Energy constant */
  tmp_param_EES_int.sigma_c              = atof(tokenized_buffer[3].c_str());   /* atomic sigma */
  tmp_param_EES_int.sig1                 = atof(tokenized_buffer[4].c_str());   /* sigma_x */
  tmp_param_EES_int.sig2                 = atof(tokenized_buffer[5].c_str());   /* sigma_y */
  tmp_param_EES_int.sig3                 = atof(tokenized_buffer[6].c_str());   /* sigma_z */
  tmp_param_EES_int.e11                  = atof(tokenized_buffer[7].c_str());   /* e_xx */
  tmp_param_EES_int.e22                  = atof(tokenized_buffer[8].c_str());   /* e_yy */
  tmp_param_EES_int.e33                  = atof(tokenized_buffer[9].c_str());   /* e_zz */
  tmp_param_EES_int.e12                  = atof(tokenized_buffer[10].c_str());   /* e_xy */
  tmp_param_EES_int.e23                  = atof(tokenized_buffer[11].c_str());   /* e_yz */
  tmp_param_EES_int.e31                  = atof(tokenized_buffer[12].c_str());   /* e_zx */

  tokenized_buffer.clear();

  vec_param_EES_int.push_back(tmp_param_EES_int);

#ifdef PARAMDB_DEBUG

  cout << "* emd::paramdb::parse_param_EES_int: adding a new EES "
       << "interaction parameter set:" << endl

       <<            "  class_id_surf:        " << tmp_param_EES_int.class_id_surf << endl
       <<            "  class_id_rigid_body:   " << tmp_param_EES_int.class_id_rigid_body << endl

       << setprecision(8) << showpos << showpoint << left

       << setw(8) << "  A:                    " << tmp_param_EES_int.A << " (kcal.mol^-1.ang^-3)" << endl
       << setw(8) << "  sigma_c:              " << tmp_param_EES_int.sigma_c << " (ang)" << endl
       << setw(8) << "  sig1:                 " << tmp_param_EES_int.sig1 << " (ang)" << endl
       << setw(8) << "  sig2:                 " << tmp_param_EES_int.sig2 << " (ang)" << endl
       << setw(8) << "  sig3:                 " << tmp_param_EES_int.sig3 << " (ang)" << endl
       << setw(8) << "  e11:                  " << tmp_param_EES_int.e11 << endl
       << setw(8) << "  e22:                  " << tmp_param_EES_int.e22 << endl
       << setw(8) << "  e33:                  " << tmp_param_EES_int.e33 << endl
       << setw(8) << "  e12:                  " << tmp_param_EES_int.e12 << endl
       << setw(8) << "  e23:                  " << tmp_param_EES_int.e23 << endl
       << setw(8) << "  e31:                  " << tmp_param_EES_int.e31 << endl

       << noshowpos << noshowpoint;

#endif

  return EXIT_SUCCESS;
}

/**************************************** post-processing *********************/
int
emd::paramdb::param_post_proc ()
{

  /* RE^2 heterogeneous interactions post processing */
  if (param_RE2_hetero_int_post_proc()==EXIT_FAILURE) return EXIT_FAILURE;

  /* GB heterogeneous interactions post processing */
  if (param_GB_hetero_int_post_proc()==EXIT_FAILURE) return EXIT_FAILURE;

  /* parameter caches for fast access */
  if (init_param_cache_rigid_body()==EXIT_FAILURE) return EXIT_FAILURE;
  if (init_param_cache_int_surf()==EXIT_FAILURE) return EXIT_FAILURE;
  if (init_param_cache_RE2_int()==EXIT_FAILURE) return EXIT_FAILURE;
  if (init_param_cache_GB_int()==EXIT_FAILURE) return EXIT_FAILURE;
  if (init_param_cache_EES_int()==EXIT_FAILURE) return EXIT_FAILURE;

  /* freeing the vectors */
  vec_param_rigid_body.clear();
  vec_param_int_surf.clear();
  vec_param_RE2_homo_int.clear();
  vec_param_RE2_hetero_int.clear();
  vec_param_GB_homo_int.clear();
  vec_param_GB_hetero_int.clear();
  vec_param_EES_int.clear();

#ifdef PARAMDB_DEBUG
  cout << "* emd::paramdb::param_post_proc: post-processing completed successfully!" << endl;
#endif

  return EXIT_SUCCESS;

}

/**********************************************************************************/

int
emd::paramdb::param_RE2_hetero_int_post_proc ()
{

  /* setting up heterogeneous interactions */
  for (int i=0; i<vec_param_RE2_hetero_int.size(); i++)
  {

    if (!vec_param_RE2_hetero_int[i].spec)
    {

      int class_id_1 = vec_param_RE2_hetero_int[i].class_id_1;
      int class_id_2 = vec_param_RE2_hetero_int[i].class_id_2;

      int i1, i2, class_homo_id_1, class_homo_id_2;
      bool flag = false;

      /* search through homo interaction parameters */
      for (int j=0; j<vec_param_RE2_homo_int.size()-1; j++)
      {
        class_homo_id_1 = vec_param_RE2_homo_int[j].class_id;
        for (int k=j+1; k<vec_param_RE2_homo_int.size(); k++)
        {
          class_homo_id_2 = vec_param_RE2_homo_int[k].class_id;

          if (class_id_1==class_homo_id_1 && class_id_2==class_homo_id_2)
            { i1 = j; i2 = k; flag = true; }
          else if (class_id_1==class_homo_id_2 && class_id_2==class_homo_id_1)
            { i1 = k; i2 = j; flag = true; }

          if (flag)
          {
            vec_param_RE2_hetero_int[i].sig11 = vec_param_RE2_homo_int[i1].sig1;
            vec_param_RE2_hetero_int[i].sig12 = vec_param_RE2_homo_int[i1].sig2;
            vec_param_RE2_hetero_int[i].sig13 = vec_param_RE2_homo_int[i1].sig3;
            vec_param_RE2_hetero_int[i].sig21 = vec_param_RE2_homo_int[i2].sig1;
            vec_param_RE2_hetero_int[i].sig22 = vec_param_RE2_homo_int[i2].sig2;
            vec_param_RE2_hetero_int[i].sig23 = vec_param_RE2_homo_int[i2].sig3;
            vec_param_RE2_hetero_int[i].e11 = vec_param_RE2_homo_int[i1].e1;
            vec_param_RE2_hetero_int[i].e12 = vec_param_RE2_homo_int[i1].e2;
            vec_param_RE2_hetero_int[i].e13 = vec_param_RE2_homo_int[i1].e3;
            vec_param_RE2_hetero_int[i].e21 = vec_param_RE2_homo_int[i2].e1;
            vec_param_RE2_hetero_int[i].e22 = vec_param_RE2_homo_int[i2].e2;
            vec_param_RE2_hetero_int[i].e23 = vec_param_RE2_homo_int[i2].e3;
            break;
          }

        }

        if (flag) break;

      }

      if (!flag)
      {
        cerr << "* emd::paramdb::param_post_proc: error setting up heterogeneous parameters!"
             << endl;
      }

    }

  }

  return EXIT_SUCCESS;

}

/**********************************************************************************/

int
emd::paramdb::param_GB_hetero_int_post_proc ()
{

  /* setting up heterogeneous interactions */
  for (int i=0; i<vec_param_GB_hetero_int.size(); i++)
  {

    if (!vec_param_GB_hetero_int[i].spec)
    {

      int class_id_1 = vec_param_GB_hetero_int[i].class_id_1;
      int class_id_2 = vec_param_GB_hetero_int[i].class_id_2;

      int i1, i2, class_homo_id_1, class_homo_id_2;
      bool flag = false;

      /* search through homo interaction parameters */
      for (int j=0; j<vec_param_GB_homo_int.size()-1; j++)
      {
        class_homo_id_1 = vec_param_GB_homo_int[j].class_id;
        for (int k=j+1; k<vec_param_GB_homo_int.size(); k++)
        {
          class_homo_id_2 = vec_param_GB_homo_int[k].class_id;

          if (class_id_1==class_homo_id_1 && class_id_2==class_homo_id_2)
            { i1 = j; i2 = k; flag = true; }
          else if (class_id_1==class_homo_id_2 && class_id_2==class_homo_id_1)
            { i1 = k; i2 = j; flag = true; }

          if (flag)
          {
            vec_param_GB_hetero_int[i].sig11 = vec_param_GB_homo_int[i1].sig1;
            vec_param_GB_hetero_int[i].sig12 = vec_param_GB_homo_int[i1].sig2;
            vec_param_GB_hetero_int[i].sig13 = vec_param_GB_homo_int[i1].sig3;
            vec_param_GB_hetero_int[i].sig21 = vec_param_GB_homo_int[i2].sig1;
            vec_param_GB_hetero_int[i].sig22 = vec_param_GB_homo_int[i2].sig2;
            vec_param_GB_hetero_int[i].sig23 = vec_param_GB_homo_int[i2].sig3;
            vec_param_GB_hetero_int[i].e11 = vec_param_GB_homo_int[i1].e1;
            vec_param_GB_hetero_int[i].e12 = vec_param_GB_homo_int[i1].e2;
            vec_param_GB_hetero_int[i].e13 = vec_param_GB_homo_int[i1].e3;
            vec_param_GB_hetero_int[i].e21 = vec_param_GB_homo_int[i2].e1;
            vec_param_GB_hetero_int[i].e22 = vec_param_GB_homo_int[i2].e2;
            vec_param_GB_hetero_int[i].e23 = vec_param_GB_homo_int[i2].e3;
            break;
          }

        }

        if (flag) break;

      }

      if (!flag)
      {
        cerr << "* emd::paramdb::param_post_proc: error setting up heterogeneous parameters!"
             << endl;
      }

    }

  }

  return EXIT_SUCCESS;

}

/**********************************************************************************/

int
emd::paramdb::init_param_cache_rigid_body ()
{

  if (vec_param_rigid_body.size()==0)
    return EXIT_SUCCESS;

  rigid_body_param_max_idx = vec_param_rigid_body.size()-1;
  rigid_body_param = new param_cache_rigid_body[rigid_body_param_max_idx+1];

  for (int i=0; i<rigid_body_param_max_idx+1; i++)
  {

    rigid_body_param[i].class_id  =  vec_param_rigid_body[i].class_id;
    rigid_body_param[i].name      =  vec_param_rigid_body[i].name;
    rigid_body_param[i].mass      =  vec_param_rigid_body[i].mass;
    rigid_body_param[i].I1        =  vec_param_rigid_body[i].I_body_x;
    rigid_body_param[i].I2        =  vec_param_rigid_body[i].I_body_y;
    rigid_body_param[i].I3        =  vec_param_rigid_body[i].I_body_z;

    gsl_matrix_set(rigid_body_param[i].I_body,     0, 0, vec_param_rigid_body[i].I_body_x);
    gsl_matrix_set(rigid_body_param[i].I_body,     1, 1, vec_param_rigid_body[i].I_body_y);
    gsl_matrix_set(rigid_body_param[i].I_body,     2, 2, vec_param_rigid_body[i].I_body_z);

    gsl_matrix_set(rigid_body_param[i].I_body_inv, 0, 0, 1/vec_param_rigid_body[i].I_body_x);
    gsl_matrix_set(rigid_body_param[i].I_body_inv, 1, 1, 1/vec_param_rigid_body[i].I_body_y);
    gsl_matrix_set(rigid_body_param[i].I_body_inv, 2, 2, 1/vec_param_rigid_body[i].I_body_z);

  }

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::init_param_cache_int_surf ()
{

  if (vec_param_rigid_body.size()==0)
    return EXIT_SUCCESS;

  int_surf_param_max_idx = vec_param_int_surf.size()-1;
  int_surf_param = new param_cache_int_surf[int_surf_param_max_idx+1];

  for (int i=0; i<int_surf_param_max_idx+1; i++)
  {
    int_surf_param[i].class_id  =  vec_param_int_surf[i].class_id;
  }

  return EXIT_SUCCESS;

}

/**********************************************************************************/

/* assert: init_param_cache_rigid_body() is run beforehand */
int
emd::paramdb::init_param_cache_RE2_int()
{

  re2_int_param = new param_cache_RE2_int*[rigid_body_param_max_idx+1];
  int class_id_1, class_id_2;

  for (int idx1=0; idx1<=rigid_body_param_max_idx; idx1++)
  {

    re2_int_param[idx1] = new param_cache_RE2_int[rigid_body_param_max_idx+1];
    class_id_1 = rigid_body_param[idx1].class_id;

    for (int idx2=0; idx2<=rigid_body_param_max_idx; idx2++)
    {
      class_id_2 = rigid_body_param[idx2].class_id;

      if (class_id_1==class_id_2) /* homogeneous */
      {

        bool flag = false;

        for (int k=0; k<vec_param_RE2_homo_int.size(); k++)
          if (vec_param_RE2_homo_int[k].class_id==class_id_1)
          {
            flag = true;

            param_RE2_homo_int *    src  =  &(vec_param_RE2_homo_int[k]);
            param_cache_RE2_int *  dest  =  &(re2_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id;                                  /* class_id_1 */
            dest->class_id_2  =  src->class_id;                                  /* class_id_2 */
            dest->h_type      =  src->h_type;                                    /* h type */

            if (src->h_type == H_FORMULA_SPH)
            {
              cout << "* emd::paramdb::initparam_cache_RE2_int: giant sphere h formula is only for"
                   << " heterogenous interaction!" << endl;
              dest->h_type = H_FORMULA_GB; /* default */
            }

            dest->hamaker     =  src->hamaker;                                   /* hamaker */
            dest->sigma_c     =  src->sigma_c;                                   /* atomic sigma */

            dest->dS1         =  src->sig1 * src->sig2 * src->sig3;              /* det[S_1] */
            dest->dS2         =  dest->dS1;                                      /* det[S_2] */

            dest->max_thickness = max(src->sig1, src->sig2);
            dest->max_thickness = max(dest->max_thickness, src->sig3);
            dest->max_thickness *= 2;

            gsl_matrix_set(dest->S1,    0,  0,  src->sig1);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig2);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig3);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig1 * src->sig1);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig2 * src->sig2);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig3 * src->sig3);
            gsl_matrix_set(dest->S1pm2, 0,  0,  1/(src->sig1 * src->sig1));
            gsl_matrix_set(dest->S1pm2, 1,  1,  1/(src->sig2 * src->sig2));
            gsl_matrix_set(dest->S1pm2, 2,  2,  1/(src->sig3 * src->sig3));
            gsl_matrix_set(dest->E1,    0,  0,  src->e1);
            gsl_matrix_set(dest->E1,    1,  1,  src->e2);
            gsl_matrix_set(dest->E1,    2,  2,  src->e3);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e1 * src->e1);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e2 * src->e2);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e3 * src->e3);

            gsl_matrix_memcpy(dest->S2,     dest->S1);
            gsl_matrix_memcpy(dest->S2p2,   dest->S1p2);
            gsl_matrix_memcpy(dest->S2pm2,  dest->S1pm2);
            gsl_matrix_memcpy(dest->E2,     dest->E1);
            gsl_matrix_memcpy(dest->E2p2,   dest->E1p2);

          }

#ifdef EMD_PARAMDB_SHOW_WARNINGS
        if (!flag)
          cerr << "* emd::paramdb::initparam_cache_RE2_int: WARNING! homogenous interaction parameter not found"
               << " (class_id: " << class_id_1 << ")!" << endl;
#endif

      }
      else /* heterogeneous */
      {
        bool flag = false;

        for (int k=0; k<vec_param_RE2_hetero_int.size(); k++)
          if ( (vec_param_RE2_hetero_int[k].class_id_1==class_id_1) &&
               (vec_param_RE2_hetero_int[k].class_id_2==class_id_2) )
          {

            flag = true;

            param_RE2_hetero_int *  src  =  &(vec_param_RE2_hetero_int[k]);
            param_cache_RE2_int *  dest  =  &(re2_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id_1;                                /* class_id_1 */
            dest->class_id_2  =  src->class_id_2;                                /* class_id_2 */
            dest->h_type      =  src->h_type;                                    /* h type */

            dest->hamaker     =  src->hamaker;                                   /* hamaker */
            dest->sigma_c     =  src->sigma_c;                                   /* atomic sigma */

            dest->dS1         =  src->sig11 * src->sig12 * src->sig13;           /* det[S_1] */
            dest->dS2         =  src->sig21 * src->sig22 * src->sig23;           /* det[S_2] */

            double tmp1       = max(src->sig11, src->sig12);
            tmp1              = max(tmp1, src->sig13);
            double tmp2       = max(src->sig21, src->sig22);
            tmp2              = max(tmp2, src->sig23);
            dest->max_thickness = tmp1 + tmp2;

            gsl_matrix_set(dest->S1,    0,  0,  src->sig11);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig12);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig13);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig11 * src->sig11);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig12 * src->sig12);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig13 * src->sig13);
            gsl_matrix_set(dest->S1pm2, 0,  0,  1/(src->sig11 * src->sig11));
            gsl_matrix_set(dest->S1pm2, 1,  1,  1/(src->sig12 * src->sig12));
            gsl_matrix_set(dest->S1pm2, 2,  2,  1/(src->sig13 * src->sig13));
            gsl_matrix_set(dest->E1,    0,  0,  src->e11);
            gsl_matrix_set(dest->E1,    1,  1,  src->e12);
            gsl_matrix_set(dest->E1,    2,  2,  src->e13);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e11 * src->e11);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e12 * src->e12);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e13 * src->e13);

            gsl_matrix_set(dest->S2,    0,  0,  src->sig21);
            gsl_matrix_set(dest->S2,    1,  1,  src->sig22);
            gsl_matrix_set(dest->S2,    2,  2,  src->sig23);
            gsl_matrix_set(dest->S2p2,  0,  0,  src->sig21 * src->sig21);
            gsl_matrix_set(dest->S2p2,  1,  1,  src->sig22 * src->sig22);
            gsl_matrix_set(dest->S2p2,  2,  2,  src->sig23 * src->sig23);
            gsl_matrix_set(dest->S2pm2, 0,  0,  1/(src->sig21 * src->sig21));
            gsl_matrix_set(dest->S2pm2, 1,  1,  1/(src->sig22 * src->sig22));
            gsl_matrix_set(dest->S2pm2, 2,  2,  1/(src->sig23 * src->sig23));
            gsl_matrix_set(dest->E2,    0,  0,  src->e21);
            gsl_matrix_set(dest->E2,    1,  1,  src->e22);
            gsl_matrix_set(dest->E2,    2,  2,  src->e23);
            gsl_matrix_set(dest->E2p2,  0,  0,  src->e21 * src->e21);
            gsl_matrix_set(dest->E2p2,  1,  1,  src->e22 * src->e22);
            gsl_matrix_set(dest->E2p2,  2,  2,  src->e23 * src->e23);

          }

          else if ( (vec_param_RE2_hetero_int[k].class_id_1==class_id_2) &&
                    (vec_param_RE2_hetero_int[k].class_id_2==class_id_1) )

          {

            flag = true;

            param_RE2_hetero_int *  src  =  &(vec_param_RE2_hetero_int[k]);
            param_cache_RE2_int *  dest  =  &(re2_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id_2;                                /* class_id_1 */
            dest->class_id_2  =  src->class_id_1;                                /* class_id_2 */
            dest->h_type      =  src->h_type;                                    /* h type */

            dest->hamaker     =  src->hamaker;                                   /* hamaker */
            dest->sigma_c     =  src->sigma_c;                                   /* atomic sigma */

            dest->dS2         =  src->sig11 * src->sig12 * src->sig13;           /* det[S_1] */
            dest->dS1         =  src->sig21 * src->sig22 * src->sig23;           /* det[S_2] */

            gsl_matrix_set(dest->S2,    0,  0,  src->sig11);
            gsl_matrix_set(dest->S2,    1,  1,  src->sig12);
            gsl_matrix_set(dest->S2,    2,  2,  src->sig13);
            gsl_matrix_set(dest->S2p2,  0,  0,  src->sig11 * src->sig11);
            gsl_matrix_set(dest->S2p2,  1,  1,  src->sig12 * src->sig12);
            gsl_matrix_set(dest->S2p2,  2,  2,  src->sig13 * src->sig13);
            gsl_matrix_set(dest->S2pm2, 0,  0,  1/(src->sig11 * src->sig11));
            gsl_matrix_set(dest->S2pm2, 1,  1,  1/(src->sig12 * src->sig12));
            gsl_matrix_set(dest->S2pm2, 2,  2,  1/(src->sig13 * src->sig13));
            gsl_matrix_set(dest->E2,    0,  0,  src->e11);
            gsl_matrix_set(dest->E2,    1,  1,  src->e12);
            gsl_matrix_set(dest->E2,    2,  2,  src->e13);
            gsl_matrix_set(dest->E2p2,  0,  0,  src->e11 * src->e11);
            gsl_matrix_set(dest->E2p2,  1,  1,  src->e12 * src->e12);
            gsl_matrix_set(dest->E2p2,  2,  2,  src->e13 * src->e13);

            gsl_matrix_set(dest->S1,    0,  0,  src->sig21);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig22);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig23);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig21 * src->sig21);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig22 * src->sig22);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig23 * src->sig23);
            gsl_matrix_set(dest->S1pm2, 0,  0,  1/(src->sig21 * src->sig21));
            gsl_matrix_set(dest->S1pm2, 1,  1,  1/(src->sig22 * src->sig22));
            gsl_matrix_set(dest->S1pm2, 2,  2,  1/(src->sig23 * src->sig23));
            gsl_matrix_set(dest->E1,    0,  0,  src->e21);
            gsl_matrix_set(dest->E1,    1,  1,  src->e22);
            gsl_matrix_set(dest->E1,    2,  2,  src->e23);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e21 * src->e21);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e22 * src->e22);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e23 * src->e23);

          }

#ifdef EMD_PARAMDB_SHOW_WARNINGS
        if (!flag)
          cerr << "* emd::paramdb::init_param_cache_RE2_int: WARNING! heterogenous interaction parameter not found"
               << " (class_id_1: " << class_id_1 << ", class_id_2: " << class_id_2 << ")!" << endl;
#endif

      }

    }

  }

  return EXIT_SUCCESS;
}

/**********************************************************************************/

/* assert: init_param_cache_rigid_body() is run beforehand */
int
emd::paramdb::init_param_cache_GB_int()
{

  gb_int_param = new param_cache_GB_int*[rigid_body_param_max_idx+1];
  int class_id_1, class_id_2;

  for (int idx1=0; idx1<=rigid_body_param_max_idx; idx1++)
  {

    gb_int_param[idx1] = new param_cache_GB_int[rigid_body_param_max_idx+1];
    class_id_1 = rigid_body_param[idx1].class_id;

    for (int idx2=0; idx2<=rigid_body_param_max_idx; idx2++)
    {
      class_id_2 = rigid_body_param[idx2].class_id;

      if (class_id_1==class_id_2) /* homogeneous */
      {

        bool flag = false;

        for (int k=0; k<vec_param_GB_homo_int.size(); k++)
          if (vec_param_GB_homo_int[k].class_id==class_id_1)
          {
            flag = true;

            param_GB_homo_int *    src  =  &(vec_param_GB_homo_int[k]);
            param_cache_GB_int *  dest  =  &(gb_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id;                                  /* class_id_1 */
            dest->class_id_2  =  src->class_id;                                  /* class_id_2 */

            dest->eps         =  src->eps;                                       /* energy */
            dest->sig         =  src->sig;                                       /* length */
            dest->mu          =  src->mu;                                        /* mu */
            dest->nu          =  src->nu;                                        /* nu */

            /* prefactor in eta */
            dest->dS1         = (src->sig1 * src->sig2 + src->sig3 * src->sig3) * sqrt(src->sig1 * src->sig2);
            dest->dS2         =  dest->dS1;

            dest->max_thickness = max(src->sig1, src->sig2);
            dest->max_thickness = max(dest->max_thickness, src->sig3);
            dest->max_thickness *= 2;

            gsl_matrix_set(dest->S1,    0,  0,  src->sig1);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig2);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig3);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig1 * src->sig1);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig2 * src->sig2);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig3 * src->sig3);
            gsl_matrix_set(dest->E1,    0,  0,  src->e1);
            gsl_matrix_set(dest->E1,    1,  1,  src->e2);
            gsl_matrix_set(dest->E1,    2,  2,  src->e3);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e1 * src->e1);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e2 * src->e2);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e3 * src->e3);

            gsl_matrix_memcpy(dest->S2,     dest->S1);
            gsl_matrix_memcpy(dest->S2p2,   dest->S1p2);
            gsl_matrix_memcpy(dest->E2,     dest->E1);
            gsl_matrix_memcpy(dest->E2p2,   dest->E1p2);

          }

#ifdef EMD_PARAMDB_SHOW_WARNINGS
        if (!flag)
          cerr << "* emd::paramdb::initparam_cache_GB_int: WARNING! homogenous interaction parameter not found"
               << " (class_id: " << class_id_1 << ")!" << endl;
#endif

      }
      else /* heterogeneous */
      {
        bool flag = false;

        for (int k=0; k<vec_param_GB_hetero_int.size(); k++)
          if ( (vec_param_GB_hetero_int[k].class_id_1==class_id_1) &&
               (vec_param_GB_hetero_int[k].class_id_2==class_id_2) )
          {

            flag = true;

            param_GB_hetero_int *  src  =  &(vec_param_GB_hetero_int[k]);
            param_cache_GB_int *  dest  =  &(gb_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id_1;                                /* class_id_1 */
            dest->class_id_2  =  src->class_id_2;                                /* class_id_2 */

            dest->eps         =  src->eps;                                       /* energy */
            dest->sig         =  src->sig;                                       /* length */
            dest->mu          =  src->mu;                                        /* mu */
            dest->nu          =  src->nu;                                        /* nu */

            /* prefactor in eta */
            dest->dS1         = (src->sig11 * src->sig12 + src->sig13 * src->sig13) * sqrt(src->sig11 * src->sig12);
            dest->dS2         = (src->sig21 * src->sig22 + src->sig23 * src->sig23) * sqrt(src->sig21 * src->sig22);

            double tmp1       = max(src->sig11, src->sig12);
            tmp1              = max(tmp1, src->sig13);
            double tmp2       = max(src->sig21, src->sig22);
            tmp2              = max(tmp2, src->sig23);
            dest->max_thickness = tmp1 + tmp2;

            gsl_matrix_set(dest->S1,    0,  0,  src->sig11);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig12);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig13);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig11 * src->sig11);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig12 * src->sig12);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig13 * src->sig13);
            gsl_matrix_set(dest->E1,    0,  0,  src->e11);
            gsl_matrix_set(dest->E1,    1,  1,  src->e12);
            gsl_matrix_set(dest->E1,    2,  2,  src->e13);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e11 * src->e11);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e12 * src->e12);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e13 * src->e13);

            gsl_matrix_set(dest->S2,    0,  0,  src->sig21);
            gsl_matrix_set(dest->S2,    1,  1,  src->sig22);
            gsl_matrix_set(dest->S2,    2,  2,  src->sig23);
            gsl_matrix_set(dest->S2p2,  0,  0,  src->sig21 * src->sig21);
            gsl_matrix_set(dest->S2p2,  1,  1,  src->sig22 * src->sig22);
            gsl_matrix_set(dest->S2p2,  2,  2,  src->sig23 * src->sig23);
            gsl_matrix_set(dest->E2,    0,  0,  src->e21);
            gsl_matrix_set(dest->E2,    1,  1,  src->e22);
            gsl_matrix_set(dest->E2,    2,  2,  src->e23);
            gsl_matrix_set(dest->E2p2,  0,  0,  src->e21 * src->e21);
            gsl_matrix_set(dest->E2p2,  1,  1,  src->e22 * src->e22);
            gsl_matrix_set(dest->E2p2,  2,  2,  src->e23 * src->e23);

          }

          else if ( (vec_param_GB_hetero_int[k].class_id_1==class_id_2) &&
                    (vec_param_GB_hetero_int[k].class_id_2==class_id_1) )

          {

            flag = true;

            param_GB_hetero_int *  src  =  &(vec_param_GB_hetero_int[k]);
            param_cache_GB_int *  dest  =  &(gb_int_param[idx1][idx2]);

            /* setting up parameters */

            dest->class_id_1  =  src->class_id_2;                                /* class_id_1 */
            dest->class_id_2  =  src->class_id_1;                                /* class_id_2 */

            dest->eps         =  src->eps;                                       /* energy */
            dest->sig         =  src->sig;                                       /* length */
            dest->mu          =  src->mu;                                        /* mu */
            dest->nu          =  src->nu;                                        /* nu */

            /* prefactor in eta */
            dest->dS2         = (src->sig11 * src->sig12 + src->sig13 * src->sig13) * sqrt(src->sig11 * src->sig12);
            dest->dS1         = (src->sig21 * src->sig22 + src->sig23 * src->sig23) * sqrt(src->sig21 * src->sig22);

            gsl_matrix_set(dest->S2,    0,  0,  src->sig11);
            gsl_matrix_set(dest->S2,    1,  1,  src->sig12);
            gsl_matrix_set(dest->S2,    2,  2,  src->sig13);
            gsl_matrix_set(dest->S2p2,  0,  0,  src->sig11 * src->sig11);
            gsl_matrix_set(dest->S2p2,  1,  1,  src->sig12 * src->sig12);
            gsl_matrix_set(dest->S2p2,  2,  2,  src->sig13 * src->sig13);
            gsl_matrix_set(dest->E2,    0,  0,  src->e11);
            gsl_matrix_set(dest->E2,    1,  1,  src->e12);
            gsl_matrix_set(dest->E2,    2,  2,  src->e13);
            gsl_matrix_set(dest->E2p2,  0,  0,  src->e11 * src->e11);
            gsl_matrix_set(dest->E2p2,  1,  1,  src->e12 * src->e12);
            gsl_matrix_set(dest->E2p2,  2,  2,  src->e13 * src->e13);

            gsl_matrix_set(dest->S1,    0,  0,  src->sig21);
            gsl_matrix_set(dest->S1,    1,  1,  src->sig22);
            gsl_matrix_set(dest->S1,    2,  2,  src->sig23);
            gsl_matrix_set(dest->S1p2,  0,  0,  src->sig21 * src->sig21);
            gsl_matrix_set(dest->S1p2,  1,  1,  src->sig22 * src->sig22);
            gsl_matrix_set(dest->S1p2,  2,  2,  src->sig23 * src->sig23);
            gsl_matrix_set(dest->E1,    0,  0,  src->e21);
            gsl_matrix_set(dest->E1,    1,  1,  src->e22);
            gsl_matrix_set(dest->E1,    2,  2,  src->e23);
            gsl_matrix_set(dest->E1p2,  0,  0,  src->e21 * src->e21);
            gsl_matrix_set(dest->E1p2,  1,  1,  src->e22 * src->e22);
            gsl_matrix_set(dest->E1p2,  2,  2,  src->e23 * src->e23);

          }

#ifdef EMD_PARAMDB_SHOW_WARNINGS
        if (!flag)
          cerr << "* emd::paramdb::init_param_cache_GB_int: WARNING! heterogenous interaction parameter not found"
               << " (class_id_1: " << class_id_1 << ", class_id_2: " << class_id_2 << ")!" << endl;
#endif

      }

    }

  }

  return EXIT_SUCCESS;
}

/**********************************************************************************/

int
emd::paramdb::init_param_cache_EES_int()
{

  ees_int_param = new param_cache_EES_int*[int_surf_param_max_idx+1];
  int class_id_surf, class_id_rigid_body;

  for (int idx1=0; idx1<=int_surf_param_max_idx; idx1++)
  {

    /* malloc */
    ees_int_param[idx1] = new param_cache_EES_int[rigid_body_param_max_idx+1];

    class_id_surf = int_surf_param[idx1].class_id;

    for (int idx2=0; idx2<=rigid_body_param_max_idx; idx2++)
    {
      class_id_rigid_body = rigid_body_param[idx2].class_id;

      bool flag = false;

      for (int k=0; k<vec_param_EES_int.size(); k++)
        if (vec_param_EES_int[k].class_id_surf == class_id_surf &&
            vec_param_EES_int[k].class_id_rigid_body == class_id_rigid_body)
        {

          flag = true;

          param_EES_int *        src   =  &(vec_param_EES_int[k]);
          param_cache_EES_int *  dest  =  &(ees_int_param[idx1][idx2]);

          /* setting up parameters */

          dest->class_id_surf       =  class_id_surf;
          dest->class_id_rigid_body  =  class_id_rigid_body;

          double sig6               =  gsl_pow_6(src->sigma_c);
          dest->A_fact              =  src->A * sig6;
          dest->R_fact              =  src->A * sig6 * sig6;
          dest->trE                 =  src->e11 + src->e22 + src->e33;
          dest->max_thickness       =  max(src->sig1, src->sig2);
          dest->max_thickness       =  max(dest->max_thickness, src->sig3);

          gsl_matrix_set(dest->S,    0,  0,  src->sig1);
          gsl_matrix_set(dest->S,    1,  1,  src->sig2);
          gsl_matrix_set(dest->S,    2,  2,  src->sig3);

          /* diagonal */
          gsl_matrix_set(dest->E,    0,  0,  src->e11);
          gsl_matrix_set(dest->E,    1,  1,  src->e22);
          gsl_matrix_set(dest->E,    2,  2,  src->e33);

          /* off diagonal */
          gsl_matrix_set(dest->E,    0,  1,  src->e12);
          gsl_matrix_set(dest->E,    1,  2,  src->e23);
          gsl_matrix_set(dest->E,    2,  0,  src->e31);
          gsl_matrix_set(dest->E,    1,  0,  src->e12);
          gsl_matrix_set(dest->E,    2,  1,  src->e23);
          gsl_matrix_set(dest->E,    0,  2,  src->e31);

        }

#ifdef EMD_PARAMDB_SHOW_WARNINGS
      if (!flag)
        cerr << "* emd::paramdb::init_param_cache_EES_int: WARNING: interaction parameter not found"
             << " (class_id_surf: " << class_id_surf << ", class_id_rigid_body: "
             << class_id_rigid_body << ")!" << endl;
#endif

    }

  }

  return EXIT_SUCCESS;
}

/**********************************************************************************/

/* assert: init_param_cache_rigid_body() is run beforehand */
int
emd::paramdb::get_rigid_body_index (int class_id)
{

  for (int i=0; i<=rigid_body_param_max_idx; i++)
    if (rigid_body_param[i].class_id==class_id)
      return i;

  cerr << "* emd::paramdb::get_rigid_body_index: no such class id is present!"
       << endl;

  return 0;

}

/**********************************************************************************/

int
emd::paramdb::get_int_surf_index (int class_id)
{

  for (int i=0; i<=int_surf_param_max_idx; i++)
    if (int_surf_param[i].class_id==class_id)
      return i;

  cerr << "* emd::paramdb::get_int_surf_index: no such class id is present!"
       << endl;

  return 0;

}

/**********************************************************************************/

bool
emd::paramdb::is_init ()
{
  return init;
}

/**********************************************************************************/

int
emd::paramdb::free_memory ()
{

  /* free param caches */

  if (init)
  {

    delete[] rigid_body_param;

    delete[] int_surf_param;

    for (int i=0; i<=rigid_body_param_max_idx; i++)
    {
      delete[] re2_int_param[i];
      delete[] gb_int_param[i];
    }

    delete[] re2_int_param;
    delete[] gb_int_param;

    for (int i=0; i<=int_surf_param_max_idx; i++)
    {
      delete[] ees_int_param[i];
    }

    delete[] ees_int_param;

  }

  init = false;

  return EXIT_SUCCESS;

}
