/* gsl_quaternion.h
 * 
 * Copyright (C) 2005 Mehrtash Babadi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __GSL_QUATERNION_H__
#define __GSL_QUATERNION_H__

#undef __BEGIN_DECLS
#undef __END_DECLS
#ifdef __cplusplus
# define __BEGIN_DECLS extern "C" {
# define __END_DECLS }
#else
# define __BEGIN_DECLS /* empty */
# define __END_DECLS /* empty */
#endif

__BEGIN_DECLS

/* two consecutive built-in types as a quaternion */
typedef double *       gsl_quaternion_packed ;
typedef float *        gsl_quaternion_packed_float ;
typedef long double *  gsl_quaternion_packed_long_double ;

typedef const double *       gsl_const_quaternion_packed ;
typedef const float *        gsl_const_quaternion_packed_float ;
typedef const long double *  gsl_const_quaternion_packed_long_double ;


/* 2N consecutive built-in types as N quaternions */
typedef double *       gsl_quaternion_packed_array ;
typedef float *        gsl_quaternion_packed_array_float  ;
typedef long double *  gsl_quaternion_packed_array_long_double ;

typedef const double *       gsl_const_quaternion_packed_array ;
typedef const float *        gsl_const_quaternion_packed_array_float  ;
typedef const long double *  gsl_const_quaternion_packed_array_long_double ;

/* Yes... this seems weird. Trust us. The point is just that
   sometimes you want to make it obvious that something is
   an output value. The fact that it lacks a 'const' may not
   be enough of a clue for people in some contexts.
 */

typedef double *       gsl_quaternion_packed_ptr ;
typedef float *        gsl_quaternion_packed_float_ptr  ;
typedef long double *  gsl_quaternion_packed_long_double_ptr ;

typedef const double *       gsl_const_quaternion_packed_ptr ;
typedef const float *        gsl_const_quaternion_packed_float_ptr  ;
typedef const long double *  gsl_const_quaternion_packed_long_double_ptr ;


// long double quaternion
typedef struct
{
    long double dat[4];
}
gsl_quaternion_long_double;

// double quaternion
typedef struct
{
    double dat[4];
}
gsl_quaternion;

// float quaternion
typedef struct
{
    float dat[24];
}
gsl_quaternion_float;

#define GSL_QUATERNION_REAL(q)      ((q).dat[0])
#define GSL_QUATERNION_IMAG1(q)     ((q).dat[1])
#define GSL_QUATERNION_IMAG2(q)     ((q).dat[2])
#define GSL_QUATERNION_IMAG3(q)     ((q).dat[3])

#define GSL_QUATERNION_P(qp) ((qp)->dat)

#define GSL_QUATERNION_P_REAL(qp)  ((qp)->dat[0])
#define GSL_QUATERNION_P_IMAG1(qp)  ((qp)->dat[1])
#define GSL_QUATERNION_P_IMAG2(qp)  ((qp)->dat[2])
#define GSL_QUATERNION_P_IMAG3(qp)  ((qp)->dat[3])

#define GSL_QUATERNION_EQ(q1,q2) (((q1).dat[0] == (q2).dat[0]) && ((q1).dat[1] == (q2).dat[1]) \
  ((q1).dat[2] == (q2).dat[2]) && ((q1).dat[3] == (q2).dat[3]))

#define GSL_SET_QUATERNION(qp,x,y1,y2,y3) do {(qp)->dat[0]=(x); (qp)->dat[1]=(y1); \
  (qp)->dat[2]=(y2); (qp)->dat[3]=(y3);} while(0)

#define GSL_QUATERNION_SET_REAL(qp,x) do {(qp)->dat[0]=(x);} while(0)
#define GSL_QUATERNION_SET_IMAG1(qp,y1) do {(qp)->dat[1]=(y1);} while(0)
#define GSL_QUATERNION_SET_IMAG2(qp,y2) do {(qp)->dat[2]=(y2);} while(0)
#define GSL_QUATERNION_SET_IMAG3(qp,y3) do {(qp)->dat[3]=(y3);} while(0)

#define GSL_SET_QUATERNION_PACKED(qp,n,x,y1,y2,y3) do {*((qp)+2*(n))=(x); \
  *((qp)+(2*(n)+1))=(y1); *((qp)+(2*(n)+2))=(y2); *((qp)+(2*(n)+3))=(y3);} while(0)

__END_DECLS

#endif /* __GSL_QUATERNION_H__ */
