/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "surface.h"

using namespace std;

/***************************************************************************/

emd::flat_int_surface::flat_int_surface()
{

  n      =  gsl_vector_alloc(3);
  r0     =  gsl_vector_alloc(3);
  force  =  gsl_vector_alloc(3);

#ifdef SURF_DEBUG
  cout << "* emd::flat_int_surface: a new flat int. surf. constructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

emd::flat_int_surface::flat_int_surface(const flat_int_surface & r_flat_int_surface)
{

  n         =  gsl_vector_alloc(3);
  r0        =  gsl_vector_alloc(3);
  force     =  gsl_vector_alloc(3);

  class_id  =  r_flat_int_surface.class_id;
  idx       =  r_flat_int_surface.idx;

  gsl_vector_memcpy(n, r_flat_int_surface.n);
  gsl_vector_memcpy(r0, r_flat_int_surface.r0);
  gsl_vector_memcpy(force, r_flat_int_surface.force);

#ifdef SURF_DEBUG
  cout << "* emd::flat_int_surface: a flat int. surf. copy constructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

emd::flat_int_surface::~flat_int_surface()
{

  gsl_vector_free(n);
  gsl_vector_free(r0);
  gsl_vector_free(force);

#ifdef SURF_DEBUG
  cout << "* emd::flat_int_surface: flat int. surf. destructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

/* assignment */
const emd::flat_int_surface & emd::flat_int_surface::operator=(
  const emd::flat_int_surface & r_flat_int_surface)
{

  class_id  =  r_flat_int_surface.class_id;
  idx       =  r_flat_int_surface.idx;

  gsl_vector_memcpy(n, r_flat_int_surface.n);
  gsl_vector_memcpy(r0, r_flat_int_surface.r0);
  gsl_vector_memcpy(force, r_flat_int_surface.force);

#ifdef SURF_DEBUG
  cout << "* emd::flat_int_surface: a flat int. surf. copied ("
       << this << ")." << endl;
#endif

  return *this;

}

/***************************************************************************/

emd::spherical_int_surface::spherical_int_surface()
{

  r0     =  gsl_vector_alloc(3);
  force  =  gsl_vector_alloc(3);

#ifdef SURF_DEBUG
  cout << "* emd::spherical_int_surface: a spherical int. surf. constructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

emd::spherical_int_surface::spherical_int_surface(const spherical_int_surface & r_spherical_int_surface)
{

  r0  =  gsl_vector_alloc(3);

  class_id  =  r_spherical_int_surface.class_id;
  idx       =  r_spherical_int_surface.idx;
  rad       =  r_spherical_int_surface.rad;

  gsl_vector_memcpy(r0, r_spherical_int_surface.r0);
  gsl_vector_memcpy(force, r_spherical_int_surface.force);

#ifdef SURF_DEBUG
  cout << "* emd::spherical_int_surface: a spherical int. surf. copy constructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

emd::spherical_int_surface::~spherical_int_surface()
{

  gsl_vector_free(r0);
  gsl_vector_free(force);

#ifdef SURF_DEBUG
  cout << "* emd::spherical_int_surface: spherical int. surf. destructed ("
       << this << ")." << endl;
#endif

}

/***************************************************************************/

/* assignment */
const emd::spherical_int_surface & emd::spherical_int_surface::operator=(
  const emd::spherical_int_surface & r_spherical_int_surface)
{

  class_id  =  r_spherical_int_surface.class_id;
  idx       =  r_spherical_int_surface.idx;
  rad       =  r_spherical_int_surface.rad;

  gsl_vector_memcpy(r0, r_spherical_int_surface.r0);
  gsl_vector_memcpy(force, r_spherical_int_surface.force);

#ifdef SURF_DEBUG
  cout << "* emd::spherical_int_surface: a spherical int. surf. copied ("
       << this << ")." << endl;
#endif

  return *this;

}

