/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * description:                                                            *
 *                                                                         *
 * - Flat and Spherical interacting surfaces                               *
 *                                                                         *
 ***************************************************************************/

#ifndef SURFACE_H
#define SURFACE_H

#include <iostream>

#include <gsl/gsl_vector.h>

#include "gslext.h"
#include "paramdb.h"

//#define SURF_DEBUG

namespace emd
{

  class flat_int_surface
  {
  public:

    int                   class_id;   /* interaction parameters class id */
    int                   idx;        /* index in paramdb caches */

    gsl_vector *          n;          /* normal vector */
    gsl_vector *          r0;         /* an arbitrary point on the surface */

    gsl_vector *          force;      /* the exerted force on the surface */

    /* constructor */
    flat_int_surface();

    /* copy constructor */
    flat_int_surface(const flat_int_surface & r_flat_int_surface);

    /* destructor */
    ~flat_int_surface();

    /* assignment */
    const flat_int_surface & operator=(const flat_int_surface & r_flat_int_surface);

  };

  /***************************************************************************/

  class spherical_int_surface
  {
  public:

    int                   class_id;   /* interaction parameters class id */
    int                   idx;        /* index in paramdb caches */

    double                rad;        /* radius */
    gsl_vector *          r0;         /* center */

    gsl_vector *          force;      /* the exerted force on the surface */

    /* constructor */
    spherical_int_surface();

    /* copy constructor */
    spherical_int_surface(const spherical_int_surface & r_spherical_int_surface);

    /* destructor */
    ~spherical_int_surface();

    /* assignment */
    const spherical_int_surface & operator=(const spherical_int_surface & r_spherical_int_surface);

  };

  /***************************************************************************/

};

#endif /* SURFACE_H */
