/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <stdlib.h>
using namespace std;

int main(int argc, char ** argv)
{

  string output_file = "new_uni.uni";

  int      class_id       =  1;
  double   rad_1          =  0.5;
  double   rad_2          =  0.5;
  double   rad_3          =  2.0;
  double   sigma          =  1.0;
  double   pad            =  0.0;
  double   x_noise_amp    =  0.1*sigma;
  double   y_noise_amp    =  0.1*sigma;
  double   z_noise_amp    =  0.1*sigma;
  double   ang_noise_amp  =  10.0;
  double   vel_noise_amp  =  0.1;


  double   x_spc     = 2*rad_1 + 0.5*sigma;
  double   y_spc     = 2*rad_2 + 0.5*sigma;
  double   z_spc     = 2*rad_3 + 0.5*sigma;

  int      N_x       = 8;
  int      N_y       = 8;
  int      N_z       = 8;

  double   max_len = 0;
  double   max_spc = 0;

  if ( x_spc > max_spc) max_spc = x_spc;
  if ( y_spc > max_spc) max_spc = y_spc;
  if ( z_spc > max_spc) max_spc = z_spc;

  if ( (max_spc*N_x) > max_len) max_len = (max_spc*N_x);
  if ( (max_spc*N_y) > max_len) max_len = (max_spc*N_y);
  if ( (max_spc*N_z) > max_len) max_len = (max_spc*N_z);


  x_spc = max_spc;
  y_spc = max_spc;
  z_spc = max_spc;

  ofstream of(output_file.c_str());

  of << " ###########################\n"
     << " # saved universe snapshot #\n"
     << " ###########################\n"
     << "\n";

  of << "param_file       /home/dt/codes/cpp/emd/workspace/param.txt\n" << endl;

  of << "x_periodic       " << -(max_len/2) << "  " << (max_len/2) << endl;
  of << "y_periodic       " << -(max_len/2) << "  " << (max_len/2) << endl;
  of << "z_periodic       " << -(max_len/2) << "  " << (max_len/2) << endl;

  of << "interior_point           0  0  0" << endl;
  of << "global_electrical_field  0  0  1" << endl;

  of << endl;

  of << "rigid_body_num  " << (N_x*N_y*N_z) << endl;

  of << endl;

  srand48(time(0));

  for (int n_x=0; n_x<N_x; n_x++)
    for (int n_y=0; n_y<N_y; n_y++)
      for (int n_z=0; n_z<N_z; n_z++)
      {

        of << "rigid_body    " << class_id << "   1" <<"    2"<< endl
           //<< "              " << 46 << endl
           << "              " << (x_spc/2 + n_x*x_spc + x_noise_amp*(2*drand48()-1) - max_len/2) << "     "
                               << (y_spc/2 + n_y*y_spc + y_noise_amp*(2*drand48()-1) - max_len/2) << "     "
                               << (z_spc/2 + n_z*z_spc + z_noise_amp*(2*drand48()-1) - max_len/2) << endl
           << "              " << (ang_noise_amp*(2*drand48()-1)) << "     "
                               << (ang_noise_amp*(2*drand48()-1)) << "     "
                               << (ang_noise_amp*(2*drand48()-1)) << endl
           << "              " << (0.5*vel_noise_amp*(2*drand48()-1)) << "     "
                               << (0.5*vel_noise_amp*(2*drand48()-1)) << "     "
                               << (2.0*vel_noise_amp*(2*drand48()-1)) << endl
           << "              " << 0 << "   " << 0 << "   " << 0 << endl;

        of << endl;

        of << "rigid_body_dipole   0  0  1" << endl;

        of << endl;

      }


}
