/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GMDUNIVERSELOADER_H
#define GMDUNIVERSELOADER_H

#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>


#include <gsl/gsl_vector.h>

#include "quaternion/gsl_quaternion.h"

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "surface.h"
#include "paramdb.h"
#include "stringext.h"
#include "moluniverse.h"

/***************************************************************************/

//#define UNIVERSELOADER_DEBUG

/***************************************************************************/

namespace emd
{

namespace uni
{

namespace loader
{

  /* parsers */
  int parse_param_file(const std::string & param_string);
  int parse_rigid_body_num(const std::string & param_string);
  int parse_flat_int_surf_num(const std::string & param_string);
  int parse_sph_int_surf_num(const std::string & param_string);
  int parse_rigid_body(const std::string & param_string);
  int parse_rigid_body_dipole(const std::string & param_string);
  int parse_flat_int_surf(const std::string & param_string);
  int parse_sph_int_surf(const std::string & param_string);
  int parse_x_periodic(const std::string & param_string);
  int parse_y_periodic(const std::string & param_string);
  int parse_z_periodic(const std::string & param_string);
  int parse_interior_point(const std::string & param_string);
  int parse_global_electrical_field(const std::string & param_string);
  //{{{restart : declaration of parse_time
  int parse_time(const std::string & param_string);
  //}}}
  /* actions */
  typedef int (*act)(const std::string & input_string);

  static int parsed_rigid_bodies   = 0;           /* the count of parsed rigid bodies */
  static int parsed_flat_int_surfs = 0;           /* the count of parsed flat int. surfs. */
  static int parsed_sph_int_surfs  = 0;           /* the count of parsed sph. int. surfs. */
  //{{{restart : definition of t_start
  extern double t_start               ;
  //}}}
  //changed
  static bool ext_param=false;
  static std::string ext_param_file;
  
  /* keywords */
  static std::string KEYWORDS[] =
  {
    "param_file",
    "rigid_body_num",
    "flat_int_surf_num",
    "sph_int_surf_num",
    "rigid_body",
    "rigid_body_dipole",
    "flat_int_surf",
    "sph_int_surf",
    "x_periodic",
    "y_periodic",
    "z_periodic",
    "interior_point",
    "global_electrical_field",
    //{{{restart: "time" field
    "time"
    //}}}
  };

  /* actions */
  static act ACTIONS[] =
  {
    &parse_param_file,
    &parse_rigid_body_num,
    &parse_flat_int_surf_num,
    &parse_sph_int_surf_num,
    &parse_rigid_body,
    &parse_rigid_body_dipole,
    &parse_flat_int_surf,
    &parse_sph_int_surf,
    &parse_x_periodic,
    &parse_y_periodic,
    &parse_z_periodic,
    &parse_interior_point,
    &parse_global_electrical_field,
    //{{{restart : parse_time
    &parse_time
    //}}}
  };

  /***************************************************************************/

  /* parses the input file and initializes the uni */
  int load_file(const std::string & input_file_name, const bool& clear = true);
  void set_ext_param_file(std::string inp_param_file);

}; /* loader */

}; /* uni */

}; /* emd */

#endif /* GMDUNIVERSELOADER_H */
