/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef EMDUNIVERSESAVER_H
#define EMDUNIVERSESAVER_H

#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include "quaternion/gsl_quaternion.h"

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "surface.h"
#include "paramdb.h"
#include "stringext.h"
#include "moluniverse.h"
#include "dynamics.h"
#include "geom.h"

/***************************************************************************/

#define UNIVERSESAVER_DEBUG

/***************************************************************************/

namespace emd
{

namespace uni
{

namespace saver
{

  static const int      DEF_PREC   = 12;
  static const int      DEF_WIDTH  = 22;

  /***************************************************************************/

  /* saves the current snapshot of unierse to file */
  int save_universe_to_file(const std::string & out_file_name);

  /* saves the current snapshot of unierse to file (only the given indices) */
  int save_universe_to_file(const std::vector<rigid_body_ref> & indices,
    const std::string & out_file_name);

  /* saves the current snapshot of unierse to file */
  int save_scaled_universe_to_file(const double & scale_factor,
    const std::string & out_file_name);

  /* saves the current snapshot of unierse as a Povray file (all) */
  int save_universe_to_povray_file(const std::string & out_file_name);

  /* saves the current snapshot of unierse as a Povray file (given indices) */
  int save_universe_to_povray_file(const std::vector<rigid_body_ref> & indices,
    const std::string & out_file_name);

  /* save nose-hoover chain variables to file */
  int save_nhc_vars_to_file(const std::string & out_file_name);

}; /* saver */

}; /* uni */

}; /* emd */

#endif /* EMDUNIVERSESAVER_H */
